/*
 * ata.c - An ATA/IDE driver
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 04/01/02
 *
 * Last CVS commit : $Id: ata.c,v 1.28 2002/02/19 16:46:31 jalh Exp $	
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * This file impliments the ATA2 to ATA5 specs which can be got from
 * http://www.t13.org. ATA5 spec is avaliable at
 * ftp://fission.dt.wdc.com/pub/standards/x3t13/project/d1321r3.pdf
 *
 * Notes/Fixme's:
 * There's an unexpected interrupt generated when setting multiple sector reads
 * on my hdd. - Mark
 * 12/10/01 Mark: fix the ATA_TEST code to use gettimeofday()
 *
 * Changelog:
 * 07/07/01 Mark: Added an optional cache of the task-register
 * 05/11/01 Mark: Added subsystems code and updated the timing functions.
 * 12/12/01 Mark: Cleaned up a bit
 */

#include <common.h>
#include <byte.h>
#include <ata.h>
#include <kmalloc.h>
#include <assert.h>
#include <printk.h>
#include <io.h>
#include <irq.h>
#include <cpu.h>
#include <delay.h>
#include <stdio.h>
#include <signals.h>
#include <device.h>
#include <block.h>
#include <time.h>
#include <pit.h>
#include <subsystems.h>
#include "ata-stuff.h"

// define this to enable caching for io (should be enabled unless you have
// problems with it...)
#define CACHE_TASK_REG

#define DRV_TIMEOUT 20000
#define CNT_TIMEOUT 100000


// FIXME: Autodetect sometime? - Mark
__uint16 bases[] = {0x1F0, 0x170};		// Channels to controllers
uchar IRQs[] = {14, 15};			// IRQ's for controllers

// Pointers to all of the cnt structs
struct cntStr *cnts[ATA_MAX_CNT] = {NULL};

/* function: errRet HDDWait(struct cntStr *cnt, __uint8 drive)
 * description: Wait for the hard drive to be ready (clear the busy bit and set
 * the ready bit).
 */
errRet HDDWait(struct cntStr *cnt, __uint8 drive)
{
	struct drvStr *drv = cnt->drive[drive];
	__uint32 timeout;
	assert(drv);

	timeout = drv->timeout;

	for(; timeout && !isPollReady(cnt) && isPollBusy(cnt); timeout--)
		udelay(1);

	if(timeout)					// No timeout
		return SUCCESS;

	if(aborted(cnt)) {		// Aborted command
		DPRINT("Aborted\n");
		return EABORT;
	}

	DPRINT("Timeout\n");

	return ETIMEOUT;
}

// Waits for a HDDDone to happen otherwise times out
errRet HDDDoneWait(struct cntStr *cnt, __uint8 disk)
{
	__uint32 timeout;
	struct drvStr *drv;

	assert(cnt);
	drv = cnt->drive[disk];
	assert(drv);

	for(timeout = drv->timeout; timeout; udelay(1), timeout--) {
		if(cnt->doing) {
			cnt->doing = FALSE;
			timeout = drv->timeout;
		}
		if(cnt->done)
			return SUCCESS;
	}

	DPRINT("Timeout while waiting for interrupt (cnt %u)\n", cnt->num);
	cnt->handler = NULL;		// Reset handler cos we failed
	return ETIMEOUT;		// Timeout
}

////////////////////////////////////////////////////////////////////////////////
// ATA Interrupt handlers
////////////////////////////////////////////////////////////////////////////////
// Just sets handler to NULL
void HDDDone(struct cntStr *cnt)
{
	cnt->handler = NULL;
	cnt->done = TRUE;
}

// Standard interrupt for data transfer from the HDD
void HDDReadTransfer(struct cntStr *cnt)
{
	struct drvStr *drv;
	ulong read;

	assert(cnt);

	cnt->doing = TRUE;

	drv = cnt->drive[cnt->curDrv];
	assert(drv);

	read = drv->sectsPerIntr;

	if(!read) {				// Catch NULL-reads
		DPRINT("Moooo: NULL READ (someone's got BSE...)!!\n");
		goto endHDDDoDataTransfer;
	}

	if(cnt->sectsToDo < read)		// Partial read?
		read = cnt->sectsToDo;

	cnt->sectsToDo -= read;

	DPRINT("sectsPerIntr: %u, left to read: %u, read: %u, buf: %p\n",
		drv->sectsPerIntr, cnt->sectsToDo, read, cnt->dataBlock);

	if(drv->info.longIO)
		insl(cnt->addr, cnt->dataBlock, (BYTES_PER_SECT * read) / 4);
	else
		insw(cnt->addr, cnt->dataBlock, (BYTES_PER_SECT * read) / 2);

	if(!cnt->sectsToDo) {
endHDDDoDataTransfer:
#if CONFIG_ATA & OPT_IS_DEBUG
		printk(KINFO, ",");
#endif /* CONFIG_ATA & OPT_IS_DEBUG */
		HDDDone(cnt);
	}

	cnt->dataBlock += BYTES_PER_SECT * read;
}

void HDDWriteTransfer(struct cntStr *cnt)
{
	struct drvStr *drv;
	ulong write;

	assert(cnt);

	cnt->doing = TRUE;

	drv = cnt->drive[cnt->curDrv];
	assert(drv);

	if(!cnt->sectsToDo) {
		HDDDone(cnt);
		DPRINT("Looks like that write ended then\n");
		return;
	}

	write = drv->sectsPerIntr;
	if (cnt->sectsToDo < write)
		write = cnt->sectsToDo;

	cnt->sectsToDo -= write;

	DPRINT("sectsPerIntr: %u, left to write: %u\n", drv->sectsPerIntr,
			cnt->sectsToDo);

	if(drv->info.longIO)		// Can use 32-bit transfers?
		outsl(cnt->addr, cnt->dataBlock, (BYTES_PER_SECT * write) / 4);
	else
		outsw(cnt->addr, cnt->dataBlock, (BYTES_PER_SECT * write) / 2);

	cnt->dataBlock += BYTES_PER_SECT * write;
}

spinlock_t ATAIntLock = INIT_UNLOCKED;

/* Interrupt handler, just calls off the current handler */
IRQ_HND(ATAIntrHandler, struct cntStr *cnt, irq)
{
	assert(cnt);

	spinlock(&ATAIntLock);

#if CONFIG_ATA & OPT_IS_DEBUG
	printk(KINFO, ".");
#endif

	if(cnt->handler)
		cnt->handler(cnt);
	else
		DPRINT("Unexpected interrupt from controller %u, hard drive %u"
				"... shoot Mark\n", cnt->num, cnt->curDrv);
	getStatus(cnt);		// Make sure it clears the interrupt pending

	spinlockUnlock(&ATAIntLock);
}

////////////////////////////////////////////////////////////////////////////////
// HDD Commands
////////////////////////////////////////////////////////////////////////////////

/* Issues a raw command to the AT register based on the data given (now
 * impliments caching for faster register access - Mark)
 */
void ATAIssue(struct cntStr *cnt, __uint8 len, __uint8 sect, __uint16 cyl,
		bool LBA, bool drive, __uint8 head, __uint8 command)
{
	__uint8 cyls[2], drvReg;
	cyls[0] = cyl & 0xFF;
	cyls[1] = cyl >> 8;
	drvReg = 0xA0 | (LBA << 6) | (drive << 4) | head;

	assert(head <= 0xF);
	DPRINT("Issuing: len %u; sect %u; cyl %u; cmd %2x; LBA %u; drive %u; "
			"head %u\n", len, sect, cyl, command, LBA, drive, head);

#ifdef CACHE_TASK_REG
#define PRINT_OUT_CACHE(a) DPRINT("%u not cached\n", a)

#define OUT_CACHE(a,b,c) do {					\
	if(!cnt->cacheValid || (cnt->cache[a] != (b)))		\
		PRINT_OUT_CACHE(a);				\
		outb(cnt->cache[a] = (b), cnt->addr + (c));	\
	} while(0)
#else
#define PRINT_OUT_CACHE(a) do {} while(0)
#define OUT_CACHE(a,b,c) outb(b, cnt->addr + (c))
#endif

	OUT_CACHE(0, len, SECT_COUNT);
	OUT_CACHE(1, sect, SECT_NUM);
	OUT_CACHE(2, cyls[0], CYL_LSB);
	OUT_CACHE(3, cyls[1], CYL_MSB);
	OUT_CACHE(4, drvReg, DRV_HEAD);
	cnt->cacheValid = TRUE;

#undef PRINT_OUT_CACHE
#undef OUT_CACHE

	// Don't cache the command!
	outb(command, cnt->addr + CMD_REG);	// And set the ball rolling :)

	udelay(1);			// Wait for BSY to get set
}

// Does a command and sets stuff up for it :)
errRet HDDDoCmd(__uint32 start, __uint16 len, void *buf, deviceCmd cmd,
		struct cntStr *cnt, __uint8 drive)
{
	struct drvStr *drv;
	__uint16 cyl = 0;
	__uint8 head = 0, sect = 0;
	bool LBA = FALSE;

	if(len > 256)			// We only support upto 256 sectors..
		return EINVALID;

	assert(cnt);

	drv = cnt->drive[drive];
	assert(drv);

	cnt->curDrv = drive;

	if(drv->init && !drv->info.LBA) {  // no LBA (but drive is initalized)
		__uint16 t = start % (drv->info.heads * drv->info.sects);
		cyl = start / (drv->info.heads * drv->info.sects);
		head = t / drv->info.sects;
		sect = t % drv->info.sects + 1;
	} else if(drv->init) {		// LBA and the drive is initalized :))
		sect = start & 0xFF;
		cyl = (start >> 8) & 0xFFFF;
		head = (start >> 24) & 0xF;
		LBA = TRUE;
	}

	cnt->dataBlock = buf;
	cnt->sectsToDo = len;
	cnt->done = FALSE;
	cnt->doing = FALSE;

	switch(cmd) {
	case DEV_WRITE:
		cnt->handler = HDDWriteTransfer;
		// Do a multi-write if at all possible
		cnt->req = (drv->sectsPerIntr > 1)?
						ATA_CMD_MULTI_WRITE :
						ATA_CMD_WRITE;
		break;

	case DEV_READ:
		cnt->handler = HDDReadTransfer;
		// Do a multi-read if at all possible
		cnt->req = (drv->sectsPerIntr > 1)?
						ATA_CMD_MULTI_READ :
						ATA_CMD_READ;
		break;

	case DEV_DETECT_PARAMS:
		cnt->handler = HDDReadTransfer;
		cnt->req = ATA_CMD_ID_DRV;
		len = 0;
		break;

	default:
		DPRINT("Unsupported operation issued on device %lu:%lu\n",
				major_ata, ATA_MINOR(cnt, drive));
		return EOHHBUGG;
	}

	// Output the command (could be split to a separate thread for sending
	// requests to the HDD

	if(HDDWait(cnt, drive) != SUCCESS)
		return ETIMEOUT;

	ATAIssue(cnt, len & 0xFF, sect, cyl, LBA, drive, head, cnt->req);

	if(cmd == DEV_WRITE)		// We have to kick-start it
		cnt->handler(cnt);

	/* FIXME: Do we really want this? - Mark */
	return HDDDoneWait(cnt, drive);
}

// Resets the specified drive on the specified controller
errRet HDDReset(struct cntStr *cnt, __uint8 disk)
{
	cnt->handler = HDDDone;
	outb(ATA_CMD_RECALIB, cnt->addr + CMD_REG);

	return HDDDoneWait(cnt, disk);
}

// Detects the drives parameters
errRet HDDDetectParam(struct cntStr *cnt, __uint8 disk)
{
	struct drvStr *drv = cnt->drive[disk];

	return HDDDoCmd(0, 1, &drv->info, DEV_DETECT_PARAMS, cnt, disk);
}

errRet ATASetMulti(struct cntStr *cnt, __uint8 disk, __uint8 sects)
{
	ATAIssue(cnt, sects, 0, 0, 0, disk, 0, ATA_CMD_SET_MULTI);

	return HDDWait(cnt, disk);
}

// Sets features
inline errRet ATASetFeat(struct cntStr *cnt, __uint8 disk, __uint8 feat)
{
	outb(feat, cnt->addr + FEAT_REG);
	outb(ATA_CMD_SET_FEATURE, cnt->addr + CMD_REG);
	// Cache is valid because we are not affecting any of the cached
	// registers

	return HDDWait(cnt, disk);
}

inline errRet fillInGeom(struct cntStr *cnt, __uint8 disk,
		struct block_geom *str)
{
	struct drvStr *drv = cnt->drive[disk];

	str->cylinders = drv->info.cyls;
	str->heads = drv->info.heads;
	str->sectors = drv->info.sects;
	str->size = drv->capacity;

	return SUCCESS;
}

inline errRet fillInOp(struct cntStr *cnt, __uint8 drive,
		struct blk_cmd *command)
{
	errRet ret;
	ulong i = command->length, read;
	void *buffer = command->data;
	ulong start = command->start;

	while(i) {
		read = (i > 256)?256:i;
		i -= read;
		ret = HDDDoCmd(start, read, buffer, command->type, cnt, drive);

		if(ret != SUCCESS)
			return ret;

		start += read;
		buffer += read * BYTES_PER_SECT;
	}

	return SUCCESS;
}

deviceHnd(ATADo)
{
	struct cntStr *cnt = cnts[(num->minor / 64) / ATA_MAX_DRV];
	__uint8 drive = (num->minor / 64) % ATA_MAX_DRV;
	struct blk_cmd *command = data;

	switch(command->type) {
	case DEV_GET_GEOM:
		return fillInGeom(cnt, drive, (void *)command->data);

	default:
		return fillInOp(cnt, drive, command);
	}
}

////////////////////////////////////////////////////////////////////////////////
// ATA Commands
////////////////////////////////////////////////////////////////////////////////

// Fixes strings from the detect parameters command
void paramFix(char *str, ulong len)
{
	ulong i;
	assert(~len & 1);			// Must be a multiple of 2

	/* OK, so ATA is pretty stupid about this, infact, I'd go as far as
	 * saying this is totally fucked up... *Everything* in ATA is x86-biast,
	 * IE Little endian, *apart from* the strings reported by the detect
	 * parameters command, which are Little endian... - A pretty pissed off
	 * Mark
	 */
	for(i = 0; i < len; i += 2) {
		__uint16 *t = (__uint16 *)(str + i);
		*t = beToCPU16(*t);
	}

	chompl(str, len);
}

// Adds and detects a new drive to the specified interface
errRet ATAAddDrv(struct cntStr *cnt, __uint8 disk)
{
	struct drvStr *drive;
	struct devNum t;
	__uint8 i;

	assert(cnt);
	assert(disk < ATA_MAX_DRV);

	// Assign memory for the drive
	drive = kmalloc(sizeof(struct drvStr));

	cnt->drive[disk] = drive;
	drive->timeout = DRV_TIMEOUT;
	drive->init = FALSE;		// Not init'd yet
	drive->sectsPerIntr = 1;	// Temp value
	drive->info.longIO = 0;

	if(HDDReset(cnt, disk) != SUCCESS) {	// Reset the disk first
		DPRINT("HDD %d reset failed. Drive prolly doesnt exist\n",
				ATA_MINOR(cnt, disk) / 64);
		return ENOEXIST;	// No error in not having a drive
	}

	if(HDDDetectParam(cnt, disk) != SUCCESS) {	// Detect params failed
		DPRINT("HDD %d detect param failed. Drive prolly doesnt exist\n"
				, ATA_MINOR(cnt, disk) / 64);
		return ENOEXIST;
	}

	printk(KINFO, "hd%c, ", 'a' + ATA_MINOR(cnt, disk) / 64);

	paramFix(drive->info.model, 40);
	if(drive->info.model)
		printk(KINFO, "%s. ", drive->info.model);

	paramFix(drive->info.serial, 20);	// Dunno why we do these - Mark
	paramFix(drive->info.firmware, 8);

	// Print our ATA version number
	for(i = 0; i < sizeof(__uint16) * CHAR_BIT - 1; i++)
		if(bt(&drive->info.ATAMajor, i))
			printk(KINFO, "ATA-%u, ", i);

	if(!drive->info.cyls || !drive->info.heads || !drive->info.sects) {
		printk(KINFO, "%u/%u/%u: CD-ROM?\n", drive->info.cyls,
				drive->info.heads, drive->info.sects);
		return EOHHBUGG;
	}

	drive->capacity = drive->info.cyls * drive->info.heads *
							drive->info.sects;

	/* If we have LBA and the c*h*s values are within 10% of the LBA
	 * capacity value, we use that, otherwise use c*h*s as the capacity
	 * - Mark
	 */
	if(drive->info.LBA && ((ulong)(drive->capacity / 10) ==
					(ulong)(drive->info.LBACap / 10)))
		drive->capacity = drive->info.LBACap;

	printk(KINFO, "%u/%u/%u %luMb, ", (uint)drive->info.cyls,
			(uint)drive->info.heads, (uint)drive->info.sects,
			(drive->capacity * BYTES_PER_SECT) / (1024 * 1024));

	if(drive->info.bufferSize)
		printk(KINFO, "%uKb cache, ",
				(uint)drive->info.bufferSize >> 1);

	if(drive->info.LBA)
		printk(KINFO, "LBA, ");

	if(drive->info.sectsPerIntr) {
		cnt->handler = HDDDone;
		drive->sectsPerIntr = drive->info.sectsPerIntr;
		printk(KINFO, "%u spi, ", (uint)drive->sectsPerIntr);

		/* Attempt to set the multisector reads up correctly */
		if(ATASetMulti(cnt, disk, drive->sectsPerIntr) == EABORT) {
			DPRINT("Failed :/\n");
			drive->sectsPerIntr = 1;

			/* If the previous failed, we try to reset the drive
			 * back to it's previous state, if this doesn't work, we
			 * just say fuck it and carry on - Mark
			 */
			if(ATASetMulti(cnt, disk, 0) == EABORT)
				DPRINT("Uhh, resetting sectsPerIntr back to 0 "
					"didn't work (shouldn't happen)\n");
		}

		cnt->handler = NULL;
	}

	drive->init = TRUE;		// Init'd properly now

	t.major = major_ata;
	t.minor = (disk + cnt->num * ATA_MAX_DRV) * 64;	// Steps of 64
	devReg(&t, GET_OFFSET(ATADo));

	printk(KINFO, "\b\b.\n");

	return SUCCESS;
}

// Removes a disk, free's memory
void ATADelDrv(struct cntStr *cnt, __uint8 disk)
{
	DPRINT("Removing device %lu:%lu\n", major_ata, ATA_MINOR(cnt, disk));
	assert(cnt->drive[disk]);
	kfree(cnt->drive[disk]);
	cnt->drive[disk] = NULL;
}

errRet cntReset(struct cntStr *cnt)
{
	__uint32 timeout = cnt->timeout;

	outb(4, cnt->addr + DOR);	// Reset and disable IRQ's
	udelay(10);			// Wait a sec
	outb(0, cnt->addr + DOR);	// Enable IRQ's

	for(; timeout && isPollBusy(cnt); timeout--)
		udelay(1);

	return SUCCESS;
}

// Removes a controller, frees memory
void ATADelCnt(__uint8 num)
{
	assert(cnts[num]);
	delHandler(cnts[num]->IRQ, ATAIntrHandler);
	kfree(cnts[num]);
	cnts[num] = NULL;		// NULL-ify it :)
}

// Configures a controller
errRet ATAConfCnt(__uint8 num)
{
	__uint8 disk;

	DPRINT("Adding controller %d\n", num);

	// Allocate memory
	cnts[num] = (struct cntStr *) kmalloc(sizeof(struct cntStr));

	kmemset((uchar *)cnts[num], 0, sizeof(struct cntStr));	// 0 the struct

	cnts[num]->num = num;
	cnts[num]->IRQ = IRQs[num];
	cnts[num]->addr = bases[num];
	cnts[num]->timeout = CNT_TIMEOUT;
	cnts[num]->cacheValid = FALSE;		// Cache is unintatlized
	addHandler(cnts[num]->IRQ, ATAIntrHandler, 1, (void *)cnts[num]);

	// Reset the controller (test if it exists)
	if(cntReset(cnts[num]) != SUCCESS) {
		DPRINT("Controller %d doesn't look like it exists\n", num);
		ATADelCnt(num);
		return ENOEXIST;
	}

	DPRINT("Controller %d exists, getting drive info from it\n", num);
	for(disk = 0; disk < ATA_MAX_DRV; disk++) {
		if(ATAAddDrv(cnts[num], disk) != SUCCESS)
			ATADelDrv(cnts[num], disk);
	}

	// If there's no disks, there's not really much point in keeping the cnt
	// struct around
	for(disk = 0; disk < ATA_MAX_DRV; disk++)
		if(cnts[num]->drive)
			return SUCCESS;

	ATADelCnt(num);

	return SUCCESS;
}

#ifdef ATA_TIME

#define MB_BURN 32
#define MB_S 2
#define MB_READ (MB_S * MB_BURN)
#define MB ((MB_S * 1024 * 1024) / BYTES_PER_SECT)

errRet ATATime(__uint8 disk)
{
	ulong i, millisec;
	void *v;
	struct devNum device;
	struct blk_cmd cmd;
	struct timeval tStart, tStop;

	// Test to see that the controller and the disk exist
	if(!(cnts[disk / ATA_MAX_DRV] && cnts[disk / ATA_MAX_DRV]->drive[disk %
			ATA_MAX_DRV]))
		return ENOEXIST;

	if(subsysIsLoaded("time") == FALSE)
		return ENODEP;

	printk(KINFO, "hd%c:\n", 'a' + disk);

	if(gettimeofday(&tStart) != SUCCESS)
		return EOHHBUGG;

	printk(KINFO, "Testing read with blk_read: ");
	for(i = 0; i < MB_BURN; i++) {
		printk(KINFO, ".");
		v = blk_read(major_ata, disk * 64, 0, MB);

		kfree(v);
		if(!v) {
			DPRINT("Error with read\n");
			return EOHHBUGG;
		}
	}

	if(gettimeofday(&tStop) != SUCCESS)
		return EOHHBUGG;

	millisec = ((tStop.tv_sec - tStart.tv_sec) * 1000)
			+ ((tStop.tv_usec - tStart.tv_usec) / 1000);

	printk(KINFO, "\n%luMB in %lu.%3u seconds = %lu.%2luMB/s\n", MB_READ,
			millisec / 1000, millisec % 1000,
			MB_READ * 1000 / millisec,
			MB_READ * 1000000 / millisec);

	printk(KINFO, "Testing read with devices and a single kmalloc: ");

	if(gettimeofday(&tStart) != SUCCESS)
		return EOHHBUGG;

	v = kmalloc(MB * BYTES_PER_SECT);
	if(!v) {
		DPRINT("Error with kmalloc\n");
		return EOHHBUGG;
	}

	device.major = major_ata;
	device.minor = disk * 64;
	cmd.start = 0;
	cmd.length = MB;
	cmd.data = v;
	cmd.type = DEV_READ;

	for(i = 0; i < MB_BURN; i++) {
		printk(KINFO, ".");
		if(devCall(&device, &cmd) != SUCCESS) {
			DPRINT("Read was not a SUCCESS");
			return EOHHBUGG;
		}
	}

	kfree(v);
	if(gettimeofday(&tStop) != SUCCESS)
		return EOHHBUGG;

	millisec = ((tStop.tv_sec - tStart.tv_sec) * 1000)
			+ ((tStop.tv_usec - tStart.tv_usec) / 1000);

	printk(KINFO, "\n%luMB in %lu.%3u seconds = %lu.%2luMB/s\n", MB_READ,
			millisec / 1000, millisec % 1000,
			MB_READ * 1000 / millisec,
			MB_READ * 1000000 / millisec);

	printk(KINFO, "Testing read directly with a single kmalloc: ");

	if(gettimeofday(&tStart) != SUCCESS)
		return EOHHBUGG;

	v = kmalloc(MB * BYTES_PER_SECT);
	if(!v) {
		DPRINT("Error with kmalloc\n");
		return EOHHBUGG;
	}

	for(i = 0; i < MB_BURN; i++) {
		ulong j = MB;
		printk(KINFO, ".");
		while(j) {
			ulong read = (j > 256)?256:j;
			if(HDDDoCmd(MB - j, read, (v + (MB - j) *
					BYTES_PER_SECT), DEV_READ,
					cnts[disk / ATA_MAX_DRV],
					disk % ATA_MAX_DRV) != SUCCESS) {
				DPRINT("Read was not a SUCCESS");
				return EOHHBUGG;
			}
			j -= read;
		}
	}

	kfree(v);

	if(gettimeofday(&tStop) != SUCCESS)
		return EOHHBUGG;

	millisec = ((tStop.tv_sec - tStart.tv_sec) * 1000)
			+ ((tStop.tv_usec - tStart.tv_usec) / 1000);

	printk(KINFO, "\n%luMB in %lu.%3u seconds = %lu.%2luMB/s\n", MB_READ,
			millisec / 1000, millisec % 1000,
			MB_READ * 1000 / millisec,
			MB_READ * 1000000 / millisec);

	return SUCCESS;
}

#endif

initFunc __preInitFunc ATAInit()
{
	__uint8 i = 0;
	printk(KINFO, "Init'ing the HDD's\n");

//	for(; i < ATA_MAX_CNT; i++)
		ATAConfCnt(i);

	SUBSYSTEM_ON("ATA", NULL, NULL, NULL);

#ifdef ATA_TIME
	for(i = 0; i < ATA_MAX_CNT * ATA_MAX_DRV; i++)
		ATATime(i);
#endif

	return SUCCESS;
}
