/*
 * keyboard.c - A keyboard driver
 *
 * Added to the Zealos project and (C) by Daniel Abraham
 * Also (C) by Reggy
 *
 * Current maintainer: Reggy <reggy@ekky.org>
 * Status: Maintained
 * Last update: 04/01/02
 *
 * Last CVS commit: $Id: keyboard.c,v 1.24 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * Programed to the keyboard spec found at
 * www.nondot.org/sabre/os/H4HumanInterface/keyboard2.txt.
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 11/08/01 silas: now it can kgetc etc (;
 * 10/09/01 reggy: Well it seems keyboard selftest is (no longer)
 *		   supported by keyboard, will be testing this on old
 *		   computer soon.
 * 11/09/01 reggy: fixed (i hope) the Keyboard wants data resend.
 */

#include <common.h>
#include <keyboard.h>
#include <kmalloc.h>
#include <assert.h>
#include <printk.h>
#include <io.h>
#include <irq.h>
#include <stdio.h>
#include <signals.h>
#include <cpu.h>
#include <device.h>
#include "keyboard.h"
#include "keymaps/us.kbm"
#include <sysrq.h>



/* Function: keybSelftest
 * Description: This is a function to do the keyboard selftest
 */

bool keybSelftest()
{
	outb(0xaa, KEYB_INTERFACEPORT);
	delay(2);
	if(keybState & KEYB_WANTS_DATA_RESEND)
		outb(0xaa, KEYB_INTERFACEPORT);

        if(keybState & KEYB_SELFTEST_SUCCESS) {
                printk(KINFO, " Success\n");
		return TRUE;
        } else {
                printk(KINFO, " Failure\n");
		return FALSE;
        }
}

bool keybWait()
{
	while((inb(KEYB_INTERFACEPORT) & 2))
		/* do nada */;

	return TRUE;
}

/*
 * Function: keybClearBuf
 * Description: This function is called from within the kernel to clear out the
 * keyboard event buffer. This means all buffered keypresses will be lost, use
 * this with caution.
 */
void keybClearBuf()
{
	scanPos = 0;
	readPos = 0;
	kmemset(scanBuf, 0, sizeof(scanBuf));
}

/*
 * Function: keybHandler
 * Description: This is the signal handler for the keyboard, its triggered
 * whenever a key has been pressed or released, or when there is an error or a
 * buffer overflow, ie, too many keys pressed.
 */
IRQ_DEF_HND(keybHandler)
{
	// Read input *before* we check for buffer overflow so we clear out the
	// interrupt
	uchar inp = inb(KEYB_IOPORT);

#ifdef CONFIG_KEYBOARD & OPT_IS_DEBUG
	uchar inp2 = inb(KEYB_ACKPORT);
	uchar inp3 = inb(KEYB_INTERFACEPORT);
#endif

#ifdef CONFIG_KEYBOARD & OPT_IS_DEBUG
	printk(KINFO, "INPUT FROM KEYB_IOPORT: ");
	printk(KINFO, "%u - 0x%x\n", inp, inp);

	printk(KINFO, "INPUT FROM KEYB_ACKPORT: ");
	printk(KINFO, "%u - 0x%x\n", inp2, inp2);

	printk(KINFO, "INPUT FROM KEYB_INTERFACEPORT: ");
	printk(KINFO, "%u - 0x%x\n", inp3, inp3);

	printk(KINFO, "This should be:");
	printk(KINFO, " %c", kbmap[scanBuf[readPos++]][keybMapMod]);
#endif

	switch (inp)
	{
	case 0x00: /* Keyboard error - Too many keys */
		beep();
		return; // why would anybody else care?!
	case 0xFE:
		keybState |= KEYB_WANTS_DATA_RESEND;
		return;
	case 0xFC:
		/* Basic Assurance Test (BAT) Failed */
		DPRINT("Keyboard Error: Basic Assurance Test Failed!!\n");
		return;
	case 0x55:
		/* Keyboard Selftest Success */
		printk(KINFO, "Keyboard Selftest Success!\n");
		keybState |= KEYB_SELFTEST_SUCCESS;
		return;
	case 0xFF:
		/* Keyboard ERROR! */
		DPRINT("Keyboard Error: Error!! :)\n");
		return;
	case 58:			// caps lock
		if(keybMapMod < 2)
			keybMapMod += 2;
		else
			keybMapMod -= 2;
		keybLedState ^= LED_CAPS;
		keybSetLeds();
		return;
	case 56:
		keybCtrlState +=2;	// Alt pressed
		return;
	case 184:
		keybCtrlState -=2;	// Alt released
		return;

	case 29:
		keybCtrlState +=1;	// CTRL pressed
		return;
	case 157:
		keybCtrlState -=1;	// CTRL released
		return;

	case 224:
		if(keybCtrlState == 3);	// Del Pressed, checking CTRL+ALT
			reboot(REBOOT_WARM);
		break;

	case 42: case 54:
		keybMapMod += 1;        // shift pressed
		if(keybCtrlState == 3)	// Shift pressed, checking CTRL+ALT
			halt();
		return;
	case 170: case 182:
		keybMapMod -= 1;       // shift released
		return;
	default:
		/* Acknowledge the receival of the scancode by disabling and
		 * enabling the keyboard */
		outb(0, KEYB_ACKPORT);
		if(keybState & KEYB_WANTS_DATA_RESEND)
			outb(0, KEYB_ACKPORT);

		outb(1 << 7, KEYB_ACKPORT);
		if(keybState & KEYB_WANTS_DATA_RESEND)
			outb(1 << 7, KEYB_ACKPORT);
	}
	do_sysrq(inp, irq->regs);

	if(inp < 128) {
		scanBuf[scanPos] = inp;
		scanPos++;
		if(scanPos == KEYB_BUFFERSIZE) {
			if(readPos == 0) {
				DPRINT("Keyboard buffer overflow! Clearing.\n");
				keybClearBuf();
			}
			scanPos = 0;
		}

	}
}


/* Function: keyBufEmpty
 * Description: Returns a boolean value indicating whether the keyboard buffer
 * is empty. This can be indicated by the first keyboard event having a scancode
 * of zero.
 */
bool keybBufEmpty()
{
	return !!scanPos;
}

/*
 * Function: keybSetLeds
 * Description: Sets a specific led to either lit or not-lit.
 */
void keybSetLed(uchar LED, bool lit)
{
	if(lit)
		keybLedState |= LED;
	else
		keybLedState &= ~LED;
	keybSetLeds();
}

void keybSetLeds()
{
	outb(0xED, KEYB_IOPORT);
	if(keybState & KEYB_WANTS_DATA_RESEND)
		outb(0xED, KEYB_IOPORT);

	keybWait();
	outb(keybLedState, KEYB_IOPORT);
	if(keybState & KEYB_WANTS_DATA_RESEND)
		outb(keybLedState, KEYB_IOPORT);
}

/*
 * Function: keybGetLeds
 * Description: returns the raw data about which LED's are lit and which aren't.
 */
__uint8 keybGetLeds()
{
	return keybLedState;
}

/*
 * Function: int nextchar()
 * Description: Gets the next character from the buffer..
 */
int nextchar()
{
	while(1) {
		if(readPos == KEYB_BUFFERSIZE)
			readPos = 0;
		if(readPos != scanPos && scanBuf[readPos])
			break; // got it!
	}
	return kbmap[scanBuf[readPos++]][keybMapMod];
}

/*
 * Function: kgetc
 * Description: Simular to the libc 'getc' function, it retrieves a scancode
 * from the keyboard event buffer, translates it to an ascii value and then
 * returns that value.
 */
int kgetc()
{
	int ret = nextchar();
	printk(KINFO, "%c", ret);
	return ret;
}

/*
 * Function: kgets
 * Description: Simular to the libc 'gets' function, it takes a pointer to a
 * buffer of an arbitrary length, writing to it until it receives a carridge
 * return from the keyboard.
 */
char *kgets(char *buf)
{
	char *tmp = buf;
	while ((*tmp = kgetc()) != '\n') tmp++;
	return buf;
}

/*
 * Function: kkbhit
 * Description: Similar to the libc 'kbhit' function, it waits for the user to
 * hit a key and once it does that it returns TRUE, it should not return any
 * other value.
 */
bool kkbhit()
{
	nextchar();

	return TRUE;
}

/*
 * Function: keyboardInit
 * Description: This function initialises the keyboard hardware, telling it to
 * activate itself by sending a byte with bit 7 enabled to its ACKPORT. It then
 * requets a quick self-test and installs its interrupt handler.
 */
initFunc __preInitFunc keyboardInit()
{
	scanBuf = kmalloc(KEYB_BUFFERSIZE);
	keybClearBuf();
	/* Enable keyboard */
	outb(inb(KEYB_ACKPORT) | 1 << 7, KEYB_ACKPORT);
	if(keybState & KEYB_WANTS_DATA_RESEND)
		outb(inb(KEYB_ACKPORT) | 1 << 7, KEYB_ACKPORT);

	outb(0xAE, KEYB_INTERFACEPORT);
	if(keybState & KEYB_WANTS_DATA_RESEND)
		outb(0xAE, KEYB_INTERFACEPORT);

	/* Switch all LEDs off */
	keybLedState = 0;
	keybSetLeds();

	/* Install handler */
	addHandler(KEYB_IRQ, keybHandler, NO_DELAY, NULL);
	printk(KINFO, "keyboard initalised successfully\n");
	/* Do the Keyboard Selftest */
	printk(KINFO, "Doing Keyboard Selftest .. \n");
	keybSelftest();			// why the f00k does this fail?!!
	return SUCCESS;
}
