/*
 * partition.c - A partition locator and registra
 *
 * Added to the Zealos project and (C) by Silas Albrecht
 *
 * Current maintainer: Silas Albrecht <silas_a_a@yahoo.com>
 * Status: Maintained
 * Last update: 13/01/02
 *
 * Last CVS commit : $Id: partition.c,v 1.26 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * How a partition table works:
 * This document was made with the semi-useful doc at
 * http://nondot.org/sabre/os/S2Partitions/PartitionTables.txt, which doesn't
 * tell you the whole story. Here's what it misses out:
 * * Offsets of the partition table do *NOT* start at offset 0x1c0, but instead,
 *   because of the 2 signiture bytes, they start at offset 0x1be.
 * * Partition type 0x5 is an extended partition
 * -- Mark
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 10/04/01 silas: no extended partition support yet but that's easily fixed.
 *	also... note the hdd nums are (major:minor) 1:0 (hda) 1:64 (hdb) etc..
 *	partitionCmd() gets 1:1 - 1:63 for hda partitions, 1:65 - 1:123 for hdb,
 *	etc.. :)
 * 10/06/2001 mark: Misc fixups and enabling of extended partitions.
 * 05/11/01 Mark: Added subsystems code
 */

#include <common.h>
#include <partition.h>
#include <assert.h>
#include <device.h>
#include <kmalloc.h>
#include <block/ata.h>
#include <block/block.h>
#include <video.h>
#include <stdio.h>
#include <subsystems.h>

// The entry in the 'type' field for an extended partition
#define EXTENDED_TYPE 0x5

// Mark did this properly, silas fucked it up ;)
struct rawPartition {
	unsigned active		:1;
	unsigned crap		:7;
	__uint8 headStart;
	__uint8 sectStart;
	__uint8 cylStart;
	__uint8 type;
	__uint8 headEnd;
	__uint8 sectEnd;
	__uint8 cylEnd;
	__uint32 LBAStart;
	__uint32 LBALength;
} __PACKED__;

// The list we store our partition data in (private!)
struct partStore {
	bool active;
	__uint8 type;
	ulong start;
	ulong length;
	uchar num;
	struct partStore *next;
} *partitions[ATA_MAX_DRV * ATA_MAX_CNT] = {NULL, };

char *readMBR(struct devNum *dn, ulong offset)
{
	char *raw;

	assert(dn);
	raw = blk_read(dn->major, dn->minor, offset, 1);
	if(!raw) return NULL;

	// Check for magic signature:
	if(*(__uint16 *)(raw + 0x200 - sizeof(__uint16)) != 0xAA55) {
		kfree(raw);
		return NULL;
	}
	
	return raw;
}

// To recurse is devine...
struct partStore *readExtended(struct devNum *dn, ulong offset, uchar *c)
{
	char *raw = readMBR(dn, offset);
	struct rawPartition *p = (void *)(raw + 0x1be);
	struct partStore *t;
	if(!p || !p[0].type)
		return NULL;

	printk(KINFO, "%d ", *c);
	t = kmalloc(sizeof(struct partStore));
	t->type = p[0].type;
	t->active = p[0].active;
	t->start = p[0].LBAStart + offset;
	t->length = p[0].LBALength;
	t->num = (*c)++;
	DPRINT("active: %d; type: 0x%X; start (LBA): %d; len (LBA): %d; "
			"offset: %d\n", t->active, t->type, t->start, t->length,
			offset);
	t->next = p[1].LBAStart?readExtended(dn, p[1].LBAStart + offset,c):NULL;
	kfree(raw);
	return t;
}

struct partStore *readTable(struct devNum *dn)
{
	char *raw = readMBR(dn, 0);
	struct rawPartition *p;
	struct partStore *d = NULL;
	uchar i;

	if(!raw) return NULL;		// No entry found

	p = (void *)(raw + 0x1be);

	for(i = 0; i < 4; i++) {	// move it up 16 bytes each time
		struct partStore *t;
		if(!p[i].type)		// Blank entry
			continue;

		printk(KINFO, "%d ", i + 1);
		DPRINT("active: %d; type: 0x%X; start (LBA): %d; len (LBA):"
				" %d\n", p[i].active, p[i].type, p[i].LBAStart,
				p[i].LBALength);
		t = kmalloc(sizeof(struct partStore));
		t->active = p[i].active;
		t->type = p[i].type;
		t->start = p[i].LBAStart;
		t->length = p[i].LBALength;
		t->num = i + 1;
		t->next = d;
		d = t;
	}

	kfree(raw);

	return d;
}

/* this is called when somebody does something like cat /dev/hda1 */
deviceHnd(partCmd)
{
	struct blk_cmd *cmd;
	struct devNum dn;
	struct partStore *t;

	cmd = data;
	dn.major = num->major;
	dn.minor = (minor)(num->minor / 64) * 64;

	DPRINT("Called with num %d:%d and cmd %d", num->major, num->minor,
			cmd->type);

	for(t = partitions[num->minor / 64]; t; t = t->next)
		if(t->num == num->minor % 64) {
			/* Whoops, spent 3 or 4 days rewriting ATA to find out
			 * I'd forgotten to do the field adjustments here...
			 * - Mark
			 */
			cmd->start += t->start;
			return devCall(&dn, cmd);
		}

	return ENOEXIST;
}

// this function takes a partition type number and 
// returns a linked list of the partitions of that type
struct partList *partGetCommonTypes(__uint8 type, __uint8 disk)
{
	struct partStore *t;
	struct partList *start = NULL, *v;

	assert(disk < ATA_MAX_CNT * ATA_MAX_DRV);

	for(t = partitions[disk]; t; t = t->next) {
		if(t->type == type) {	// Found, link it in
			v = kmalloc(sizeof(struct partList));
			v->active = t->active;
			v->type = t->type;
			v->start = t->start;
			v->length = t->length;
			v->next = start;
			v->device.major = 1;
			v->device.minor = disk * 64 + t->num;
			start = v;
		}
	}

	return start;
}

initFunc __preInitFunc partitionInit()
{
	uchar i;

	REQUIRES_SUBSYSTEM("partition", "ATA");

	for(i = 0; i < ATA_MAX_CNT * ATA_MAX_DRV; i++) {
		struct partList *t;
		struct partStore *v;
		struct devNum d;
		uchar a = 5;

		d.major = 1;
		d.minor = i * 64;
		if(!devExists(&d))
			continue;

		printk(KINFO, "hd%d: ", i);
		partitions[i] = readTable(&d);

		printk(KINFO, "<");
		t = partGetCommonTypes(EXTENDED_TYPE, i);
		while(t) {
			struct partList *old;
			struct partStore *end, *s;
			for(end = NULL, s = partitions[i]; s; s = s->next)
				end = s;

			(end?end->next:partitions[i]) = readExtended(&d,
								t->start, &a);
			old = t;
			t = t->next;
			kfree(old);		// Remove the copy
		}
		printk(KINFO, "\b>\n");

		// Now add them all in
		for(v = partitions[i]; v; v = v->next) {
			d.minor = i * 64 + v->num;
			devReg(&d, partCmd);
		}
	}

	SUBSYSTEM_ON("partition", NULL, NULL, NULL);

	return SUCCESS;
}

EXPORT_SYMBOL(partGetCommonTypes);
