/*
 * vfs_cache.c - the vfs buffer cache system
 * 
 * Added to the Zealos project and (C) by silas albrecht
 * 
 * Current maintainer: silas albrecht <silas_a_a@yahoo.com>
 * Status: Maintained
 * Last update: 12/12/01
 * 
 * Last CVS commit : $Id: vfs_cache.c,v 1.9 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * Description:
 * this code manages a list of buffers currently in memory. when a device
 * reads from a hdd(or anything else that can use a blk_cmd i suppose)
 * cache_read reads it into memory and puts it on the list (*if* it isn't
 * already there). It's kept in memory until it's written.
 * 
 * Notes/Fixme's:
 * something is needed to occasionally write some cache to disk. i'm planning
 * on having a routine called on the timer int which would write some if
 * there isn't much activity going on.. the purpose of this is to not have too
 * much in the cache when we *need* memory because there's already a lot of
 * pressure on the system then..
 * 
 * Changelog:
 * 19/07/01 silas: restructured for per device caches
 */

#include <common.h>
#include <vfs.h>
#include <llist.h>
#include <block/block.h>
#include <printk.h>
#include <assert.h>
#include <kmalloc.h>
#include <signals.h>
#include <pit.h>
#include <irq.h>


#define WRITE_TIMEOUT NO_DELAY

struct vfs_cache {
	__uint8 dirty;
	__uint32 start;
	__uint32 length;
	char *data;
	struct vfs_cache *next;
};

// we maintain a list for each device
struct vfs_cache_dev {
	major major;
	minor minor;
	// activity vars in here..
	struct vfs_cache *cache;
	struct vfs_cache_dev *next;
} *dev_list = NULL;

void try_to_write()
{
	DPRINT("BOOM!");
}

inline struct vfs_cache_dev *cache_search_dev(major major, minor minor)
{
	struct vfs_cache_dev *dev;

	listItt(dev_list, dev)
		if(dev->major == major && dev->minor == minor)
			break;

	return dev;
}

// NOTE!! this doesn't need a length var because cache_read makes sure there
// is only one cache node at 'start'..
inline struct vfs_cache *cache_search_cache(struct vfs_cache *cache,
						__uint32 start)
{
	struct vfs_cache *node;

	listItt(cache, node)
		if(node->start == start)
			break;
	
	return node;
}

// bwahaha ;)
void cache_marks_dirty(major major, minor minor, __uint32 start,
				__uint32 length, char *data)
{
	struct vfs_cache_dev *my_dev = cache_search_dev(major, minor);
	struct vfs_cache *node = cache_search_cache(my_dev->cache, start);

	if(node)
		node->dirty = 1;
	else	// not in cache so write what they have..
		blk_write(major, minor, start, length, data);

	return;
}

/* function: errRet _cache_write(struct vfs_cache *cache, __uint32 start, __uint32 length, char *data)
 * description: see cache_write. this is just optomized so internal functions
 * don't have to search for the device's cache list..
 */
errRet _cache_write(struct vfs_cache_dev *my_dev, __uint32 start,
				__uint32 length, char *data)
{
	errRet ret = SUCCESS;
	struct vfs_cache *tmp;

	DPRINT("writing node\n");

	// ugh.. condition as an arg :(
	listDel(my_dev->cache, tmp, tmp->start == start);
	if(!tmp || tmp->dirty)
		ret = blk_write(my_dev->major, my_dev->minor, start, length,
				data);

	kfree(tmp);
	return ret;
}


/* function: char *cache_read(major major, minor minor, __uint32 start, __uint32 length)
 * description: searches for a buffer in the list, if part of it is in mem it
 * writes it and reads the whole thing. if none is in memory.. :)..
 */
char *cache_read(major major, minor minor, __uint32 start, __uint32 length)
{
	struct vfs_cache *node = NULL;
	struct vfs_cache_dev *my_dev = cache_search_dev(major, minor);

	if(my_dev == NULL) {
			DPRINT("making a new device cache and\n");
		if(dev_list == NULL) {
			// we are now making the cache.. so add the timer
			// int handler
	                addHandler(PIT_IRQ, try_to_write, WRITE_TIMEOUT, NULL);
		}
		my_dev = (struct vfs_cache_dev *)kmalloc(
						sizeof(struct vfs_cache_dev));
		my_dev->major = major;
		my_dev->minor = minor;
		my_dev->cache = NULL;
		listPrepend(dev_list, my_dev);
	} else {
		node = cache_search_cache(my_dev->cache, start);

		if(node) {
			if(node->length == length) // exact match
				return node->data;
			// part of what we want is in mem so write and reread
			_cache_write(my_dev, node->start, node->length,
					node->data);
			// FIXME: errmm.. i guess we could have some problems
			// if we kfree here but if we don't we could still have
			//  problems :( locks?
		}
	}

	DPRINT("making a new cache node\n");

	// make a new cache node
	node = (struct vfs_cache *) kmalloc(sizeof(struct vfs_cache));
	node->start = start;
	node->length = length;
	node->dirty = 0;
	node->data = blk_read(major, minor, start, length);
	listPrepend(my_dev->cache, node);
	
	return node->data;
}

/* function: inline errRet cache_write(major major, minor minor, __uint32 start, __uint32, length, char *data)
 * description: writes data to disk if its cache node is dirty or it doesn't
 * have a cache node..
 */
inline errRet cache_write(major major, minor minor, __uint32 start,
					__uint32 length, char *data)
{
	return _cache_write(cache_search_dev(major, minor), start, length,
			data);
}

// these should be optomized sometime, because cache_write will search for the
// node again..

void cache_write_dev(major major, minor minor)
{
	struct vfs_cache *node;
	struct vfs_cache_dev *my_dev = cache_search_dev(major, minor);

	listItt(my_dev->cache, node)
		cache_write(major, minor, node->start, node->length,
				node->data);
}

void cache_write_all()
{
	struct vfs_cache *node;
	struct vfs_cache_dev *cur_dev;
	listItt(dev_list, cur_dev)			// each device
		listItt(cur_dev->cache, node)		// each node of that
			_cache_write(cur_dev, node->start, node->length,
					node->data);
}

EXPORT_SYMBOL(cache_search_dev);
EXPORT_SYMBOL(cache_search_cache);
EXPORT_SYMBOL(cache_marks_dirty);
EXPORT_SYMBOL(cache_read);
EXPORT_SYMBOL(cache_write);
EXPORT_SYMBOL(_cache_write);
EXPORT_SYMBOL(cache_write_dev);
EXPORT_SYMBOL(cache_write_all);
