/*
 * vfs_inode.c - The VFS layer for inodes
 *
 * Added to the Zealos project and (C) by Silas Albrecht
 *
 * Current maintainer: Silas Albrecht <silas_a_a@yahoo.com>
 * Status: Maintained
 * Last update: 13/01/02
 *
 * Last CVS commit : $Id: vfs_inode.c,v 1.13 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 * we need some locks here :) (silas: agreed =) )
 *
 * Changelog:
 * 16/05/01 aris: fixed some bugs in list management
 * 16/05/01 silas: misc fixes
 */

#include <common.h>
#include <vfs.h>
#include <kmalloc.h>
#include <stdio.h>
#include <printk.h>
#include <llist.h>


/* function: struct inode *inode_lookup(struct inode *list, ulong i_num)
 * description: searches through 'list' for the inode of 'i_num'
 */
struct inode *inode_lookup(struct inode *list, ulong i_num)
{
	for ( ; list && list->i_number != i_num; list = list->next);
	return list;
}

/* function: void inode_write(struct inode *i)
 * description: calls 'i''s write_inode function if it's dirty and that
 * function exists
 */
void inode_write(struct inode *i)
{
	if (!i->dirty)
		return;

	if (!i->sb || !i->sb->super_ops || !i->sb->super_ops->write_inode) {
		i->dirty = 0;   // blah!
		return;
	}

	i->sb->super_ops->write_inode(i);
}

/* function: struct inode *inode_make_new(struct sblock *sb, ulong i_num)
 * description: kmallocs and initializes an inode with 'sb' in mind
 */
struct inode *inode_make_new(struct sblock *sb, ulong i_num)
{
	struct inode *inode;

	inode = (struct inode *)kmalloc (sizeof(struct inode));
	if (!inode) {
		DPRINT("ack! out of mem!!\n");
		return NULL;
	}

	memset(inode, 0, sizeof(struct inode));

	inode->i_number = i_num;
	inode->sb = sb;
	inode->usenum = 1;

	return inode;
}

/* function: inline void inode_read(struct inode * inode)
 * description: calls the inodes read_inode function if it exist
 */
inline void inode_read(struct inode * i)
{
        if (i->sb && i->sb->super_ops && i->sb->super_ops->read_inode)
                i->sb->super_ops->read_inode(i);
}

// used on the dirty list.. so we can assume it's dirty unlike inode_write
inline void inode_sync_one(struct inode *i)
{
        if (i->sb && i->sb->super_ops && i->sb->super_ops->write_inode)
                i->sb->super_ops->write_inode(i);
}

/* function: void inode_sync_list(struct inode *list)
 * description: syncs a list..
 */
void inode_sync_list(struct inode *list)
{
	ulong t;

	dlistItt(list, list, t)
		if (list->dirty)
			inode_sync_one(list);

	return;
}

/* function: struct inode *iget (struct sblock *sb, __uint16 ino)
 * description: iget is called to find an inode descriptor with a number
 * that matches 'i_num'. if one isn't found it alocates one by calling
 * inode_make_new.
 */
struct inode *iget (struct sblock *sb, __uint16 i_num)
{
	struct inode *i = inode_lookup(sb->inodes, i_num);
	if(i) {
		i->usenum++;
		return i;
	}
	
	i = inode_make_new(sb, i_num);
	if (i)
		dlistPrepend(sb->inodes, i);
	return i;
}

/* function: void iput (struct inode *i)
 * description: iput writes an inode ('i') to disk if usenum == 1.
 */
void iput (struct inode *i)
{
	if(!i) return;

	if(!i->usenum) {
		DPRINT("ugh! attempting to iput unused inode (inode=%lu)\n",
							i->i_number);
		return;
	}

	if(--i->usenum) return;

	if(i->sb && i->sb->super_ops && i->sb->super_ops->put_inode)
                i->sb->super_ops->put_inode(i);
	if(i->dirty)
		inode_write(i);

	memset(i, 0, sizeof(struct inode)); // save it for later, eh?
}

EXPORT_SYMBOL(inode_lookup);
EXPORT_SYMBOL(inode_make_new);
EXPORT_SYMBOL(inode_write);
EXPORT_SYMBOL(inode_read);
EXPORT_SYMBOL(inode_sync_one);
EXPORT_SYMBOL(inode_sync_list);
EXPORT_SYMBOL(iget);
EXPORT_SYMBOL(iput);
