/*
 * vfs_super.c - Superblock handling in the VFS layer
 *
 * Added to the Zealos project and (C) by Silas Albrecht
 *
 * Current maintainer: Silas Albrecht <silas_a_a@yahoo.com>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * Last CVS commit : $Id: vfs_super.c,v 1.33 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * VFS -- aka 'All Your FS' are belong to me!' - Mark
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 16/05/01 aris: what about don't use strings to identify fs types?
 * 16/05/01 silas: char *fsname -> int type and other misc changes
 * 22/05/01 mark: mopups etc
 * 23/05/01 silas: mopup mopups
 * 23/06/01 silas: changed all vfs code to use llist.h
 * 05/11/01 Mark: Added subsystems code
 */

#include <common.h>
#include <vfs.h>
#include <kmalloc.h>
#include <stdio.h>
#include <device.h>
#include <assert.h>
#include <partition.h>
#include <init.h>
#include <block/ata.h>
#include <llist.h>
#include <subsystems.h>


struct fs_type *file_systems = NULL;
struct sblock *super_blocks = NULL, *last_sblock = NULL;

struct fs_type *fs_type_find(__uint8 idnum)
{
	struct fs_type *tmp;
	ulong t;

	dlistItt(file_systems, tmp, t)
		if (tmp->idnum == idnum)
			break;

	return tmp;
}

errRet fs_type_reg(struct fs_type *fs)
{
	if (fs_type_find(fs->idnum))
		return EEXIST;

	dlistPrepend(file_systems, fs);
	return SUCCESS;
}

errRet fs_type_unreg (struct fs_type *fs)
{
	struct fs_type *tmp = fs_type_find(fs->idnum);

	if (!tmp) return ENOEXIST;

	dlistDel(file_systems, tmp);
	kfree(tmp);

	return SUCCESS;
}

// super_blocks management below!!
struct sblock *sb_create()
{
	struct sblock *new;

	new = kmalloc(sizeof(struct sblock));
	if (!new) return NULL;

	kmemset(new, 0, sizeof(struct sblock));

	return new;
}

struct sblock *sb_find(struct devNum *dev)
{
	struct sblock *tmp;
	ulong t;

	if (last_sblock && last_sblock->dev->major == dev->major &&
	    last_sblock->dev->minor == dev->minor)
		return last_sblock;

	dlistItt(super_blocks, tmp, t)
		if (tmp->dev->major == dev->major &&
		    tmp->dev->minor == dev->minor) {
			last_sblock = tmp;
			break;
		}

	return tmp;
}

errRet sb_reg (struct sblock *sb)
{
	if(!sb) // the fses read_super messed up somehow..
		return EOHHBUGG;

	if (sb_find(sb->dev))
		return EEXIST;

	dlistPrepend(super_blocks, sb);
	return SUCCESS;
}

errRet sb_unreg (struct devNum *dev)
{
	struct sblock *tmp;

	tmp = sb_find(dev);
	if (!tmp) return ENOEXIST;

	dlistDel(super_blocks, tmp);

	if (last_sblock == tmp)
		last_sblock = NULL;

	kfree(tmp);
	return SUCCESS;
}

/* function: errRet fsReg(struct fs_type *fs)
 * description: Register an fs given as a fs_type struct.
 */
errRet fsReg(struct fs_type *fs)
{
	ulong drv;
	// add it to the file_systems list
	fs_type_reg(fs);

	// for each drive
	for(drv = 0; drv < ATA_MAX_DRV * ATA_MAX_CNT; drv++) {
		struct partList *t, *p = partGetCommonTypes(fs->idnum, drv);
		// init all of this type
		for(; p; t = p, p = p->next, kfree(t)) {
			DPRINT("Type: 0x%x, device: %lu:%lu\n", fs->idnum,
					p->device.major, p->device.minor);
			sb_reg(fs->super_ops.read_super(&p->device));
		}
	}

	return SUCCESS;
}

/* function: initFunc __preInitFunc vfsInit()
 * type: init
 * description: Goes through each fs_type in 'fsStart' registers it.
 */
initFunc __preInitFunc vfsInit()
{
	ulong i;
	struct fs_type *syms = (void *)&fsStart;

	// Loop thru all the structs
	for(i = 0; i < (memory)&fsLen / sizeof(struct fs_type); i++) {
		printk(KINFO, "Detected fs reporting type: 0x%x\n",
				syms[i].idnum);
		fsReg(&syms[i]);
	}

	SUBSYSTEM_ON("VFS", NULL, NULL, NULL);

	return SUCCESS;
}

EXPORT_SYMBOL(fs_type_reg);
EXPORT_SYMBOL(fs_type_unreg);
EXPORT_SYMBOL(sb_create);
EXPORT_SYMBOL(sb_find);
EXPORT_SYMBOL(sb_reg);
EXPORT_SYMBOL(sb_unreg);
