/*
 * bitmap.c - A set of functions for making bitmap searches faster
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * Last CVS commit : $Id: bitmap.c,v 1.15 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 14/05/01 mark: Added the block-search thingie to hopefully speed up searches 
 *	(a *lot*)
 * 26/06/01 Mark: Converted it to take a bitmap * for most options. Kept
 *	bmpInit() for kmalloc's use only.
 */


#include <common.h>
#include <bitmap.h>
#include <stdio.h>
#include <assert.h>
#include <kmalloc.h>

#define blkSize (sizeof(ulong) * CHAR_BIT)

/* function: ulong bmpFind(bitmap *bitmap)
 * type: global
 * description: Finds a free entry in a bitmap and returns it's offset. If no
 * free bits are returned, returns 'length'.
 */
ulong bmpFind(bitmap *map)
{
	ulong i = 0;
	ulong *tmp = map->map;

	for(; i < map->length / blkSize; i++)
		if(tmp[i] != ~0UL)
			break;

	for(i *= blkSize; i < map->length; i++)
		if(!bmpIsUsed(map->map, i))
			break;

	return i;
}

/* function: ulong bmpFindSetFree(bitmap *bitmap)
 * type: global
 * description: Finds a free entry in a bitmap, marks it as used and returns
 * it's offset. If no free bits are returned, returns 'length'. See Also
 * bmpFree()
 */
ulong bmpFindSetFree(bitmap *map)
{
	ulong i = bmpFind(map);

	if(i != map->length)
		bmpMarkUsed(map->map, i);

	return i;
}

/* function: ulong bmpFindBlock(bitmap *map, ulong size)
 * type: global
 * desc: This finds 'size' free blocks in a row in a bitmap and returns the
 * offset of the beginning of the block.
 */
ulong bmpFindBlock(bitmap *map, ulong size)
{
	ulong i = bmpFind(map), j = size, k;
	ulong *t = map->map;

	DPRINT("Looking for %lu pages in bmp %p\n", size, map);

	assert(map);

	/* OK, I don't understand what this does, so I'm not expecting you to ;)
	 * - Mark (FIXME: anyone suggest something better?)
	 *
	 * Done tests and this is about twice as fast as the old one (about
	 * 30000 cycles to allocate 2MB memory :)
	 */
	for(i /= blkSize; i < map->length / blkSize; i++) {
		if(j < blkSize)
			goto loop2;

		if(t[i] == 0)
			j -= blkSize;
		else
loop1:			j = size;
	}

	i = 0;		// Fall thru

loop2:	for(k = i * blkSize; k < map->length; k++) {
		if(!j)
			return k - size;	// Size, not j

		if(bmpIsUsed(map->map, k)) {
			if(i)
				goto loop1;
			else
				j = size;
		} else
			j--;
	}

	return map->length;
}

/* function: ulong bmpFindBlockSetFree(bitmap *map, ulong size)
 * type: global
 * desc: This calls bmpFindBlock() and then marks the blocks as used. Returns
 * the offset of the beginning of the block.
 */
ulong bmpFindBlockSetFree(bitmap *map, ulong size)
{
	ulong i = bmpFindBlock(map, size);

	if(i != map->length)
		while(size)
			bmpMarkUsed(map->map, i + --size);

	return i;
}

/* function: errRet bmpResize(bitmap *map, ulong newLen)
 * desc: Resizes the bitmap 'map' to the size specified by 'newLen'.
 */
errRet bmpResize(bitmap *map, ulong newLen)
{
	void *new;
	spinlock(&map->resizeLock);

	new = kmalloc(newLen / CHAR_BIT + ((newLen % CHAR_BIT)?1:0));
	if(!new) {
		spinlockUnlock(&map->resizeLock);
		return ENOMEM;
	}

	if(newLen > map->length)
		bmpSetup(new + (map->length / blkSize), newLen -
				((ulong)(map->length / blkSize)) * blkSize);

	kmemcopy(new, map->map, MIN(map->length, newLen));

	kfree(map->map);
	map->map = new;
	map->length = newLen;
	spinlockUnlock(&map->resizeLock);

	return SUCCESS;
}

/* function: ulong bmpSetup(void *start, ulong length)
 * desc: Initalizes a bitmap table of length 'length' starting at 'start'.
 * Returns the number of bytes the table takes up.
 */
ulong bmpSetup(void *start, ulong length)
{
	ulong i;

	DPRINT("Initalizing bitmap at 0x%p, len %lu\n", start, length);

	i = kmemset(start, 0, length / CHAR_BIT) * CHAR_BIT;
	for(; i < length; i++)
		bmpMarkFree(start, i);		// Clear on the bit level

	return ROUND(i, CHAR_BIT);
}

/* function: ulong bmpInit(bitmap *map, void *start, ulong length)
 * type: global
 * description: Initalizes the specified bitmap. Returns as bmpSetup does.
 */
ulong bmpInit(bitmap *map, void *start, ulong length)
{
	map->length = length;
	map->map = start;
	spinlockInit(&map->resizeLock);

	return bmpSetup(start, length);
}

/* function: bitmap *bmpNew(ulong length)
 * desc: Creates and initalizes a new bitmap and returns the pointer to it.
 */
bitmap *bmpNew(ulong length)
{
	bitmap *new;

	new = kmalloc(sizeof(bitmap));

	bmpInit(new, kmalloc(length / CHAR_BIT + ((length % CHAR_BIT)?1:0)),
			length);

	return new;
}

/* function: void bmpMarkFree(void *start, ulong offset)
 * type: global
 * description: Marks a bit as free in a bitmap.
 */
inline void bmpMarkFree(void *start, ulong offset)
{
	btc(start, offset);
}

/* function: void bmpMarkUsed(void *start, ulong offset)
 * type: global
 * description: Marks a bit as used in a bitmap
 */
inline void bmpMarkUsed(void *start, ulong offset)
{
	bts(start, offset);
}

/* function: bool bmpIsUsed(void *start, ulong offset)
 * type: global
 * description: Returns true if bit is set to used, otherwise FALSE
 */
inline bool bmpIsUsed(void *start, ulong offset)
{
	return bt(start, offset);
}

EXPORT_SYMBOL(bmpFind);
EXPORT_SYMBOL(bmpFindSetFree);
EXPORT_SYMBOL(bmpFindBlock);
EXPORT_SYMBOL(bmpFindBlockSetFree);
EXPORT_SYMBOL(bmpResize);
EXPORT_SYMBOL(bmpSetup);
EXPORT_SYMBOL(bmpInit);
EXPORT_SYMBOL(bmpNew);
EXPORT_SYMBOL(bmpMarkFree);
EXPORT_SYMBOL(bmpMarkUsed);
EXPORT_SYMBOL(bmpIsUsed);
