/*
 * cpu.c - CPU-independant CPU functions (not an oxymoron)
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * Last CVS commit : $Id: cpu.c,v 1.25 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 * Add multiple CPU stuff in (See the bottom of the file).
 *
 * Changelog:
 *
 */

#include <common.h>
#include <cpu.h>
#include <delay.h>
#include <desc.h>
#include <irq.h>
#include <printk.h>
#include <stdio.h>
#include <pit.h>
#include <syscall.h>
#include <machine.h>
#include <kmalloc.h>
#include <debug.h>

machine thisBox;


// Dumps the registers
syscallFunc(dump)
{
	dumpRegs(regs);
	return SUCCESS;
}

/*
 * OK, the way we calibrate the delay loop is like this:
 * 1) Find a rough value
 * 2) go back down that xor'ing a mask to get the precise value
 * Simple eh?
 * This is quite like how linux does it except it's a bit shorter and a bit
 * quicker.
 *
 * Step 1:
 *	We set an initial delay (of 1) to caliberate with. This then
 *	gets shifted up each time, till the loop is slower than 1000 Hz. It then
 *	shifts it back one.
 * Step 2:
 *	A mask is applied (1 bit) and is rippled up till it's faster than 1000Hz
 *	as before, it's then shifted right by 1 and perminantly applied.
 */

#define PIT_CALIB_HZ 1000
// About 33 Mhz
#define INITAL_BITS 14
#define CALIB_BITS 8

volatile uchar tick = 0;

IRQ_DEF_HND(__preInitFunc delayISR)
{
	tick++;
	DPRINT(".");
}

errRet CPUDelaySetup(ulong num)
{
	ulong mask;
	ulong loopsPerMs = 1 << INITAL_BITS;
	__uint8 count;
	bool callAgain = TRUE;
	struct CPU *CPU = thisBox.CPU[num];

	printk(KINFO, "CPU[%lu]: Calibrating delay loop: ", num);

	// Set the PIT to 1000 interrupts a second.
	pitSet(PIT_CALIB_HZ, 0);
	// We add a handler to PIT's IRQ
	addHandler(PIT_IRQ, delayISR, NO_DELAY, NULL);

start:	do {
		loopsPerMs <<= 1;		// * 2
		tick = 0;
		while(!tick) /* Wait */;	// Wait for the next tick
		__delay(loopsPerMs);		// (tick==1)
	} while(tick == 1);

	loopsPerMs >>= 1;			// back 1

	/* This can only be called once */
	if(loopsPerMs == 1 << INITAL_BITS && callAgain) {  // proc is too slow
		loopsPerMs = 1;		// not 0, 0<<1 == 0... - Mark
		callAgain = 0;
		goto start;
	}

	/* We now do the fine caliberation */
	for(mask = loopsPerMs, count = CALIB_BITS; count; count--) {
		if(mask == 1)		// Prevent underflows -- Mark
			break;
		mask >>= 1;
		loopsPerMs |= mask;		// Apply the mask
		tick = 0;
		while(!tick) /* Wait */;
		__delay(loopsPerMs);		// (tick now = 1)
		if(tick != 1)			// If a tick has passed
			loopsPerMs ^= mask;	// remove it
	}

	delHandler(0, delayISR);
	pitSet(Hz, 0);				// Reset the Pic

	CPU->loopsPerMs = loopsPerMs;

	printk(KINFO, "%lu loops/ms\n", CPU->loopsPerMs);
	return SUCCESS;
}

/* function: delay(ulong milisec)
 * type: global
 * desc: Delays (does nothing) for the specified number of miliseconds. See also
 * udelay().
 */
void delay(ulong milisec)
{
	for(; milisec; milisec--)
		__delay(curCPU->loopsPerMs);
}

/* function: udelay(ulong milisec)
 * type: global
 * desc: Delays (does nothing) for the specified number of microseconds. See
 * also delay().
 */
void udelay(ulong microsec)
{
	__udelay(microsec);
}

/* function: initFunc CPUInit()
 * type: init
 * description: Generally detects the first CPU... Extend this sometime maybe?
 * :)
 */
initFunc __preInitFunc CPUInit()
{
	thisBox.CPU = kmalloc(sizeof(void *) * 1 /* nrCPUS */);
	CPUSetup(0);
	CPUDelaySetup(0);

#if 0
	DPRINT("Now delaying 10 seconds to varify that the value is correct\n");
	delay(10000);
	DPRINT("10 Secs is up. Let's get on wit it! :)\n");	// Test :)
	DPRINT("Now delaying for 1 second with udelay()\n");
	for(i = 0; i < 10; i++)
		udelay(100000);
	DPRINT("Done.\n");
#endif /* 0 */

	return SUCCESS;
}

EXPORT_SYMBOL(delay);
EXPORT_SYMBOL(udelay);
