/*
 * device.c - Management for device-communication
 *
 * Added to the Zealos project and (C) by Silas Albrecht
 *
 * Current maintainer: Silas Albrecht <silas_a_a@yahoo.com>
 * Status: Maintained
 * Last update: 13/01/02
 *
 * Last CVS commit : $Id: device.c,v 1.20 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 * This code will be gone soon when we have the new devices layer
 *
 * Changelog:
 * 04/04/01 mark: Heh, fixed up code for ya silas
 * 23/06/01 silas: thanks mark :)... changed to use llist.h code
 * 13/01/02 Mark: converted back to using direct calls rather than offsets
 */

#include <common.h>
#include <device.h>
#include <kmalloc.h>
#include <printk.h>
#include <assert.h>
#include <llist.h>


/* struct: majorDev, minorDev
 * type: local
 * description: see struct listSig, listTrig for a description of how these
 * functions work.
 */
struct majorDev {
	major majorNum;
	struct minorDev *minorList;
	LIST(struct majorDev);
} *devStart = NULL;

struct minorDev {
	minor minorNum;
	devHandler *function;
	LIST(struct minorDev);
};

/* function: struct majorDev *devFind(struct devNum *device)
 * type: local
 * description: This function finds a given major in a device list.
 */
struct majorDev *devFind(struct devNum *device)
{
	struct majorDev *t;

	assert(device);

	listItt(devStart, t)
		if(t->majorNum == device->major)
			return t;

	DPRINT("Couldn't find device with major %d\n", device->major);

	return NULL;
}

/* function: errRet devReg(struct devNum *device, offset func)
 * description: Registers 'device' with 'func' so that when a request is made
 * through that major/minor, it will pass it on to that device. See also,
 * devCall(), devDel() and devDellAll().
 */
errRet devReg(struct devNum *device, devHandler *func)
{
	struct majorDev *t;
	struct minorDev *tm;

	assert(device);
	assert(func);

	DPRINT("Registring device %d:%d with function 0x%p\n", device->major,
			device->minor, func);

	t = devFind(device);

	if(!t) {	// if NULL we have to create a majorDev node also
		struct majorDev *r = kmalloc(sizeof(struct majorDev));
		r->majorNum = device->major;
		r->minorList = NULL;

		listPrepend(devStart, r);	// Maybe we should lock this?
		t = r;
	}

	/* now that the major list has been dealt with we check if the
	 * minor already exists or not. if it hasn't then we add the
	 * device... otherwise... BWAHAHA *<|:-| ..
	 */
	listItt(t->minorList, tm)
		if(tm->minorNum == device->minor) {
			DPRINT("Minor exists\n");
			return EEXIST;		// Something here already.. hmm
		}

	tm = (struct minorDev *)kmalloc(sizeof(struct minorDev));
	tm->minorNum = device->minor;
	tm->function = func;

	listPrepend(t->minorList, tm);	// Lock this too ?

	return SUCCESS;
}

/* function: bool devExists(struct devNum *device)
 * description: Returns TRUE if the specified device exists, FALSE if it
 * doesn't. Code is based on the devCall() code, but with minor changes.
 */
bool devExists(struct devNum *device)
{
	struct majorDev *t;
	struct minorDev *tm = NULL;

	assert(device);		// No need to assert data, not our prob

	t = devFind(device);
	if(!t) return FALSE;

	DPRINT("Looking for %d:%d\n", device->major, device->minor);

	listItt(t->minorList, tm)
		if (tm->minorNum == device->minor)	// this is it! :)
			return TRUE;

	return FALSE;
}

/* function: errRet devDelAll(struct devNum *device, bool delAll)
 * description: If delAll is TRUE, deletes all in the major list and the
 * major, if it is FALSE, just deletes the major/minor. See also devDel().
 */
errRet devDelAll(struct devNum *device, bool delAll)
{
	struct majorDev *t = devStart, *old = NULL;
	struct minorDev *tm;
	
	assert(device);

	listItt(devStart, t) {
		if (t->majorNum == device->major)
			break;
		old = t;
	}

	if(!t) return ENOEXIST;			// No such entry

	tm = t->minorList;

	if(delAll) {			// delete the whole tree 
		struct minorDev *c;
		if(old) old->next = t->next;
		else devStart = t->next;

		DPRINT("Deleting all with major %d\n", device->major);

		while (tm) {
			/* Next first, then free (could be interrupted and
			 * kmalloc'd that space while we're still running,
			 * thus the ->next would be overwritten causing a big
			 * kernel fubar..
			 */
			c = tm;
			tm = tm->next;
			kfree(c);
		}
		kfree(t);
	} else {
	/* otherwise we just wanna delete a minor (unless it's the only one) */
		struct minorDev *n;

		DPRINT("Deleting %d:%d\n", device->major, device->minor);
		listDel(tm, n, n->minorNum == device->minor);

		if(!n) return ENOEXIST;

		if(!tm) {
			if(old) old->next = t->next;
			else devStart = t->next;
			kfree(t);
		}
		kfree(n);
	}

	return SUCCESS;
}

/* function: errRet devCall(struct devNum *device, void *data)
 * description: Makes a call to device 'device' with a data area at 'data'. Does
 * not check what 'data' is, so you can pass NULL or something like that
 * depending on the function that you're calling.
 */
errRet devCall(struct devNum *device, void *data)
{
	struct majorDev *t;
	struct minorDev *tm = NULL;

	assert(device);		// No need to assert data, not our prob

	t = devFind(device);
	if(!t) return ENOEXIST;

	DPRINT("Calling %lu:%lu\n", device->major, device->minor);

	listItt(t->minorList, tm)
		if (tm->minorNum == device->minor)	// this is it! :)
			return (tm->function)(device, data);

	DPRINT("Device %lu:%lu not found...\n", device->major, device->minor);
	return ENOEXIST;
}

#if CONFIG_DEVICE & OPT_IS_DEBUG
deviceHnd(testdev)
{
	DPRINT("Device: %d %d called with data %s\n", num->major, num->minor,
			data);
	return SUCCESS;
}
#endif /* CONFIG_DEVICE & OPT_IS_DEBUG */

/* function: initFunc deviceInit()
 * type: init
 * description: Initalizes the device handler for the kernel.
 */
initFunc __preInitFunc deviceInit()
{
#if CONFIG_DEVICE & OPT_IS_DEBUG
	struct devNum devnums;
	// Ideas taken from the signals code

	devnums.minor = 1;
	devnums.major = 1;

	DPRINT("Now testing devReg:\n");

	if(devReg(&devnums, GET_OFFSET(testdev)) == SUCCESS) {
		DPRINT("devReg seems to work, testing devCall\n");
		if(devCall(&devnums, "f00") == SUCCESS) {
			DPRINT("devCall seems successful, testing devDel\n");
			if(devDel(&devnums) == SUCCESS) {
				DPRINT("devDel successful, device code looks ok\n");
				return SUCCESS;
			}
		}
	}

	DPRINT("Failed\n");
	return EOHHBUGG;
#else
	return SUCCESS;
#endif
}

EXPORT_SYMBOL(devCall);
EXPORT_SYMBOL(devReg);
EXPORT_SYMBOL(devDelAll);
EXPORT_SYMBOL(devExists);
