/*
 * hash.c - A generic set of functions for making hash-tables
 * 
 * Added to the Zealos project and (C) by Mark Zealey
 * 
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 * 
 * Last CVS commit : $Id: hash.c,v 1.9 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * Description:
 * 
 * Notes/Fixme's:
 * hashOld (destructor) needs to free the linked lists from the hash-buckets.
 * 
 * Changelog:
 * 25/06/01 Mark: Moved this stuff over from functions.c and fixed up a bit.
 */

#include <common.h>
#include <hash.h>
#include <stdio.h>
#include <assert.h>
#include <kmalloc.h>
#include <printk.h>
#include <spinlock.h>

/* function: ulong hashGen(char *name)
 * type: local
 * description: Works out the position of the string in the hash table.
 */
ulong hashGen(struct hash *hash, char *name)
{
	ulong ret = 0;
	char *t;

	assert(name);
	assert(*name);

	for(t = name; *t; t++)
		ret = (ret * 64 + *t) % hash->entries;

	DPRINT("name: `%s' goes to hashed value %d\n", name, ret);
	return ret;
}

/* function: struct hashBucket *hashLocate(char *name, struct hashBucket *start)
 * type: local
 * description: Locates and returns the position of the matching entry in a
 * linked list from a hash. If none is found, returns NULL
 */
struct hashEntry *hashLocate(char *name, struct hashEntry *start)
{
	assert(name);
	assert(*name);

	listItt(start, start)
		if(kstrcmp(name, start->name)) {	// duplicate!
			DPRINT("Found `%s' at address 0x%p\n", name, start);
			return start;
		}

	return NULL;
}

/* function: errRet hashAdd(struct hash *hash, char *name, ulong data)
 * type: global
 * description: Adds 'data' into the hash table 'hash' and links it with 'name'.
 * If 'name' already exists, returns EEXIST.
 */
errRet hashAdd(struct hash *hash, char *str, ulong data)
{
	ulong i;
	struct hashEntry *t;

	assert(hash);
	assert(str);
	assert(*str);

	i = hashGen(hash, str);

	DPRINT("hash: 0x%p; name: `%s'; data: %d; entry: %d\n", hash, str,
			data, i);

	if(hash->entrys[i]) {  // Something's already here, check for dupes
		DPRINT("Entry into this place in the hash table already exists."
				" Checking for dupes\n");
		if(hashLocate(str, hash->entrys[i])) {	// Duplicate!
			DPRINT("Dupe found. Returning EEXIST.\n");
			return EEXIST;
		}
	}

	t = kmalloc(sizeof(struct hashEntry));
	t->data = data;
	t->name = kmalloc(length(str));
	kstrcopy(t->name, str);

	listPrepend(hash->entrys[i], t);

	return SUCCESS;
}

/* function: errRet hashFind(struct hash *hash, char *name, ulong *data)
 * type: global
 * desc: This function returns SUCCESS if the specified 'name' was found in the
 * hash, and then stores the data you stored with it in 'data'. If it's not
 * found, it will return ENOEXIST.
 */
errRet hashFind(struct hash *hash, char *str, ulong *data)
{
	struct hashEntry *t;
	ulong i;

	assert(hash);
	assert(str);
	assert(*str);
	assert(data);

	i = hashGen(hash, str);

	t = hashLocate(str, hash->entrys[i]);
	if(!t) {
		DPRINT("Can't find `%s' (%lu) in hash 0x%p\n", str, i, hash);
		return ENOEXIST;
	}

	DPRINT("located entry `%s' in hash 0x%p, position 0x%p, data %lu\n",
			str, hash, t, t->data);

	// XXX: We could now optomize the hash somehow

	*data = t->data;
	return SUCCESS;
}

/* function: bool hashExists(struct hash *hash, char *name)
 * type: global
 * desc: Returns TRUE if the specified entry exists in the hash. Otherwise it
 * returns FALSE.
 */
bool hashExists(struct hash *hash, char *str)
{
	ulong i = hashGen(hash, str);
	return hashLocate(str, hash->entrys[i])?TRUE:0;
}

/* function: errRet hashDel(char *name)
 * type: global
 * description: Removes the specified entry from the hash table 'hash'.
 */
errRet hashDel(struct hash *hash, char *str)
{
	struct hashEntry *v;
	ulong i;

	assert(hash);
	assert(str);
	assert(*str);

	i = hashGen(hash, str);

	listDel(hash->entrys[i], v, kstrcmp(str, v->name));
	if(v) {
		DPRINT("Found `%s' in hash 0x%p at address 0x%p links 0x%p\n",
				str, hash, v, v->next);
		kfree(v->name);		// And free the memory
		kfree(v);
		return SUCCESS;
	}

	DPRINT("Couldn't find `%s' (%lu) in hash 0x%p\n", str, i, hash);
	return ENOEXIST;
}

/* function: struct hash *hashNew(ulong entrys)
 * type: global
 * desc: Creates a new hash and returns a pointer to it.
 */
struct hash *hashNew(ulong entrys)
{
	struct hash *hash;
	hash = kmalloc(sizeof(struct hash));
	if(!hash)
		return NULL;

	hash->entries = entrys;

	hash->entrys = kmalloc(sizeof(struct func) * hash->entries);
	if(!hash->entrys) {
		kfree(hash);
		return NULL;
	}

	kmemset((void *)hash->entrys, 0, sizeof(struct hashEntry) * entrys);

	DPRINT("hash 0x%p has %lu entrys with an entry table at 0x%p", hash,
			hash->entries, hash->entrys);

	return hash;
}

/* function: errRet hashOld(struct hash *hash)
 * type: global
 * desc: Destroys the hash 'hash'.
 */
errRet hashOld(struct hash *hash)
{
	assert(hash);

	// FIXME: we should remove the data ref'd from the entrys... - Mark

	kfree(hash->entrys);
	kfree(hash);

	return SUCCESS;
}

EXPORT_SYMBOL(hashAdd);
EXPORT_SYMBOL(hashFind);
EXPORT_SYMBOL(hashExists);
EXPORT_SYMBOL(hashDel);
EXPORT_SYMBOL(hashNew);
EXPORT_SYMBOL(hashOld);
