/*
 * irq.c - Generic IRQ interface
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * Last CVS commit : $Id: irq.c,v 1.9 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 27/06/01 Mark: Moved this stuff over from i386/
 * 05/11/01 Mark: Added subsystems code
 */

#include <common.h>
#include <irq.h>
#include <stdio.h>
#include <kmalloc.h>
#include <signals.h>
#include <machine.h>
#include <subsystems.h>
#include <debug.h>


IRQData_t *IRQData;

/* variable: ulong IRQNum
 * description: The number of IRQ's the there are on this system
 */
ulong IRQNum;

/* variable: struct IRQHandler *IRQHandlers
 * desc: This is an array, one entry for every IRQ this system has.
 */
struct IRQHandler *IRQHandlers;

/* variable: __uint32 IRQSigRange
 * desc: This is a 'pointer' to the base signal number of the range of signals
 * reserved for fast IRQ handling.
 */
__uint32 IRQSigRange;

// This is the function that handles all IRQ's (setup in desc.c)
handleInt(triggerIRQ)
{
	uint num = r->num - IRQAddr(0);
	struct IRQ irq = {
		num: num,
		regs: r
	};

	EOIIntr(num);

	DPRINT("IRQ %lu called\n", num);

	sigTrigNum(num + IRQSigRange, &irq);		// TRIGGER! :)
}

errRet IRQAdd(ulong IRQ)
{
	if(!IRQHandlers[IRQ].things++)		// Do we need to enable?
		return IRQMask(IRQ, ENABLE);

	return SUCCESS;
}

/* function: errRet addHandler(ulong IRQ, sigHandler *address, ulong delay, void
 *						*infoAddr)
 * description: This function is called when you want to add a handler to an
 * IRQ. It makes sure that enough memory is avaliable, and if it isnt, it
 * kmalloc()'s more. Thus, there is no limit on how many handlers there are for
 * each function. (This is all handled by signals...)
 */
errRet addHandler(ulong IRQ, sigHandler *address, ulong delay, void *infoAddr)
{
	if(IRQ >= IRQNum)		// Out of range
		return ETOOBIG;

	// Interrupts are highest priority
	if(sigConNum(IRQ + IRQSigRange, address, 0, delay, infoAddr) != SUCCESS)
		return EOHHBUGG;

	return IRQAdd(IRQ);
}

/* This finction is called to remove the specified handler that was set in
 * addHandler().
 */
errRet delHandler(ulong IRQ, sigHandler *address)
{
	char buf[8];

	if(IRQ >= IRQNum)		// Out of range
		return ETOOBIG;

	/* This makes a string of the form 'IRQX' where X is the number...
	 * This is so we can add a handler to the signals register.
	 */
	sprintf(buf, "IRQ%d", IRQ);

	DPRINT("Removing signal name \"%s\"\n", buf);

	// If there's nothing connected, we can disable the IRQ
	if(!--IRQHandlers[IRQ].things)		// Do we need to disable?
		return IRQMask(IRQ, DISABLE);

	if(sigDisNum(IRQSigRange + IRQ, address) != SUCCESS)
		return EOHHBUGG;

	return SUCCESS;
}

initFunc __preInitFunc IRQInit()
{
	DPRINT("Initing IRQ's... ");

	IRQData = kmalloc(sizeof(IRQData_t));
	if (!IRQData)
		return ENOMEM;

	IRQNum = IRQInitArch();

	if(!IRQNum) {
		DPRINT("Returned 0 IRQ's.. something gone wrong?\n");
		return EOHHBUGG;
	}

	IRQSigRange = sigReqReg(IRQNum);
	IRQHandlers = kmalloc(sizeof(struct IRQHandler) * IRQNum);
	if(!IRQHandlers)
		return ENOMEM;

	kmemset(IRQHandlers, 0, sizeof(struct IRQHandler) * IRQNum);

	SUBSYSTEM_ON("IRQ", IRQData, IRQInit, NULL);

	DPRINT("Done\n");

	return SUCCESS;
}

EXPORT_SYMBOL(addHandler);
EXPORT_SYMBOL(delHandler);
EXPORT_SYMBOL(IRQAdd);

