/*
 * signals.c - Inter-task signaling...
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 13/01/02
 *
 * Last CVS commit : $Id: signals.c,v 1.22 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 * 30/06/01 Mark: This could do with using hashes (?)
 *
 * Changelog:
 * 25/10/00 mark: Re-wrote this file twice today after help from sauoq and
 *	erikm. Thanks to these people :)
 * 20/12/00 mark: Added a locking implimentation to it (Thanks to arjan)
 * 21/12/00 mark: Fixed the locking and implimented the trigger() function so
 *	you can trigger other signals, but not one thats currently
 *	being trigger()d :)
 * 24/06/01 silas: Converted to use llist code.
 * 25/06/01 silas: changed to use numbers instead of char *, made sigDelNumAll
 * 26/06/01 Mark: A few bug fixes, made the name->number layer. This has
 *	improved IRQ handling (IRQ's bypass the name->number layer) by about
 *	25% speed-wise.
 * 13/09/01 Mark: Removed some old lock.h stuff
 */

#include <common.h>
#include <signals.h>
#include <stdio.h>
#include <desc.h>
#include <kmalloc.h>
#include <assert.h>
#include <bitmap.h>
#include <hash.h>
#include <llist.h>
#include <atomic.h>
#include <signals.h>
#include <debug.h>


#if CONFIG_SIGNALS & OPT_IS_DEBUG
bool signalsDebug = 0;
#endif

////////////////////////////////////////////////////////////////////////////////
/* OK, This is the grunt layer of signals that does the useful stuff. If you
 * want to optomize, optomize here - Mark
 */

/* struct: listSig, listTrig
 * type: local
 * description: These are the main elements of the signals implimentation in
 * Zealos. The way this works is:
 * sig1 --> sig2 --> sig3 --> ...
 *   |        |        |
 * trig1    trig1    trig1
 *   |        |        |
 * trig2    trig2    trig2
 *   |        |        |
 * trig3    trig3    trig3
 *   |        |        |
 *  ...      ...      ...
 *
 * We handle this by having a linked list of names, listSig and then, from
 * each of those lists, we have a list of functions and their priorities in a
 * linked list referanced from the listSig. This is (probabily) the fastest
 * way to do searches.
 */
struct listTrig {
	sigHandler *function;	// Address of function to be called
	__uint32 delay;		// How many we ignore before called
	volatile __uint32 tdelay;  // The counter for delay.
	void *dataArea;		// Data to pass
	__uint8 priority;	// Priority of the function
	atomic read;		// Are we currently being read?
	LIST(struct listTrig);	// Link to the next one
};

static struct listSig {
	__uint32 signal;		// Signal Number
	struct listTrig *trig;		// Link to the chain coming off.
	LIST(struct listSig);		// Link to the next element
} *sBeg = NULL;

spinlock_t globalLock = INIT_UNLOCKED;

/* function: struct listSig *sigSearch(__uint32 signum)
 * type: local
 * Description: This function searches through listSig for the appropriate
 * signal and then returns a pointer to the listSig entry for it.
 */
struct listSig *sigSearch(__uint32 signum)
{
	struct listSig *sig;

	DPRINT("Looking for signal '%u': ", signum);

	spinlock(&globalLock);
	listItt(sBeg, sig)
		if(signum == sig->signal)
			break;
	spinlockUnlock(&globalLock);

	if(!sig) DPRINT("Not found.\n");
	else DPRINT("Found.\n");
	return sig;
}

/*
 * function: errRet sigConNum(__uint32 signum, sigHandler *function,
 *			__uint8 priority, __uint32 delay, void *dataArea)
 * type: global
 * arg0: signal number.
 * arg1: Pointer to the function that will be called.
 * arg2: The priority that it gets.
 * arg3: The 'delay' (IE the number of times that the signal has to be
 * trigger()'d before it gets called. The value gets reset to this every time it
 * is actually called. (see connect() if you don't want to use a delay))
 * arg4: An area of data to pass to 'function' every time the function is
 * called.
 *
 * description: Every time 'signum' is sigTrig()'d, 'function' will be
 * called. 'function' should be the address of a function created with
 * signalHnd(). There can be many functions connected to the same signal, so,
 * the ones with the lowest 'priority' get called first (Unless you really
 * want your function to be called first, please use the DEF_PRIO macro to
 * specify the default priority level). If 2 or more functions are
 * sigConnectd()'d on the same priority level, they're called on a 'last come,
 * first served' basis. the 'dataArea' is passed to 'function' every time it
 * is called, as the argument of 'void *dataBlk'. For more info, please see
 * the signalHnd() function or sigCon().
 */
errRet sigConNum(__uint32 signum, sigHandler *function, __uint8 priority,
		 __uint32 delay, void *dataArea)
{
	struct listTrig *t, *next;
	struct listSig *sig = sigSearch(signum);

	DPRINT("Trying to connect function 0x%p with signal %u at priority %d "
			"and delay %d:\n", function, signum, priority, delay);

	if(!sig) {			// If there is no signal created yet
		DPRINT("Creating a new signal\n");

		sig = kmalloc(sizeof(struct listSig));	// create a new one
		sig->signal = signum;			// Set the values
		sig->trig = NULL;
		next = sig->trig;			// Forces the next code

		spinlock(&globalLock);
		listPrepend(sBeg, sig);
	} else {
		// Else, zoom down to the place where we will insert the signal
		spinlock(&globalLock);
		listItt(sig->trig, next)
			if(priority <= next->priority)
				break;
	}

	// Create the new bit in the chain
	t = kmalloc(sizeof(struct listTrig));

	if(sig->trig == next) listPrepend(sig->trig, t);
	else listPrepend(next, t);

	t->priority = priority;		// Set the priority
	t->function = function;		// Set the address
	t->delay = delay;		// Delay
	t->tdelay = delay;		// Start off wating
	t->dataArea = dataArea;
	atomic(t->read) = 0;
	spinlockUnlock(&globalLock);

	DPRINT("Success!\n");

	return SUCCESS;
}

/* function: errRet sigDisNum(__uint32 signum, sigHandler *function)
 * type: global
 * desc: Disconnects the specified function from the sigTrigNum() list of the
 * specified number. If function is NULL, it removes all functions connected to
 * the number. See Also: sigDel().
 */
errRet sigDisNum(__uint32 signum, sigHandler *function)
{
	struct listTrig *trig;
	struct listSig *sig = sigSearch(signum);

	DPRINT("Trying to remove function 0x%p from signal %u: ", function,
			signum);

	if(!sig) return ENOSIG;

	if(function) {
		spinlock(&globalLock);
		listDel(sig->trig, trig,
			trig->function == function && !atomic(trig->read));
		spinlockUnlock(&globalLock);

		if(!trig) return ENOSIG;	// Not found
		kfree(trig);			// Free the memory
	}

	if(!sig->trig || !function) {	// We need to remove it from the list
		struct listSig *s;
		spinlock(&globalLock);
		listDel(sBeg, s, s == sig);
		spinlockUnlock(&globalLock);
		trig = sig->trig;
		kfree(s);
	}

	if(!function) {			// Remove the whole list.
		struct listTrig *n;
		// We can't listItt here
		while(trig) {
			n = trig;
			trig = trig->next;
			kfree(n);
		}
	}

	DPRINT("Success\n");

	return SUCCESS;
}

/* function: errRet sigTrigNum(__uint32 signum)
 * type: global
 * desc: Triggers all of the functions connected to the specified signal number.
 * See also sigTrig() for triggering based on a string.
 */
errRet sigTrigNum(__uint32 signum, void *data)
{
	struct listTrig *trig;
	struct listSig *sig = sigSearch(signum);

	DPRINT("Signal %lu called. sBeg is 0x%p\n", signum, sBeg);

	if(!sig) return ENOSIG;		// Cant find the signal in the list

	listItt(sig->trig, trig)
		if(!--trig->tdelay) {	// dec it
			trig->tdelay = trig->delay;	// Reset

			DPRINT("calling function: %p\n",trig->function);

			atomicInc(&trig->read);
			// Call it!
			(trig->function)(trig->dataArea, data);
			atomicDec(&trig->read);
		} else
			DPRINT("Not calling function: %p\n",
						trig->function);

	DPRINT("Signal %lu ended\n", signum, sBeg);

	return SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
/* OK, this is the linking layer. IRQ's should bypass this for speed. */
#define INIT_BMP_SZE 1023
#define BMP_STEP_SZE 1024
#define HASH_SZE 1023

bitmap *sigUsed;
struct hash *sigHash;

/* function: __uint32 sigReqReg(__uint32 length)
 * desc: Reserves a range of signal numbers and returns the start of it.
 * 'length' is the number you want. A return of 0 means failure.
 */
__uint32 sigReqReg(__uint32 length)
{
	ulong num;
	do {
		num = bmpFindBlockSetFree(sigUsed, length);
		if(num == sigUsed->length) {	// No free places found
			errRet ret;
			// Grow bitmap
			ret = bmpResize(sigUsed, sigUsed->length +BMP_STEP_SZE);
			if(ret != SUCCESS)
				return 0;
		}
	} while(num == sigUsed->length);

	return num;
}

/* function: errRet sigCon(char *name, sigHandler *func, __uint8 prio,
 *		__uint32 delay, void *data)
 * type: global
 * desc: Converts 'str' to a number and calls sigConNum().
 */
errRet sigCon(char *name, sigHandler *func, __uint8 prio, __uint32 delay,
		void *data)
{
	errRet ret;
	ulong val;

	if(hashFind(sigHash, name, &val) != SUCCESS) {
		/* Doesn't exist, create a new entry */

		DPRINT("Signal '%s' not found, making new one: ", name);

		val = bmpFindSetFree(sigUsed);
		if(val == sigUsed->length) {
			ret = bmpResize(sigUsed, sigUsed->length +BMP_STEP_SZE);
			if(ret != SUCCESS)
				return ret;

			val = bmpFindSetFree(sigUsed);
			if(val == sigUsed->length) { // Wierd.. bitmap problem?
				DPRINT("Umm, fuck. Think bitmaps barfed.\n");
				return EOHHBUGG;
			}
		}

		DPRINT("given entry %lu\n", val);

		if((ret = hashAdd(sigHash, name, val)) != SUCCESS) {
			bmpMarkFree(sigUsed->map, val);
			return ret;
		}
	}

	return sigConNum(val, func, prio, delay, data);
}

/* function: errRet sigDis(char *str, sigHandler *func)
 * type: global
 * desc: Converts 'str' to a number and calls sigDisNum().
 */
errRet sigDis(char *str, sigHandler *func)
{
	ulong num;
	errRet ret;

	if(hashFind(sigHash, str, &num) != SUCCESS) {
		DPRINT("Signal '%s' not found\n", str);
		return ENOEXIST;
	}

	if((ret = sigDisNum(num, func)) != SUCCESS)
		return ret;

	/* If there arn't any left, destroy the hash table entry and free the
	 * bit in the list. FIXME: Is there a better way of doing this?
	 *  - Mark
	 */
	if(!sigSearch(num)) {
		DPRINT("Removing %lu from has name `%s'\n", num, str);
		ret = hashDel(sigHash, str);
		bmpMarkFree(sigUsed->map, num);
		return ret;
	}

	return SUCCESS;
}

/* function: errRet sigTrig(char *str)
 * type: global
 * desc: This calls off sigTrigNum with the number linked to the name 'str'.
 */
errRet sigTrig(char *str, void *data)
{
	ulong num;

	if(hashFind(sigHash, str, &num) != SUCCESS) {
		DPRINT("Signal '%s' not found\n", str);
		return ENOEXIST;
	}

	return sigTrigNum(num, data);
}

/* function: initFunc signalsInit()
 * type: init
 * description: Initalizes signals for the kernel.
 */
initFunc __preInitFunc signalsInit()
{
	sigUsed = bmpNew(INIT_BMP_SZE);
	sigHash = hashNew(HASH_SZE);
	bmpMarkUsed(sigUsed->map, 0);		// Entry 0 is special (reserved)

	return SUCCESS;
}

EXPORT_SYMBOL(sigCon);
EXPORT_SYMBOL(sigDis);
EXPORT_SYMBOL(sigTrig);
EXPORT_SYMBOL(sigReqReg);

EXPORT_SYMBOL(sigConNum);
EXPORT_SYMBOL(sigDisNum);
EXPORT_SYMBOL(sigTrigNum);
