/*
 * stdio.c - Standard routines used in the kernel
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 11/03/02
 *
 * Last CVS commit : $Id: stdio.c,v 1.29 2002/03/06 16:12:01 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 13/05/01 mark: Totally redid the vsprintf() and numberToString() functions.
 *	Should be a lot cleaner now.
 * 07/11/01 Mark: Improved a lot of stuff in here.
 * 18/11/01 Zwane: Redid the *upper/lower functions and split some up.
 */


#include <common.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <video.h>
#include <kmalloc.h>
#include <assert.h>
#include <spinlock.h>

spinlock_t printLock = INIT_UNLOCKED;

#define SIGN 1		// is it signed
#define UPPER 2		// do it upper-case

#define LONG 1		// Long
#define LONG_LONG 2	// Long long

/* function: void putsColor(char *string, __uint8 color)
 * description: Prints a string to the screen using the specified colours. See
 * also FG() and BG()
 */
void putsColor(char *string, __uint8 color)
{
	ulong i, t;

	spinlockIRQSave(&printLock, &t);

	for(i = 0; string[i]; i++)
		putcColor(string[i], color);

	spinlockIRQRestore(&printLock, &t);
}

/* function: ulong atoi(char *str, __uint8 base)
 * description: Converts from string 'str' to a number of base 'base'.
 */
ulong atoi(char *str, __uint8 base)
{
	ulong ret = 0;

	while(*str) {
		ret *= base;
		if(*str >= '0' && *str <= '9')
			ret += *str - '0';
		else if(*str >= 'a' && *str <= 'z')
			ret += *str - 'a' + 10;
		else if(*str >= 'A' && *str <= 'Z')
			ret += *str - 'A' + 10;

		str++;
	}

	return ret;
}

/* function: void numberToString(ulong num,char *out,uint *outPos,__uint8
 *					base,char flags, uint precision)
 * description: Converts an integer (in binary) to a string (in any base)
 */
void numberToString(char **args, __uint8 fill, char *out, uint *outPos,
		__uint8 base, char flags, uint precision)
{
	uint len = 0;
	char *chr = "0123456789abcdef";
	ulong long num = 0;

	if(fill & LONG_LONG)
		num = vaArg(*args, ulong long);
	else if(fill & LONG)
		num = vaArg(*args, ulong);
	else
		num = vaArg(*args, uint);

	if(flags & UPPER)
		chr = "0123456789ABCDEF";

	if(!precision) {
		ulong long t = num;
		__uint32 n;

		do {
			len++;
			n = base;
		} while((t = mathDiv64(t, &n)));
	} else
		len = precision;

	if((flags & SIGN) && ((long long)num < 0)) {
		*outPos++ = '-';		// print a '-'
		num = -(long long)num;		// invert it
	}

	*outPos += len;
	out += *outPos;
	for(; len; len--) {		// Do it reversed...
		__uint32 rem = base;
		num = mathDiv64(num, &rem);
		*--out = chr[rem];
	}
}

/* function: void vsprintf(char *out, char *fmt, char *args)
 * description: Does the formating of the string for functions like printf() and
 * sprintf()
 */
void vsprintf(char *out, char *fmt, char *args)
{
	uint outPos = 0;
	char *chr;

	assert(out);
	assert(fmt);

	for(; *fmt; fmt++)
		if(*fmt == '%') {
			__uint8 base = 0, fill = 0, flags = 0;
			uint precision = 0;

			fmt++;
			/* If someone specifies a precision like '%2d', get
			 * the '2', otherwise, we set it to 0, or autodetect
			 * (assuming base 10)
			 */
			while(*fmt >= '0' && *fmt <= '9')
				precision = (precision * 10) + *fmt++ - '0';

			if(*fmt == 'l')	{	// Filling
				fill |= LONG;
				fmt++;
			} else if(*fmt == 'L') {
				fill |= LONG_LONG;
				fmt++;
			}

			/* Convert uppercase stuff to lowercase and set the
			 * uppercase output flag
			 */
			if(*fmt >= 'A' && *fmt <= 'Z') {
				flags |= UPPER;
				*fmt += 'a' - 'A';
			}

			switch(*fmt) {
			case 'p':
				fill = LONG;
			case 'x':
				base = 16;
				break;

			case 'd':
			case 'i':
				flags |= SIGN;
			case 'u':
				base = 10;
				break;

			case 'o':
				base = 8;
				break;

			case 'b':		// Binary
				base = 2;
				break;

			case 's':
				chr = vaArg(args, char *);
				while(*chr) out[outPos++] = *chr++;
				break;

			case 'c':
				out[outPos++] = vaArg(args, char);
				break;
			}

			if(base)
				numberToString(&args, fill, out, &outPos, base,
						flags, precision);
		} else
			out[outPos++] = *fmt;

	out[outPos] = 0;
}

/* function: void printf(char *fmt, ...)
 * description: Formats a string and prints it to the screen using the
 * universially recognised % characters. uses vsprintf() to format it.
 */
void printf(char *f, ...)
{
	char *args;
	char out[1000];

	vaStart(args, f);
	vsprintf(out, f, args);
	puts(out);
}

/* function: void sprintf(char *buf, char *fmt, ...)
 * description: Formats the string with 'fmt' and rather than putting to the
 * screen like printf(), it puts it into 'buf'.
 */
void sprintf(char *buf, char *f, ...)
{
	char *args;

	vaStart(args,f);
	vsprintf(buf,f,args);
}

/* function: bool kstrcmp(char *s1, char *s2)
 * description: Compares 2 strings. Returns TRUE is they are equal, otherwise
 * returns FALSE
 */
bool __WEAK__ kstrcmp(char *s1, char *s2)
{
	while(*s1 == *s2 && *s1 && *s2) { s1++; s2++; }
	if(*s1 || *s2) return FALSE;
	return TRUE;
}

/* function: bool kbeginswith(char *s1, char *s2)
 * description: Compares 2 strings. Returns TRUE if the second string begins
 * with the 1st, otherwise returns FALSE
 */
bool __WEAK__ kbeginswith(char *s1, char *s2)
{
	while(*s1 == *s2++ && *s1) s1++;
	if(*s1) return FALSE;
	return TRUE;
}

/* function: void kstrcopy(char *p1, char *p2)
 * description: copies 'p2' to 'p1' until it reaches a \0 in 'p2', for example
 * kstrcopy(foo,"bar"); would copy "bar" to char *foo
 */
void __WEAK__ kstrcopy(char *p1, char *p2)
{
	assert(p1);
	assert(p2);
	while(*p2) {
		*p1 = *p2;
		p1++; p2++;
	}
	*p1 = 0;
}

/* function: void kmemcopy(char *p1, char *p2, ulong len)
 * description: Copys a block of memory from p2 to p1 (len bytes)
 */
void __WEAK__ kmemcopy(char *a, char *b, ulong l)
{
	while(l--)
		*a++ = *b++;
}

/* function: ulong length(char *string)
 * description: Returns the length of the specified string (including the \0)
 */
ulong __WEAK__ length(char *s)
{
	ulong len = 0;

	assert(s);

	while(s[len])
		len++;

	return len + 1;
}

/* function: char toLower(char c)
 * description: Returns 'c' in lower-case
 */
char __WEAK__ toLower(char c)
{
	if ((c >= 'A') && (c <= 'Z'))
		c = 'a' + (c - 'A');

	return c;
}

/* function: char * strToLower(char *s)
 * description: Converts all chars in the string to lower case.
 */
char * __WEAK__ strToLower(char *s)
{
	char *p = s;
	while (*p) {
		*p = toLower(*p);
		p++;
	}

	return s;
}

/* function: char toUpper(char c)
 * description: Returns 'c' in upper-case
 */
char __WEAK__ toUpper(char c)
{
	if ((c >= 'a') && (c <= 'z'))
		c = 'A' + (c - 'a');

	return c;
}

/* function: char * strToUpper(char *s)
 * description: Converts all chars in the string to upper case.
 */
char * __WEAK__ strToUpper(char *s)
{
	char *p = s;
	while (*p) {
		*p = toUpper(*p);
		p++;
	}

	return s;
}

/* function: void bts(char *str, ulong bit)
 * description: Sets a bit in the string 'str'. See also bt() and btc().
 */
void __WEAK__ bts(char *str, ulong bit)
{
	assert(str);

	str[bit / CHAR_BIT] |= 1 << (bit % CHAR_BIT);
}

/* function: void btc(char *str, ulong bit)
 * description: Clears a bit in the specified string. See also bts() and bt().
 */
void __WEAK__ btc(char *str, ulong bit)
{
	assert(str);

	str[bit / CHAR_BIT] &= ~(1 << (bit % CHAR_BIT));
}

/* function: bool bt(char *str, ulong bit)
 * description: Tests a bit in the specified string and returns TRUE if it is
 * set. See also bts() and btc().
 */
bool __WEAK__ bt(char *str, ulong bit)
{
	assert(str);

	return str[bit / CHAR_BIT] & (1 << (bit % CHAR_BIT));
}

/* function: ulong chompl(char *str, ulong len)
 * description: removes all non-printable characters from the end of 'str'. (and
 * also spaces). It replaces them with end-of-string (\0) and returns the
 * length of the new string.
 */
ulong __WEAK__ chompl(char *str, ulong len)
{
	while(str[--len] <= ' ') {
		str[len] = '\0';
		if(!len)		// Prevent buffer-underflows - Mark
			break;
	}
	return len;
}

/* function: ulong chomp(char *str)
 * description: removes all non-printable characters from the end of 'str'. (and
 * also spaces). It replaces them with end-of-string (\0) and returns the
 * length of the new string.
 */
ulong __WEAK__ chomp(char *str)
{
	return chompl(str, length(str));
}

/* function: void kmemdump(char *base, ulong length)
 * description: dumps 'length' bytes of memory as hex, starting at base address
 * 'base'.
 */
void __WEAK__ kmemdump(char *base, ulong length)
{
	__uint8 k;
	__uint16 *i = (void *) base;
	memory j = 0;

	length /= 2 * CHAR_BIT;

	for(; j < length; j++) {
		printk(KINFO, "%8x: ", j * 16 + (memory)base);
		for(k = 0; k < 8; k++)
			printk(KINFO, "%4x ", i[j * 2 * CHAR_BIT + k]);
		printk(KINFO, "\n");
	}
}

EXPORT_SYMBOL(putsColor);
EXPORT_SYMBOL(atoi);
EXPORT_SYMBOL(numberToString);
EXPORT_SYMBOL(sprintf);
EXPORT_SYMBOL(vsprintf);
EXPORT_SYMBOL(kstrcmp);
EXPORT_SYMBOL(kbeginswith);
EXPORT_SYMBOL(kstrcopy);
EXPORT_SYMBOL(kmemcopy);
EXPORT_SYMBOL(length);
EXPORT_SYMBOL(toLower);
EXPORT_SYMBOL(toUpper);
EXPORT_SYMBOL(strToUpper);
EXPORT_SYMBOL(strToLower);
EXPORT_SYMBOL(bts);
EXPORT_SYMBOL(btc);
EXPORT_SYMBOL(bt);
EXPORT_SYMBOL(chomp);
EXPORT_SYMBOL(chompl);
EXPORT_SYMBOL(kmemset);
EXPORT_SYMBOL(kmemdump);
