/*
 * subsystems.c - Sub-system tracking fuctions
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * This file handles all aspects of sub-system registration.
 *
 * Notes/Fixme's:
 * Fix subsysDel() when force = TRUE
 *
 * Changelog:
 * 03/11/01 Mark: Created
 * 21/11/01 Zwane: Added init/exit subsystem members
 * 12/12/01 Mark: Fixed zwane's mess ;>
 */


#include <common.h>
#include <subsystems.h>
#include <hash.h>
#include <assert.h>
#include <kmalloc.h>
#include <stdio.h>
#include <init.h>
#include <debug.h>

#define HASH_SIZE 1024

/* Main hash table for looking these sub systems up */
struct hash *subsysTable = NULL;

/* Structure of a single dependany in the chain */
struct subsysDep {
	LIST(struct subsysDep);
	char name[0];
};

/* Structure of a hash-table 'value' */
struct subsysEnt {
	void *data;
	uint num_deps;
	struct subsysDep *deps;
	errRet (*init)();		/* the subsystem's startup routine */
	errRet (*exit)();		/* the subsystem's cleanup routine */
};

/* Helper function */
struct subsysEnt *findsubsys(char *name)
{
	void *t;

	assert(name);

	/* Check it existed there */
	if(hashFind(subsysTable, name, (ulong *)&t) != SUCCESS)
		return NULL;

	assert(t);

	return t;
}

errRet subsysAdd(char *name, void *data, errRet (*init)(), errRet (*exit)())
{
	struct subsysEnt *ent = NULL;

	assert(name);

	ent = kmalloc(sizeof(struct subsysEnt));
	if(!ent) {
		DPRINT("Out of memory!\n");
		return ENOMEM;
	}

	ent->data = data;
	ent->num_deps = 0;
	ent->deps = NULL;
	/* register the subsystems startup/cleanup functions */
	ent->init = init;
	ent->exit = exit;

	if(hashAdd(subsysTable, name, (ulong)ent) != SUCCESS) {
		kfree(ent);
		return EOHHBUGG;
	}

	return SUCCESS;
}

errRet subsysDepAdd(char *thisSubsys, char *name)
{
	struct subsysEnt *ent;
	struct subsysDep *dep;

	assert(thisSubsys);
	assert(name);

	ent = findsubsys(name);

	if(!ent)
		return ENOEXIST;

	dep = kmalloc(sizeof(struct subsysDep) + length(thisSubsys));
	if(!dep) {
		kfree(ent);
		return ENOMEM;
	}

	kstrcopy(dep->name, thisSubsys);
	if(ent->num_deps == 0) {
		assert(ent->deps == NULL);
		listInit(dep);
		ent->deps = dep;
	} else
		listAdd(ent->deps, dep);

	ent->num_deps++;

	return SUCCESS;
}

errRet subsysDepDel(char *thisSubsys, char *name)
{
	struct subsysEnt *ent;
	struct subsysDep *dep;

	assert(thisSubsys);
	assert(name);

	ent = findsubsys(name);
	if(!ent || !ent->deps)
		return ENOEXIST;

	listDel(ent->deps, dep, kstrcmp(dep->name, thisSubsys));

	if(!dep)
		return ENOEXIST;

	kfree(dep);

	ent->num_deps--;
	/* XXX: This isn't needed, I think */
	if(ent->num_deps == 0)
		ent->deps = NULL;

	return SUCCESS;
}

errRet subsysDel(char *name, bool force)
{
	struct subsysEnt *ent;
	struct subsysDep *t, *o = NULL;

	assert(name);

	ent = findsubsys(name);
	if(!ent)
		return ENOEXIST;

	/* First, weed out deps that don't actually exist */
	if(ent->deps)
		listItt(ent->deps, t) {
			if(!subsysIsLoaded(t->name)) {
				/* Sub-system is not loaded, remove dep */
				listRm(ent->deps, t, o);
				ent->num_deps--;
			}

			o = t;
		}

	if(force) {
		/* XXX: Force removal of dependancies?? */
		/* FIXME: Memory leak */
		ent->deps = NULL;
		ent->num_deps = 0;
	}

	/* Check to see if there's any deps.. */
	if(ent->deps) {
		assert(ent->num_deps == 0);
		return EEXIST;
	}

	if(hashDel(subsysTable, name) != SUCCESS)
		return EOHHBUGG;

	kfree(ent);

	return SUCCESS;
}

bool subsysIsLoaded(char *name)
{
	return hashExists(subsysTable, name);
}

errRet subsysGetData(char *name, void **data)
{
	struct subsysEnt *ent;

	assert(name);
	assert(*data);

	ent = findsubsys(name);
	if(!ent)
		return ENOEXIST;

	*data = ent->data;

	return SUCCESS;
}

initFunc subsysInit()
{
	struct subsystem *subs;
	ulong i;

	subsysTable = hashNew(HASH_SIZE);

	/* Add in statically registered subsystems */
	subs = (void *)&subStart;

	for(i = 0; i < (memory)&subLen / sizeof(struct subsystem); i++) {
		DPRINT("Adding '%s' with ptr %p, init %p, exit %p\n",
				subs[i].name, subs[i].ptr, subs[i].init,
				subs[i].exit);
		subsysAdd(subs[i].name, subs[i].ptr, subs[i].init,
				subs[i].exit);
	}

	return SUCCESS;
}

EXPORT_SYMBOL(subsysAdd);
EXPORT_SYMBOL(subsysDepAdd);
EXPORT_SYMBOL(subsysDepDel);
EXPORT_SYMBOL(subsysDel);
EXPORT_SYMBOL(subsysIsLoaded);
EXPORT_SYMBOL(subsysGetData);
