/*
 * time.c - Generic time functions
 *
 * Added to the Zealos project and (C) by Reggy
 * Parts (C) by Mark Zealey <mark@itsolve.co.uk>
 *
 * Current maintainer: Reggy Ekkebus <reggy@ekky.org>
 * Status: Maintained
 * Last update: 13/01/02
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * Generic time interface for zealos.
 *
 * jiffies is a 64-bit number, which avoids most problems when it rolls over, I
 * calculated that it will take aproximatly 5845580504 (about 6 billion) years
 * to roll over if Hz is 100. If anyone gets zealos uptime that long, I'll buy
 * them a beer ;)) - Mark
 *
 * Notes/Fixme's:
 * Timezones need fixing? ATM the seconds since the epoch are measured in the
 * current timezone's time... - Mark
 *
 * Changelog:
 * 22/09/01 Mark: Fixed mktime().
 * 25/09/01 Mark: Fixed some of reggy's stuff, fixed RTCGetofday().
 * 28/09/01 Mark: Fixed it, now reads time correctly etc..
 * 28/09/01 Mark: Renamed and moved most stuff over to generic/kern/time.c.
 * 28/09/01 Mark: Created from i386/kern/time.c
 * 29/09/01 Mark: Added a RTCGetofday() for those boxen which have a cycle
 *	counter.
 * 31/10/01 Mark: Added jiffies protection (via jiffies_lock)
 * 05/11/01 Mark: Added subsystems code
 */

#include <common.h>
#include <time.h>
#include <pit.h>
#include <cpu.h>
#include <irq.h>
#include <math64.h>
#include <printk.h>
#include <subsystems.h>

/* Include specifics from our time driver */
#ifdef CONFIG_MC146818
#include <char/mc146818.h>
#endif /* CONFIG_MC146818 */

/* All time drivers should export these functions */
void RTCWaitUpdate();
ulong RTCGet(__uint8 type);

struct timeval last_time;
/* We must hold this lock to write or read jiffies. (on a 32-bit box...) */
spinlock_t jiffies_lock = INIT_UNLOCKED;
__uint64 jiffies64 = 0;

#ifdef HAVE_CYCLE_COUNT
/* The last cycle_count() that we did in gettimeofday() gets recorded here */
__uint64 last_cycle_count = 0;
#endif /* HAVE_CYCLE_COUNT */

/* Function: errRet gettimeofday(struct timeval *ts)
 * Description: Puts time struct containing 'the time' into the pointer
 * specified :)
 */
errRet gettimeofday(struct timeval *ts)
{
	/* Part of the reason that this code is so ugly is that egcs doesn't
	 * handle 64-bit aithmatic too well, and is buggy. This should work with
	 * egcs... - Mark */
#ifdef HAVE_CYCLE_COUNT
	register __uint64 count = cycle_count();
	__uint32 rem = curCPU->loopsPerMs;
	__uint32 t = rem / 1000;
	__uint32 ms = mathDiv64(count - last_cycle_count, &rem);

	last_time.tv_sec += ms / 1000;
	ms %= 1000;
	last_time.tv_usec += ms * 1000 + rem / (t ?: 1);

	last_cycle_count = count;
#else /* HAVE_CYCLE_COUNT */
	static __uint64 last_jiffies = 0;
	__uint64 cur_jiffies = get_jiffies64();

	/* spinlockIRQSave() needed? I suppose it doesn't matter if a tick
	 * happens while we are in here... - Mark*/

	/* The following code is euck and non-portable, but it has to be this
	 * way so we can work with 64-bit jiffies... - Mark
	 */
	if(cur_jiffies - last_jiffies) {
		/* Divide the difference by 100 (or whatever Hz is), result in
		 * 'res' and remainder in 'rem'. */
		ulong rem = Hz, res = mathDiv64(cur_jiffies - last_jiffies,
				&rem);

		last_time.tv_sec += res;
		/* Adjust for varying values of Hz */
		last_time.tv_usec += rem * 1000000 / Hz;

		last_jiffies = cur_jiffies;
	}
#endif /* HAVE_CYCLE_COUNT */

	while(last_time.tv_usec >= 1000000) {
		last_time.tv_usec -= 1000000;
		last_time.tv_sec++;
	}

	*ts = last_time;		// Copy the struct

	return SUCCESS;
}

/* Function: ulong mktime(uint sec, uint min, uint hour, uint day, uint month,
 *				uint year)
 * Description: 'normal' time -> unix time Year is 4 digits.
 */
ulong mktime(uint sec, uint min, uint hour, uint day, uint month, uint year)
{
	/* This algorithm was originally invented by Gauss - clever guy */

	/* Rearrange months so that february goes last because it has the
	 * leap-day. */
	if((int)(month -= 2) <= 0) {
		month += 12;
		year--;
	}

	return (ulong)(((
					// Do leap-years
					year / 4
					- year / 100
					+ year / 400
					// Years
					+ year * 365
					// Months
					+ 367 * month / 12
					// Days
					+ day
					/* Take the magic number (all the days
					 * from 0 AD to 1970 AD)
					 */
					- 719499
				) * 24 + hour	// Hours
			) * 60 + min		// Minutes
		) * 60 + sec;			// Seconds
}

IRQ_DEF_HND(jiffies_handler)
{
	/* No need for a lock as we are locking against ourselves in an irq
	 * handler */
	/* Actually on SMP you can have someone accessing jiffies64 whilst
	 * you're in the jiffieshandler */
	jiffies64++;
}

initFunc __preInitFunc timeInit()
{
	uint sec, min, hour, day, month, year;

	/* Wait for the RTC to update the number of seconds, so we know that
	 * we'll be on exactly 0 usec. */
	RTCWaitUpdate();
	sec = RTCGet(TIME_SECONDS);
	min = RTCGet(TIME_MINUTES);
	hour = RTCGet(TIME_HOURS);
	day = RTCGet(TIME_DAYS);
	month = RTCGet(TIME_MONTHS);
	year = RTCGet(TIME_YEARS);

	last_time.tv_sec = mktime(sec, min, hour, day, month, year);
	last_time.tv_usec = 0;

#ifdef HAVE_CYCLE_COUNT
	last_cycle_count = cycle_count();
#endif /* HAVE_CYCLE_COUNT */

	printk(KINFO, "Current time: %2u/%2u/%4u %2u:%2u:%2u (%lu seconds "
			"since the epoch)\n", day, month, year, hour, min, sec,
			last_time.tv_sec);

	addHandler(PIT_IRQ, jiffies_handler, NO_DELAY, NULL);

	SUBSYSTEM_ON("time", NULL, NULL, NULL);

	return SUCCESS;
}

EXPORT_SYMBOL(mktime);
EXPORT_SYMBOL(jiffies64);
EXPORT_SYMBOL(gettimeofday);
