/*
 * mm/page.c - Give a page or range of pages
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * Last CVS commit : $Id: page.c,v 1.23 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * All the functions in this file do is give a range of pages or free a range of
 * pages. They all use page *'s, it is up to the paging layer to convert that
 * into an address.
 *
 * This file uses the buddy system (Anyone want to impliment lazy buddy?) to
 * allocate and merge pages. It is zone'd, like linux's and each page maintains
 * lists of free pages.
 *
 * Notes/Fixme's:
 * 04/08/01 Mark: Anyone want to convert this from buddy to lazy buddy?
 *
 * Changelog:
 * 04/08/01 Mark: Rewrote to use the buddy system and nicer interface etc from
 *	the previous method which used a bitmap.
 */


#include <common.h>
#include <kmalloc.h>
#include <init.h>
#include <stdio.h>
#include <printk.h>
#include <cpu.h>
#include <assert.h>
#include <mem.h>
#include <page.h>
#include <machine.h>
#include <debug.h>

memArea memZone[MAX_AREAS];	// List of memory areas
bool swapSetup = FALSE;		// Is swap setup yet?
page *pageBase;			// Pointer to the base of the page tables

page *tryToGet(memArea *area, uint order)
{
	uint newOrder;
	page *p;

	/* Look through for a level where there are free pages */
	for(newOrder = order; newOrder < MAX_ORDER; newOrder++)
		if(area->free[newOrder])
			break;

	/* Damn, none found. Cut our losses and quit now */
	if(newOrder >= MAX_ORDER)
		return NULL;

	/* Remove the first entry from the list */
	p = area->free[newOrder];

	DPRINT("p = %p:%lu ", p, p?p->page:0);

	dlistDel(area->free[newOrder], p);

	/* Add the other half to the next list down, then add the next half 1
	 * level down etc... */
	while(newOrder-- > order)
		dlistPrepend(area->free[newOrder], p + (1 << newOrder));

	return p;
}

/* function: page *pageAlloc(uint order, uint options)
 * desc: Returns 2 ^ 'order' pages that are free and marks them as used.
 */
page *pageAlloc(uint order, uint options)
{
	page *p;
	uint i;

	if(order >= MAX_ORDER) {
		DPRINT("order (%ld) >= MAX_ORDER\n", order);
		return NULL;
	}

	for(i = 0; i < MAX_AREAS; i++) {
		// Don't mess with areas we shouldn't mess in
		if(i == AREA_DMA && !(options & OPT_DMA))
			continue;
		if(i == AREA_NORMAL && !(options & OPT_NORMAL))
			continue;

		/* Try to shift it off first, if that fails we go on to the next
		 * zone.
		 */
		p = tryToGet(memZone + i, order);
		if(p) {
			assert(p->refs == 0);
			p->refs++;
			DPRINT("Given page %lu (zone %ld, order %ld)\n",
					p->page, i, order);

#ifdef CONFIG_MM_STATS
			stats.pagesUsed += 1 << order;
			stats.pagesCurrent += 1 << order;
#endif /* CONFIG_MM_STATS */
			return p;
		}
	}

	DPRINT("No zones with pages free were found");

	return NULL;
}

/* function: errRet __pageFree(page *page, uint order, bool merge)
 * desc: This function should never be called directly. You should use either
 * pageFree() or pageFreeNoBuddy(). This function frees the specified pages and
 * if 'buddy' is TRUE, the function will attempt to merge the pages up.
 */
errRet __pageFree(page *p, uint order, bool merge)
{
	uchar zone;

	if(!p) {
		DPRINT("Given NULL PAGE!!\n");
		return EINVALID;
	}

#ifdef MM_STATS
	/* Must be here before order gets ordered by the buddy stuff */
	stats.pagesFree += 1 << order;
	stats.pagesCurrent -= 1 << order;
#endif /* MM_STATS */

	DPRINT("page: %lu, order: %u\n", p->page, order);

	zone = pageZone(p->page);

	if(merge)
		while(order < MAX_ORDER - 1) {
			ulong buddy;

			/* Work out who our buddy is */
			buddy = p->page ^ (1 << order);

			DPRINT("page %lu's buddy is %lu. order %u\n", p->page,
					buddy, order);

			/* Check to see if our buddy is being used, if it is, we
			 * can't go any further
			 */
			if(pageBase[buddy].refs > 0)
				break;

			/* Remove our buddy from it's free list */
			dlistDel(memZone[zone].free[order], pageBase + buddy);

			/* Clear this bit of page (to take it up another level)
			 */
			p = (p->page > buddy)?pageBase + buddy:p;

			/* And go onto the next order */
			order++;
		}

	DPRINT("%u/%u: %p %p (%lu) merge: %d\n", zone, order,
			memZone[zone].free[order],p, p->page, merge);

	p->refs = 0;

	/* Add it in to the new list */
	dlistAdd(memZone[zone].free[order], p);

	return SUCCESS;
}

/* function: errRet pageFree(page *page, uint order)
 * desc: Frees the specified page at the specified order.
 */
errRet pageFree(page *p, uint order)
{
	/* XXX: Maybe lazy buddy could be implimented here? - Mark */
	return __pageFree(p, order, TRUE);
}

/* function: errRet pageFreeNoBuddy(page *page, uint order)
 * desc: Frees the specified page at the specified order. Does not attempt to
 * merge the pages.
 */
errRet pageFreeNoBuddy(page *p, uint order)
{
	return __pageFree(p, order, FALSE);
}

void *kmallocPages(uint num, uint options)
{
	return (void *)(pageAlloc(pageGetOrder(num),options)->page * PAGE_SIZE);
}

errRet kfreePages(void *p, uint num)
{
	return pageFree(pageBase + ((memory)p / PAGE_SIZE), pageGetOrder(num));
}

uint pageGetOrder(ulong size)
{
	uint ret = 0;

	size--;
	while(size) {
		ret++;
		size >>= 1;
	}

	return ret;
}

// Initalizes pages etc
errRet __preInitFunc pageInit()
{
	ulong i, j;
	memory end = kernelStart + kernelLength;
	ulong memPages, pagesLeft;

	/* Clear out the memory areas */
	for(i = 0; i < MAX_AREAS; i++) {
		for(j = 0; j < MAX_ORDER; j++)
			memZone[i].free[j] = NULL;
		memZone[i].pagesFree = 0;
		memZone[i].pagesTotal = 0;
	}

	memPages = thisBox.memCnt / (PAGE_SIZE / 1024);
	pagesLeft = 0;
	pageBase = (void *) end;

	/* Try and get a straight range that we can put our list of pages in */
	for(i = end; i < memPages * PAGE_SIZE; i += sizeof(page)) {
		if(isMemory((memory)(pageBase + pagesLeft))) {
			/* Yep, this is normal memory */
			if(++pagesLeft == memPages)
				break;
		} else {
			/* Damn, reset the count... */
			pagesLeft = 0;
			pageBase = (void *) i;
		}
	}

	/* If we run out of pages, maybe we should limit memory to what we have?
	 * but then I suppose there wouldn't be any left over for other
	 * things... tough. Maybe you should just get some more memory (or a
	 * better bios)?
	 */
	if(i == memPages * PAGE_SIZE) {
		DPRINT("Umm, ran out of pages for allocating pages.. sorry.\n");
		while(1);
	}

	end = (memory)(pageBase + pagesLeft + 1);
	DPRINT("Beginning of pages: 0x%p, end: 0x%lx\n", pageBase, end);

	/* Itterate through all the pages */
	for(i = 0; i < memPages; i++) {
		pageBase[i].page = i;
		pageBase[i].options = isMemory(i * PAGE_SIZE)	? PAGE_UNUSED
								: PAGE_USED;
		pageBase[i].refs = 0;
	}

	/* Free individual pages till we can do block frees */
	for(i = ROUND(end, PAGE_SIZE); (i & (~0UL << (MAX_ORDER - 1))) != i;
			i++)
		pageFreeNoBuddy(pageBase + i, 0);

	/* Quick free */
	for(; i < memPages; i += (1 << (MAX_ORDER-1)))
		pageFreeNoBuddy(pageBase + i, MAX_ORDER - 1);

#ifdef MM_STATS
	stats.pagesUsed = 0;
	stats.pagesFree = 0;
	stats.pagesCurrent = 0;
#endif /* MM_STATS */

	/* Other initalization stuff? */

	kernelLength = end - kernelStart;

	return SUCCESS;
}

EXPORT_SYMBOL(kfreePages);
EXPORT_SYMBOL(kmallocPages);

