/*
 * slab.c - Split pages up into slabs and give to the user
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * Last CVS commit : $Id: slab.c,v 1.26 2002/02/19 16:46:32 jalh Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * There are a set of lists with sizes of the form 2^n of all the memory
 *	'leases' that are currently being let out. These lists are struct
 *	buddyList. For each 'lease' a new entry is put into the chain, of the
 *	form struct buddySlab. It contains the start memory address of it, and
 *	a link to the next member of the list. These lists are kept in special
 *	pages, which are on the pageBeg list. Each page has a header with a
 *	bitmap of the pages free in it and also a link to the next page in the
 *	list. These pages are kept separate from other kmalloc'able pages.
 *
 * Notes/Fixme's:
 * mark: Make kfree try to join slabs so it doesn't fragment memory so badly
 *
 * Changelog:
 * 07/10/01	Zwane: Added check for NULL pointer in kfree 
 */


#include <common.h>
#include <printk.h>
#include <stdio.h>
#include <assert.h>
#include <bitmap.h>
#include <kmalloc.h>
#include <llist.h>
#include <mem.h>
#include <vmalloc.h>
#include <debug.h>

#undef MAX_ORDER
#define MAX_ORDER PAGE_SIZE
#define BMP_END (PAGE_SIZE / sizeof(struct buddySlab))
#define BUDDY_TABLE(x) (((memory)(x)) & ~(PAGE_SIZE - 1))
#define BUDDY_POS(x) ((((memory)(x)) & (PAGE_SIZE - 1)) /	\
		sizeof(struct buddySlab))

// The structure for a single entry into a buddy list
struct buddySlab {
	void *start;		// Where we start
	LIST(struct buddySlab);	// next free but in the chain
};

// The header for a page that we alloc'd for the buddy linked list tables
struct buddyHeader {
	char bmp[BMP_END / CHAR_BIT];
	LIST(struct buddyHeader);
} __PACKED__ *pageBeg = NULL;

// List of the lists that we use (different block sizes)
struct buddyList {
	__uint32 size;			// Size of the blocks in this list
	struct buddySlab *start;	// Pointer to the start of the list
} slabs[] = {
	{  16, NULL},		// Avoid wasting space for linked lists etc
	{  32, NULL},
	{  64, NULL},
	{ 128, NULL},
	{ 256, NULL},
	{ 512, NULL},
	{1024, NULL},
	{2048, NULL},
#if PAGE_SIZE > 4096
	{4096, NULL},		// What's the point of using the buddy stuff on
				// PAGE_SIZE and above when we just kmallocPages
				// them
#endif
	{   0, NULL}		// Null ending
};

typedef struct {
	memory size;
} slabHeader;

#ifdef CONFIG_MM_STATS
struct stats stats;
#endif

// Looks thru a buddyHeader per-page and returns the address of a un-used
// buddySlab, or NULL if none found
inline struct buddySlab *findBuddyFreeSlotPage(struct buddyHeader *hdr)
{
	bitmap b = {
		map: hdr->bmp,
		length: BMP_END
	};
	ulong i = bmpFindSetFree(&b);
	if(i == BMP_END)		// No free space
		return NULL;

	DPRINT("Found buddy slab %ld free and base address 0x%p\n", i, hdr);

	return (struct buddySlab *)((memory)hdr + i * sizeof(struct buddySlab));
}

// Finds and returns a the address of a free entry into the buddy list table if
// there's no space in any of the tables, creates a new new one
struct buddySlab *findBuddyFreeSlot()
{
	struct buddyHeader *t;
	struct buddySlab *i;
	ulong j;

	listItt(pageBeg, t) {
		i = findBuddyFreeSlotPage(t);
		if((memory)i)
			return i;
	}

	// OK, looks like there was no space in any of the tables, so we just
	// make a new one and link it in >:)

	DPRINT("Creating new buddy list table: ");

	t = kmallocPage();
	assert(t);		// Out of memory??
#ifdef CONFIG_MM_STATS
	stats.buddyTables++;
#endif

	DPRINT("t is: 0x%p\n", t);

	bmpSetup(t->bmp, sizeof(struct buddyHeader));

	// Mark the first few as used so we don't write over the header
	for(j = 0; j <= (sizeof(struct buddyHeader) / sizeof(struct buddySlab));
			j++)
		bmpMarkUsed(t->bmp, j);

	listPrepend(pageBeg, t);	// Link it with the next one
	return findBuddyFreeSlotPage(pageBeg);	// To recurse is devine
}

// Adds and makes space for an entry in the specified slab table
errRet addSlab(__uint16 order, void *address)
{
	struct buddySlab *t = findBuddyFreeSlot();	// Get address
	assert(slabs[order].size);

	DPRINT("Adding slab of order %u and base mem address of 0x%p, "
				"in table %p:%lu\n", (uint)order, address,
				BUDDY_TABLE(t), BUDDY_POS(t));

	t->start = address;
	listPrepend(slabs[order].start, t);	// And add it in :)
	return SUCCESS;
}

// Removes a slab and will free its space in the table, but not delink it from
// the chain!!
void delSlab(struct buddySlab *slab)
{
	struct buddyHeader *t;
	memory offset = BUDDY_POS(slab);
	memory base = BUDDY_TABLE(slab);

	listItt(pageBeg, t)
		if((memory)t == base) {
			DPRINT("Found slab to delete in table %p:%lu, "
						"reporting start of 0x%p\n",
					t, offset, slab->start);

			if(bmpIsUsed(t->bmp, offset))
				bmpMarkFree(t->bmp, offset);
			else
				DPRINT("Attempted to free a slab that was "
					"already free'd: table %p:%lu\n", t,
					offset);
			return;
		}

	// Cant find it, print an error and forget about it
	DPRINT("Cant locate position to free... (%p)\n", slab);
}

// This will split the slabs until we get a linked list of them at the order
// which we want
void splitSlabs(__uint16 order)
{
	struct buddySlab *slab;
	ulong i;

	// If we've allready got a free slab at this level, we just return
	// straight away
	if(slabs[order].start)
		return;

	if(!slabs[order+1].size) {	// Allocate a page
		void *t;

		DPRINT("Allocating new page for slab-ifying\n");

		t = kmallocPage();

		if(t)
			for(i = 0; i < PAGE_SIZE; i += slabs[order].size)
				addSlab(order, t + i);
		else
			DPRINT("Out of memory!!!\n");

		return;
	}

	splitSlabs(order + 1);	// Look for one at the next level

	DPRINT("Splitting slab pages down a bit (order == %d)\n",order);

	// Delink from chain
	slab = slabs[order+1].start;
	slabs[order+1].start = slab->next;

	// Add in the entrys for it
	for(i = 0; i < slabs[order+1].size; i += slabs[order].size)
		addSlab(order, slab->start + i);

	// And remove it
	delSlab(slab);
}

// Works out what list we have to use for this chunk of memory
__uint8 kgetOrder(memory size)
{
	__uint8 order = 0;
	while(size > slabs[order].size && slabs[order].size)
		order++;
	return order;
}

memory kgetSize(void *ptr)
{
	slabHeader *head;
	assert(ptr);
	head = (void *)(ptr - sizeof(slabHeader));
	return head->size;
}

void setSize(void *ptr, memory len)
{
	slabHeader *head;
	assert(ptr);
	head = (void *)(ptr - sizeof(slabHeader));
	head->size = len;
}

void *kmalloc(memory size)
{
	__uint8 order;
	void *t;

	size += sizeof(slabHeader);		// extra bytes for the info

	// Work out which list we have to use
	order = kgetOrder(size);

	DPRINT("Someone requested %lu bytes (== order %u)\n", size,
			(uint)order);

	if(slabs[order].size) {
		struct buddySlab *v;

		splitSlabs(order);	//Make sure we have a slab at this level

		// Remove slab and delink from the list
		v = slabs[order].start;
		assert(v);
		slabs[order].start = slabs[order].start->next;
		t = v->start;

		delSlab(v);
	} else {
		/* We've got a request larger than we could handle with just a
		 * single page, so we request the number of pages, and remember
		 * to treat this as a special case later for kfree
		 */
		t = vmalloc(ROUND(size, PAGE_SIZE), OPT_DMA | OPT_NORMAL);
		DPRINT("Got range of pages %lu, len %lu\n", (memory)t
				/ PAGE_SIZE, ROUND(size, PAGE_SIZE));
	}

	t += sizeof(slabHeader);
	setSize(t, size);

	DPRINT("Gave memory address 0x%p, size %ld bytes (order %u)\n", t, size,
			order);

#ifdef CONFIG_MM_POISON
	kmemset(t, KMALLOC_POISON, size - sizeof(slabHeader));
#endif

#ifdef CONFIG_MM_STATS
	stats.kmallocs++;
	stats.kmallocCurrent++;
#endif
	return t;
}

void kfree(void *start)
{
	memory size;
	__uint8 order;

	if (!start)
		return;

	size = kgetSize(start);
	start -= sizeof(slabHeader);
	order = kgetOrder(size);		// Find the order out

	DPRINT("Trying to free memory address 0x%p, which reports a "
				"size of %ld bytes and order of %u\n", start,
				size, order);

#ifdef CONFIG_MM_POISON
	kmemset(start, KFREE_POISON, size);
#endif

#ifdef CONFIG_MM_STATS
	stats.kfrees++;
	assert(stats.kmallocCurrent);  // kmalloc's < kfree's (shouldn't happen)
	stats.kmallocCurrent--;
#endif
	if(!slabs[order].size) {
		DPRINT("Free'ing via kfreePages with an order of %u and a size "
				"of %lu\n", order, size);

		vfree(start, ROUND(size, PAGE_SIZE));
	} else
		addSlab(order, start);		// Add it as free space
}

void *krealloc(void *ptr, memory newSize)
{
	void *ret = ptr;

	if(!ptr)
		return kmalloc(newSize);

	if(!newSize) {
		kfree(ptr);
		return NULL;		// Success
	}

	// See if we have enough free space to just return (free space in
	// current place)
	if(slabs[kgetOrder(kgetSize(ptr))].size > newSize) {
		// We have to re-get memory
		ret = kmalloc(newSize);
		kmemcopy(ret, ptr, MIN(kgetSize(ptr), newSize));

		kfree(ptr);
	} else
		setSize(ptr, newSize + sizeof(slabHeader));

	return ret;
}

#ifdef CONFIG_MM_STATS
void dump_stats()
{
	printk(KINFO, "Current memory stats for ZealOS MM:\n");
	printk(KINFO, "Slab stats:\n");
	printk(KINFO, "kmalloc's: %lu, kfree'd: %lu, current slabs in use: "
			"%lu\n", stats.kmallocs, stats.kfrees,
			stats.kmallocCurrent);
	printk(KINFO, "buddy tables: %lu\n\n", stats.buddyTables);
	printk(KINFO, "Page stats:\n");
	printk(KINFO, "kmalloc'd: %lu, kfree'd: %lu, current: %lu\n",
			stats.pagesUsed, stats.pagesFree, stats.pagesCurrent);
}
#endif

errRet __preInitFunc slabInit()
{
	// Print some useful(?) info
	printk(KINFO, "Firing up mm\n");
	DPRINT("PAGE_SIZE: %d sizeof buddySlab: %d; sizeof buddyHeader: %d\n",
			PAGE_SIZE, sizeof(struct buddySlab),
			sizeof(struct buddyHeader));

#ifdef CONFIG_MM_STATS
	kmemset((void *)&stats, 0, sizeof(struct stats));
#endif

	return SUCCESS;
}

EXPORT_SYMBOL(kmalloc);
EXPORT_SYMBOL(kfree);
EXPORT_SYMBOL(krealloc);
EXPORT_SYMBOL(kgetSize);

