/*
 * 8259.c - A generic 8259 Programmable IRQ controller
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 26/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 06/07/00 Mark: Many thanks to 'believer' who made this work ;-)
 * 27/06/01 Mark: Moved most of the stuff over to generic/
 * 23/08/01 Zwane: Added spinlock stuff.
 * 13/09/01 Zwane: Added IRQ in service and pending functions, added lock in
 * 		   init function.
 * 13/09/01 Mark: Fixed up some stuff
 */

#include <common.h>
#include <irq.h>
#include <io.h>
#include <stdio.h>
#include <spinlock.h>
#include <kmalloc.h>
#include <printk.h>
#include <debug.h>


spinlock_t i8259Lock = INIT_UNLOCKED;

/* function: ulong IRQInitArch()
 * desc: This initalizes the 8259 PIC's
 */
ulong __preInitFunc IRQInitArch()
{
	ulong flags;

	/* we actually REALLY do not want interrupts happening
	 * during this...*/

	spinlockIRQSave(&i8259Lock, &flags);

        outb_p(0x11, 0x20);		// init
	outb_p(0x11, 0xA0);		// init
	outb_p(IRQ_OFFSET, 0x21);		// offsets
	outb_p(IRQ_OFFSET + 8, 0xA1);		// offsets
	outb_p(4, 0x21);			// cascading
	outb_p(2, 0xA1);			// cascading
	outb_p(1, 0x21);			// 80x86 mode
	outb_p(1, 0xA1);			// 80x86 mode

	// Start values with the controller line to slave enabled
	IRQData->arch.masks[0] = ~(1 << 2);
	IRQData->arch.masks[1] = 0xFF;

	outb_p(IRQData->arch.masks[0], 0x21);
	outb_p(IRQData->arch.masks[1], 0xA1);

	IRQData->arch.cached_reg = IRR_REG;	/* default is IRR anyway */
	spinlockIRQRestore(&i8259Lock, &flags);

	return IRQ_NUM;		// Number of IRQ's we've got
}

/* function: void EOIIntr(ulong IRQ)
 * desc: This function clears the current interrupt from the 8259(s).
 */
void EOIIntr(ulong IRQ)
{
        /* This function is only safe if our calling function has already locked
         *  access to the 8259 with a spinlockIRQSave -Zwane
        */

	/* We used to have it so it just EOI'd all, but that would mess w/ the
	 * queuing, so now we mask a specific IRQ */

	if(IRQ > 7) {				/* Second PIC */
		outb(0x60 | (IRQ & 7), 0xA0);	/* IRQ number */
		IRQ = 2;
	}

	outb(0x60 | IRQ, 0x20);			/* EOI the first PIC */
}

/* function: errRet IRQMask(ulong IRQ, bool enable)
 * desc: This function enables or disables a certain IRQ on the 8259
 */
errRet IRQMask(ulong IRQ, bool enable)
{
	uchar chip = PIC_NUM(IRQ);
        ulong flags;

	if(IRQ >= IRQ_NUM)		// Out of range
		return ETOOBIG;

        spinlockIRQSave(&i8259Lock, &flags);

        if(chip == 1)
		IRQ -= 8;

	if(enable == ENABLE)		// Enable
		btc(&IRQData->arch.masks[chip], IRQ);
	else				// Disable
		bts(&IRQData->arch.masks[chip], IRQ);

	outb(IRQData->arch.masks[chip], (chip == 1)?0xA1:0x21);

	DPRINT("IRQ: %d, chip: %d, mask: 0x%X\n", IRQ, chip,
			IRQData->arch.masks[chip]);

        spinlockIRQRestore(&i8259Lock, &flags);

	return SUCCESS;
}

/* checks for any pending requests for IRQ */
__uint32 IRQPending(ulong IRQ)
{
	uchar chip = PIC_NUM(IRQ);
	__uint32 mask = 1 << IRQ;
	ulong flags;
	int ret;

	/* The IRR can be read when, prior to the RD pulse, a
	 * Read Register Command is issued with OCW3 (RR = 1, RIS = 0.) */

	spinlockIRQSave(&i8259Lock, &flags);

	if (IRQData->arch.cached_reg != IRR_REG) {
		outb(IRR_REG, ((chip == 0) ? 0x20 : 0xA0));
		IRQData->arch.cached_reg = IRR_REG;
	}

	if (chip == 0)
		ret = inb(0x20) & mask;
	else
		ret = inb(0xA0) & (mask >> 8);

	spinlockIRQRestore(&i8259Lock, &flags);

	return ret;
}

/* returns IRQ being serviced */
__uint32 IRQInService(ulong IRQ)
{
	uchar chip = PIC_NUM(IRQ);
	__uint32 mask = 1 << IRQ;
	ulong flags;
	int ret;

	/* The ISR can be read, when, prior to the RD pulse, a
	 * Read Register Command is issued with OCW3 (RR = 1, RIS = 1).*/

	spinlockIRQSave(&i8259Lock, &flags);

	if (IRQData->arch.cached_reg != ISR_REG) {
		outb(ISR_REG, ((chip == 0) ? 0x20 : 0xA0));
		IRQData->arch.cached_reg = ISR_REG;
	}

	if (chip == 0)
		ret = inb(0x20) & mask;
	else
		ret = inb(0xA0) & (mask >> 8);

	spinlockIRQRestore(&i8259Lock, &flags);

	return ret;

}

EXPORT_SYMBOL(IRQMask);
EXPORT_SYMBOL(EOIIntr);
