/*
 * apic.c - Local APIC functions
 *
 * Added to the Zealos project and (C) by Zwane Mwaikambo
 *
 * Current maintainer: Zwane Mwaikambo <zwane@commfireservices.com>
 * Status: Maintained
 * Last update: 25/10/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * Support routines for smp.c and cpu.c, therefore no APICINit the relevant
 * subsystems should call these.
 * 
 * Notes/Fixme's:
 *
 * Changelog:
 * 25/10/01	Zwane: general fixes to try and actually make it *work*
 * 02/12/01	Zwane: the braindamage in this code is beyond even my normal capabilities!!
 * 17/02/02	Zwane: should work RSN ;)
 */

#include <common.h>
#include <apic.h>
#include <cpu.h>
#include <stdio.h>
#include <printk.h>
#include <debug.h>


/* description: a wrapper for reading and writing to the APICs, no need for locks
 * when reading/writing from this area, its done atomically by the CPUs
 */
static __uint32 rdapic(__uint8 num, __uint32 reg)
{
	return *(__uint32 *)(_APIC_ADDRESS_MAP(num) + reg);
}

/* Paging will probably have to be disabled when doing this, note the 32bit value,
 P5 has a problem with 8bit APIC reads/writes
 */
static void wrapic(__uint8 num, __uint32 reg, __uint32 val)
{
	*(_APIC_ADDRESS_MAP(num) + reg) = val;
}	

/* look for a local APIC and enable it if disabled then assign num to the
 * APIC id XXX Turns out we should check wether the BIOS supplied id numbers
 * are acceptable first.
 */
errRet APICDetect(struct CPU *cpu, __uint8 num)
{
	__uint32 h, l;

	/* do a few basic sanity tests */
	if (!cpu)
		return ENOEXIST;

	if (!bt(&cpu->arch.capability, CPU_CAP_MSR))
		return ENOEXIST;

	/* FIXME only i686 and K7 where model > 1 can enable APIC via software
	 * hmm.. how to test for K7 without doing ugly string tests? -Zwane
	 */
	if (cpu->arch.family < 6)
		return ENOEXIST;
	
	rdmsr(APIC_BASE_MSR, h, l);
	if (!bt(&l, APIC_BASE_MSR_FLAG_GE)) {	/* check for global enable flag */
		DPRINT(DEBUG_APIC, "CPU[%lu]: MSR_FLAG_GE disabled, enabling...\n", num);
		l &= ~APIC_BASE_MSR;
		l |= (APIC_BASE_MSR_FLAG_GE | _APIC_ADDRESS_MAP(num));
		wrmsr(APIC_BASE_MSR, h, l);
		DPRINT(DEBUG_APIC, "OK");
	}

	DPRINT(DEBUG_APIC, "CPU[%lu] Enabling APIC...", num);
#if 0
	/* we found a local APIC, now enable it */
	h = rdapic(num, APIC_SPIV);
	h &= ~APIC_VECTOR_MASK;
	h |= APIC_SIV_APIC_EN;
	h |= APIC_SIV_FOCUS_DIS;
	wrapic(num, APIC_SIV_REG, h);
#endif
	/* check the capability flags to see wether it is now really enabled */
	CPUDetectCapability(cpu);
	if (!bt(&cpu->arch.capability, CPU_CAP_APIC)) {
		DPRINT(DEBUG_APIC, "FAIL\n");
		return ENOEXIST;
	}
	
	DPRINT(DEBUG_APIC, "OK\n");
	// wrapic((__uint32*) APIC_ID_REG, num);	/* assign this CPU an id */
	cpu->arch.apic_id = rdapic(num, APIC_ID_REG);
	cpu->arch.apic_ver = rdapic(num, APIC_VERSION);
	
	printk("CPU[%lu] APIC ver %u ID=%d\n", num, cpu->arch.apic_ver,
		cpu->arch.apic_id);
	
	return SUCCESS;
}

