/*
 * cpu.c - Non-generic CPU handling functions for the i386 family of processors
 *
 * Added to the Zealos project and (C) by Mark Zealey
 * Parts (C) by Zwane Mwaikambo <zwane@commfireservices.com>
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 24/06/01 Mark: Fixed up stuff for the new archCPU stuff
 * 26/06/01 Mark: Fixed in_lock to not __local_sti before we have the interrupt
 *	table set up.
 * 26/08/01 Zwane: Fixed CPU detection (previous code assumed cpu to be pentium
 *	when cpuid exists) plus a few cpuid cleanups.
 * 31/08/01 Zwane: Added cpuid cache detection
 * 03/09/01 Zwane: Cleaned up cache detection, more cpuid cleanups.
 * 17/09/01 Zwane: Cache detection fixes (didnt work on some boxes) and more CPU
 *		   detection voodoo ;)
 * 21/09/01 Zwane: Ditto...
 * 23/09/01 Zwane: 386/486 EFlags check fixes, enabled caching
 * 23/09/01 Mark: 386/486 checking fixes
 * 24/09/01 Mark: Fixed caching stuff, moved wbinvd() to the .h file.
 * 30/09/01 Mark: Added reboot() and halt()
 * 30/09/01 Zwane: Modified reboot()
 * 14/10/01 Zwane: Started local APIC code, getting ready for the SMP merge
 * 16/10/01 Zwane: Moved APIC stuff to its own file
 */

#include <common.h>
#include <cpu.h>
#include <machine.h>
#include <desc.h>
#include <printk.h>
#include <stdio.h>
#include <signals.h>
#include <irq.h>
#include <kmalloc.h>
#include <atomic.h>
#include <kernarea.h>
#include <char/keyboard.h>
#include <io.h>
#include <debug.h>
#ifdef APIC
#include <apic.h>
#endif


typedef struct {
	int cap_num;
	char *cap_str;
} CPUCapabilities_t;

const CPUCapabilities_t CPUCapabilities [] = {
	{CPU_CAP_FPU,	"FPU"},
	{CPU_CAP_VME,	"VME"},
	{CPU_CAP_DE,	"DE"},
	{CPU_CAP_PSE,	"PSE"},
	{CPU_CAP_TSC,	"TSC"},
	{CPU_CAP_MSR,	"MSR"},
	{CPU_CAP_PAE,	"PAE"},
	{CPU_CAP_MCE,	"MCE"},
	{CPU_CAP_CX8,	"CX8"},
	{CPU_CAP_APIC,	"APIC"},
	{CPU_CAP_SEP,	"SEP"},
	{CPU_CAP_MTRR,	"MTRR"},
	{CPU_CAP_PGE,	"PGE"},
	{CPU_CAP_MCA,	"MCA"},
	{CPU_CAP_CMOV,	"CMOV"},
	{CPU_CAP_PAT,	"PAT"},
	{CPU_CAP_PSE36,	"PSE36"},
	{CPU_CAP_MMX,	"MMX"},
	{CPU_CAP_FXSR,	"FXSR"},
	{CPU_CAP_SSE,	"SSE"}
};

// Reboot the system
void reboot(__uint32 type)
{
	struct desc null_desc = {limit: 0, base: 0};

#ifdef KEYBOARD
//	int count = 10;		See below
#endif

	/* Disable paging on the CPU, so we can write to the bios area. */
	paging_disable();

	/* BIOS checks what kind of reboot came about by reading this memory
	 * location. Check cpu.h for types */
	*(uint *)0x472 = type;

	/* Doesn't work properly... (keyboard code is fucked) */
#if 0
	while (count--) {
		keybWait();		/* wait for kboard to empty out */
		outb(0xFE, 0x64);	/* now pulse the reset line */
	}
#endif /* KEYBOARD */

	/* Just fuck the system generally... */

	lidt(&null_desc);
	lgdt(&null_desc);

	/* do it the VERY raw way (Call the bios reboot code, prolly cause a
	 * gpf, which won't be handled as we removed the idt, ergo we get a
	 * tripple-fault == reboot ;) */
	((void (*)())0xFFFF0)();
}

// Halt the system
void halt()
{
	disable_local_interrupts();		// Needed?
	while(1);
}

handleInt(exception)
{
	printk(KERROR, "ERROR: %u\n", (uint)r->num);
	dumpRegs(r);
	dumpStack((ulong *)(inKernel(r->eip) ? 0 : r->esp));
	printk(KWARN, "That's all folks, maybe someone wants to debug? ;-)\n");
	halt();
}

#define STACK_DUMP_LEN 40
#define STACK_DUMP_STEP 8
void dumpStack(ulong *esp)
{
	ulong l, length;

	if(!esp)		// If NULL is given grab current stack pos :-)
		esp = (ulong *)&esp;	// (Thanks linux! ;)

	printk(KWARN, "Stack dump, ESP == %p\n", esp);

	// FIXME: We just want the stack depth really...
	for(l = 0; l < STACK_DUMP_LEN; l += STACK_DUMP_STEP) {
		for(length = STACK_DUMP_STEP; length; length--) {
			ulong addr = *esp++;
			printk(KWARN, inKernel(addr)?"<%8lx> ":"%8lx ", addr);
		}
		printk(KWARN, "\n");
	}
}

/* function: void dumpRegs(struct allRegs *)
 * description: This function just spews the current contents of the registers
 * as the function name suggests.
 */
void dumpRegs(struct allRegs *r)
{
	ulong t[3];

	printk(KWARN, "CS: %4x EIP: %8lx\n", (uint)r->cs, r->eip);
	if(inKernel(r->eip))
		printk(KWARN, "In-kernel, old SS and ESP will not be reliable"
				"\n");
	else
		printk(KWARN, "SS: %4x ESP: %8lx\n", (uint)r->ss, r->esp);

	getSeg("ss", t[0]);
	printk(KWARN, "CUR SS: %4x ESP: %8lx\n", t[0], get_stack_pointer());
	printk(KWARN, "EFLAGS: %8lx\n", r->eflags);
	printk(KWARN, "EAX: %8lx EBX: %8lx ECX: %8lx EDX: %8lx\n",
			r->eax, r->ebx, r->ecx, r->edx);
	printk(KWARN, "ESI: %8lx EDI: %8lx EBP: %8lx\n",
			r->esi, r->edi, r->ebp);

	getSeg("fs", t[0]);
	getSeg("gs", t[1]);
	printk(KWARN, "DS: %4x ES: %4X FS: %4x GS: %4x\n",
			r->ds, r->es, t[0], t[1]);

	getCr(0, t[0]);
	getCr(2, t[1]);
	getCr(3, t[2]);
	printk(KWARN, "CR0: %8lx CR2: %8lx CR3: %8lx\n", t[0], t[1], t[2]);
}

/* function: void CPUDetect(ulong num)
 * description: Detects the specified CPU and adds it into the current box.
 */
errRet CPUDetect(ulong num)
{
#if PROCESSOR < 5
	volatile flags eflags;
#endif /* PROCESSOR < 5 */
	ulong *v, eax, ebx, ecx, edx, i;
	struct CPU *CPU;

	CPU = kmalloc(sizeof(struct CPU));
	if (!CPU)
		return ENOMEM;		/* man we're b0rked */

	thisBox.CPU[num] = CPU;

	kmemset((char *)CPU, 0, sizeof(struct CPU));
	v = (__uint32 *)CPU->arch.cpuid;

	CPU->arch.family = 3;		// Set the beginning version
	CPU->arch.model = 0;
	CPU->arch.stepping = 0;
	CPU->arch.maxCPUID = 0;		// debatable...
	CPU->arch.maxExtCPUID = 0;
	CPU->arch.capability = 0;
	CPU->arch.cache[L1] = FALSE;
	CPU->arch.cache[L2] = FALSE;

	/* OK, the way this works is that we set the AC bit in eflags and then
	 * try to set eflags to that. If we're on a i386, it will not let us
	 * change it.
	 */
#if PROCESSOR < 5
	eflags = get_flags();
#endif /* PROCESSOR < 5 */

#if PROCESSOR < 4
	set_flags(eflags ^ 0x40000);		// Flip AC bit

	if(get_flags() == eflags)
		goto print_cpu;
#endif /* PROCESSOR < 4 */

	/* 486+ from here on...*/
	CPU->arch.family++;

	/* is there a way of finding this out programmatically without trying to
	   trigger an exception from doing a wbinvd? -Zwane
	*/
	CPU->arch.cache[L1] = TRUE;

#if PROCESSOR < 5
	set_flags(eflags ^ 0x200000);	// Try the ID Flag now

	/* this would be an old school 486 -Zwane*/
	if(get_flags() == eflags) {
		CPU->arch.cache_size[L1]=8;
		CPU->arch.cache_size[L2]=0;
		goto print_cpu;
	}

	set_flags(eflags);		// restore eflags
#endif /* PROCESSOR < 5 */

	/* 1 is the LCD for P6, P5 and 486s - Zwane */
	CPU->arch.maxCPUID = 1;
	CPUDetectMax(CPU, FALSE);
	CPUDetectMax(CPU, TRUE);
	CPUDetectCapability(CPU);
	/* Woopie, we're on a decent (well, is anything intel-designed decent?)
	 * processor now pentium+ which has the CPUID instruction. The end of
	 * our woes */

	/* Some late intel486s (DX4s) support cpuid so perhaps some more
	 * checks... - Zwane
	 */

	/* the ID string is returned in EBX, ECX, EDX. EAX has the max cpuid */
	cpuid(0, &eax, &v[0], &v[2], &v[1]);	// Do a standard cpuid
	CPU->arch.cpuid[12] = '\0';		// Null-terminate it

	cpuid(1, &eax, &ebx, &ecx, &edx);
	/* the family is in bits 8-11 so shift down 8 bits AND (get it :) get
	 * lower 4 bits. model in 4-7 and stepping in 0-3 -Zwane */

	CPU->arch.family = (eax >> 8) & 0xF;
	CPU->arch.model = (eax >> 4) & 0xF;
	CPU->arch.stepping = eax & 0xF;

	/* Get L1, L2 cache size the std cpuid way and then see if we can try
	 * extended CPUID.
	 * get L1 cache size, upper 8 bits of ECX has our data L1 size in kbytes
	 * upper 8 bits of EDX has our code L1 size in kbytes
	 */

	/* this is ugly but i have to hardcode it */
	if (CPU->arch.family == 5)
		CPU->arch.cache_size[L1] = bt(&CPU->arch.capability,
							CPU_CAP_MMX) ? 16 : 8;

	if (CPU->arch.maxCPUID >= 2) {
		cpuid(2, &eax, &ebx, &ecx, &edx);
		/* XXX we should actually loop eax number of times too */
		for (i = 0; i < 32; i += 8) {
			/* This looks wrong to me... - Mark */
			switch ((edx >> i) & 0xFF) {
				/* 8k I cache */
				case 0x6: CPU->arch.cache_size[L1] += 8; break;

				/* 16k I cache */
				case 0x8: CPU->arch.cache_size[L1] += 16; break;

				 /* 8k D cache */
				case 0xA: CPU->arch.cache_size[L1] += 8; break;

				 /* 16k D cache */
				case 0xC: CPU->arch.cache_size[L1] += 16; break;

				case 0x40: CPU->arch.cache[L2] = FALSE; break;

				case 0x41: CPU->arch.cache_size[L2] = 128;break;

				case 0x42: CPU->arch.cache_size[L2] = 256;break;

				case 0x43: CPU->arch.cache_size[L2] = 512; break;

				case 0x44: CPU->arch.cache_size[L2] = 1024; break;

				case 0x45: CPU->arch.cache_size[L2] = 2048; break;

				default: break;
			}
		}

		/* we got a value for cache so set cache[L2] to true */
		if (CPU->arch.cache_size[L2])
			CPU->arch.cache[L2] = TRUE;
	}

	if (CPU->arch.maxExtCPUID >= 0x80000005) {
		DPRINT("Entering L1 Cache detection using extended CPUID\n");
		cpuid(0x80000005, &eax, &ebx, &ecx, &edx);
		CPU->arch.cache_size[L1] = (ecx >> 24) + (edx >> 24);
	}

	if (CPU->arch.maxExtCPUID >= 0x80000006) {
		DPRINT("Entering L2 Cache detection using extended CPUID\n");
		cpuid(0x80000006, &eax, &ebx, &ecx, &edx);
		CPU->arch.cache_size[L2] = ecx >> 16;

		if (CPU->arch.cache_size[L2])
			CPU->arch.cache[L2] = TRUE;
	}
#ifdef APIC
	APICDetect(CPU, num);	// chances are if we're here we have an APIC
#endif /*APIC */

	goto print_cpu;		// Keep compiler happy
/* print the CPU detected etc. over here */
print_cpu:
	printk(KINFO, "CPU[%lu]: ", num);

	if (CPU->arch.maxCPUID)      /* if we got cpuid we got an id string */
		printk(KINFO, "%s ", (char*)&CPU->arch.cpuid);

        printk(KINFO, "%u86 ", CPU->arch.family);
	if (CPU->arch.maxCPUID)	     /* if we got cpuid we got something here */
		printk(KINFO, "Model %u Stepping %u ", CPU->arch.model,
							CPU->arch.stepping);

	printk(KINFO, "Detected\n");

	if (CPU->arch.maxCPUID)
		printk(KINFO, "CPU[%lu]: Max Std. CPUID = %d\n", num,
						CPU->arch.maxCPUID);

	if (CPU->arch.maxExtCPUID)
		printk(KINFO, "CPU[%lu]: Max Ext. CPUID = 0x%x\n", num,
						CPU->arch.maxExtCPUID);

	/*XXX how do we detect cache in P5? */
	if (CPU->arch.cache[L1])
		printk(KINFO, "CPU[%lu]: L1 cache = %dK\n", num,
						CPU->arch.cache_size[L1]);

        if (CPU->arch.cache[L2])
		printk(KINFO, "CPU[%lu]: L2 cache = %dK\n", num,
						CPU->arch.cache_size[L2]);

	if (CPU->arch.capability) {
		eax = sizeof(CPUCapabilities)/sizeof(CPUCapabilities_t);

		/* man this is ugly... -Zwane */
		printk(KINFO, "CPU[%lu]: Capabilities = 0x%x <", num,
				CPU->arch.capability);

		for (i = 0; i < eax; i++)
			if (bt(&CPU->arch.capability,
						CPUCapabilities[i].cap_num))
				printk(KINFO, "%s ",
						CPUCapabilities[i].cap_str);

		printk(KINFO, "\b>\n");
	}

        return SUCCESS;
}

/* function: void CPUDetectMax(struct CPU *p)
 * description: Detects the max CPUID command.
 * and the max extended if extended == TRUE
 */
void CPUDetectMax(struct CPU *p, bool extended)
{
	__uint32 d;

	if(!p->arch.maxCPUID)		// 0 for not allowed to CPUID
		return;

        /* highest value is returned in EAX */
	if (extended)
		cpuid(0x80000000, &p->arch.maxExtCPUID, &d, &d, &d);
	else
		cpuid(0x00000000, &p->arch.maxCPUID, &d, &d, &d);
}

/* function: void CPUDetectCapability(struct CPU *p)
 * description: Gets the capability of the CPU.
 */
void CPUDetectCapability(struct CPU *p)
{
	__uint32 d;
	if(p->arch.maxCPUID < 0x00000001)
		return;

	/* Intel processors don't support 0x80000001 ??? -Zwane */
	cpuid(0x00000001, &d, &d, &d, &p->arch.capability);
}

/* function: void CPUDetectCPUID(struct CPU *p)
 * description: Detects the CPU by using CPUID.
 */
void CPUDetectCPUID(struct CPU *p)
{
	__uint32 *v = (__uint32 *) p->arch.info;

	if(p->arch.maxExtCPUID < 0x80000004)	// Return if it dont support it
		return;

	// Grab the name etc
	cpuid(0x80000002, &v[0], &v[1], &v[2], &v[3]);
	cpuid(0x80000003, &v[4], &v[5], &v[6], &v[7]);
	cpuid(0x80000004, &v[8], &v[9], &v[10], &v[11]);

	p->arch.info[48] = '\0';		// we add a \0
}

errRet CPUSetup(ulong num)
{
	printk(KINFO, "Detecting CPU[%lu]\n", num);

	if(CPUDetect(num) != SUCCESS)
		return ENOEXIST;

	/* enable the cache lines by clearing the CD (bit 30) and NW (bit 29)
	 * flags in cr0 XXX which CPU are we doing this on, on an SMP box?
	 * -Zwane
	 *
	 * On an SMP box, I'd assume we wanted to clear the things on all the
	 * CPU's, as they each have their own cache's. I'm thinking that we
	 * should just let the bios set up the cache, as it should know better,
	 * but then you have b0rked biosen etc... - Mark
	 *
	 * We have to enable all the caches manually then do an IPI to get the
	 * cpus to wbvind so that we're in a known state, the BIOS setup is far
	 * from optimal with regards to caches, there should be an smp function
	 * to do this and we call it from here. -Zwane
	 */

	if (thisBox.CPU[num]->arch.cache[L1]) {
		__uint32 cr0;
		getCr(0, cr0);
		setCr(0, cr0 & ~0x60000000);
		wb_invalidate_cache();	// Put here to simplify code path..
	}

	return SUCCESS;
}

EXPORT_SYMBOL(dumpRegs);
EXPORT_SYMBOL(dumpStack);
EXPORT_SYMBOL(reboot);
EXPORT_SYMBOL(halt);
