/*
 * desc.c - Descriptor table handling for the i386
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 20/06/01 Mark: Re-worked the file so it uses one function for resizing all
 *	descriptor tables.
 * 01/09/01 Mark: Removed userData and userCode segments, they can be done by
 *	processes and are not needed.
 * 21/10/01 Mark: Fixed the horrid gcc kludges.
 */

#include <common.h>
#include <desc.h>
#include <irq.h>
#include <stdio.h>
#include <printk.h>
#include <cpu.h>
#include <system.h>
#include <signals.h>
#include <video.h>
#include <io.h>
#include <assert.h>
#include <kmalloc.h>
#include <syscall.h>
#include <debug.h>


bool interrupts = FALSE;

bool intSetup()
{
	return interrupts;
}

EXPORT_SYMBOL(intSetup);

errRet descResize(void **beg, ulong *curEntrys, ulong newEntrys,
		ulong maxEntrys, ulong entrySize, void (*reloadFunc)())
{
	void *t = *beg;			// This variables makes the fall-back

	if(*curEntrys > newEntrys)
		return SUCCESS;

	DPRINT("Relocating to %d entries from %d\n", newEntrys, *curEntrys);

	if(newEntrys >= maxEntrys)
		return ETOOBIG;

	if(!(*beg = kmalloc(newEntrys * entrySize))) {
		// Erk, no mem for us to relocate in... Restore the desc and
		// bail out
		*beg = t;
		return ENOMEM;
	}

	kmemset(*beg, 0, newEntrys * entrySize);

	kmemcopy(*beg, t, MIN(newEntrys, *curEntrys) * entrySize);
	*curEntrys = newEntrys;
	kfree(t);	// Free the old table

	return SUCCESS;
}

/* variable: struct gateDesc *IDT
 * description: The IDT. It's defaultly set to NULL, and then kmalloc()'d later
 * on.
 */
struct gateDesc *IDTArea = NULL;

/* variable: ulong IDTEntries
 * description: The variable that holds the current number of entries into the
 * IDT.
 */
ulong IDTEntries = 0;

/* function: void IDTReload()
 * description: Reloads the idt register of the processor, starts up interrupts,
 * and clears NT in eflags.
 */
void IDTReload()
{
	struct desc loadIDT;
	flags eflags;

	loadIDT.limit = IDTEntries * sizeof(struct gateDesc) - 1;
	loadIDT.base = (memory)IDTArea;
	lidt(&loadIDT);			// Load the new IDT

	/* Clear the NT bit in eflags: */
	eflags = get_flags();
	eflags &= ~(1 << 14);
	set_flags(eflags);

	enable_local_interrupts();			// Start interrupts!! :)
}

/* function: errRet setGate(void *handler, uint interrupt, char DPL,
 *				char type)
 * description: This sets a gate in the IDT of the specified type and the
 * specified privilage level. Returns ETOOBIG if the interrupt number is over
 * the maxumum number of interrupts allowed.
 */
errRet setGate(void *handler, ulong interrupt, char DPL, char type)
{
	if(descResize((void **)&IDTArea, &IDTEntries, interrupt,
				MAX_IDT_ENTRIES, sizeof(struct gateDesc),
				IDTReload) != SUCCESS)
		return ENOMEM;

	IDTArea[interrupt].offsetLow = (__uint16)((__uint32)handler & 0xFFFF);
	assert(kernCode);	// GDT prolly hasnt been set up yet...

	IDTArea[interrupt].TSS = kernCode;

	/* The amount of doubles it puts onto the stack when called. May be
	 * needed for syscalls etc
	 */
	IDTArea[interrupt].dwordCount = 0;
	IDTArea[interrupt].type = type;
	IDTArea[interrupt].DPL = DPL;
	IDTArea[interrupt].present = 1;
	IDTArea[interrupt].offsetHigh = (__uint16)((__uint32)handler >> 16);

	DPRINT("Added interrupt %lu pointing at %p\n", interrupt, handler);

	return SUCCESS;
}

/* Ugly hack to avoid waste of space (sorry, I've just been reading the IOCCC
 * entries... :) (www.ioccc.org) - Mark
 * I was making a linked list of this stuff, but I thought that as we've got
 * signals, why not use them? :)
 * (Note, this is for IRQ's :)
 */
#define I(x) startIntr(irq##x, x + IRQ_OFFSET, triggerIRQ);
I(0) I(1) I(2)  I(3)  I(4)  I(5)  I(6)  I(7)
I(8) I(9) I(10) I(11) I(12) I(13) I(14) I(15)
#undef I

// GDB will handle them, no need for us to bother
#ifndef GDB
#define I(x) startIntr(intr##x, x, exception);
I(0)  I(1)  I(2)  I(3) I(4)  I(5)  I(6)  I(7)  I(8)  I(9)
I(10) I(11) I(12) I(13) I(14) I(15) I(16) I(17) I(18) I(19)
/*I(20) I(21) I(22) I(23) I(24) I(25) I(26) I(27) I(28) I(29)
I(30) I(31)*/		// Extended exceptions
#undef I
#endif

// 1 is not the true value but...
startInt(intNull,1) {
	printk(KWARN, "Unhandled interrupt called, whoever let that happen "
			"should be shot, but we'll carry on anyway...\n");
}

startInt(beforeSys, SYSCALL_INT)
{
	panic("Syscall called before syscall's were set up. Prolly an error so"
			" I'll hang.\n");
}

/* function: void unsetGate(ulong interrupt)
 * description: Free a gate of the specified interrupt.
 */
void unsetGate(ulong interrupt)
{
	setIntGate(intNull, interrupt, 3);	// reset to the default handler
}

/* function: initFunc IDTInit()
 * type: init
 * description: The IDT initalization routines.
 */
initFunc __preInitFunc IDTInit()
{
	if(!(IDTArea = kmalloc(IRQAddr(IRQNum) * sizeof(struct gateDesc))))
		return ENOMEM;

	IDTEntries = IRQAddr(IRQNum);
	kmemset((void *)IDTArea, 0, IDTEntries * sizeof(struct gateDesc));

#ifndef GDB
	// Setup the exception handlers. (if GDB isn't going to)
#define I(x) setIntGate(intr##x,  x,  0 /* This is DPL... */);
	I(0)  I(1)  I(2)  I(3)  I(4)  I(5)  I(6)  I(7)  I(8)  I(9)
	I(10) I(11) I(12) I(13) I(14) I(15) I(16) I(17) I(18) I(19)
#undef I
#endif

	setIntGate(beforeSys, SYSCALL_INT, 0);	// Catch premature syscalls

	// Setup the interrupt handlers. (for the IRQ's)
#define I(x) setIntGate(irq##x, IRQAddr(x),  3 /* DPL (FIXME: should be 0? */);
	I(0)  I(1)  I(2)  I(3)  I(4)  I(5)  I(6)  I(7)
	I(8)  I(9) I(10) I(11) I(12) I(13) I(14) I(15)
#undef I

	enable_local_interrupts();		// Start interrupts

	IDTReload();
	interrupts = TRUE;

	return SUCCESS;
}

/* function: INTHandler
 * description: This is the handler that /all/ interrupts on the machine go
 * though. It calls off the specified handler and saves all registers.
 */
asm(
	".globl INTHandler		\n"
	".globl firstForkReturn		\n"
	".globl endIntSpecial		\n"
	"INTHandler:			\n"
	"	pushl %ebx		\n"	// Push the call-number
	"	pushl %edx		\n"	// Push all
	"	pushl %edi		\n"
	"	pushl %esi		\n"
	"	pushl %ebp		\n"
	"	pushl %ds		\n"
	"	pushl %es		\n"
	"	movw %ss, %dx		\n"	// Get correct segs (ss is
	"	movw %dx, %ds		\n"	// always valid)
	"	movw %dx, %es		\n"
	"	movl %esp, %edx		\n"	/* edx is the location of the
						 * real stack now (all args are
						 * now for the function call)
						 */
	"	jmp *%ecx		\n"	// Call the extra function(s)
	"endIntSpecial:			\n"
	"	pushl %edx		\n"	/* Push the location of the
						 * 'real' stack. This does 2
						 * things, 1 it gives the
						 * function it's struct of
						 * registers and 2, it lets us
						 * magically restore the stack
						 * when we leave the function
						 * call */
	"	call *%eax		\n"	// Call the handler!
	"firstForkReturn:		\n"	// Return location when we fork
	"	popl %edx		\n"
	"	movl %edx, %esp		\n"	// Restore our stack
	"	popl %es		\n"	// Pop all
	"	popl %ds		\n"
	"	popl %ebp		\n"
	"	popl %esi		\n"
	"	popl %edi		\n"
	"	popl %edx		\n"
	"	popl %ebx		\n"	// this one for regs->num
	"	popl %ecx		\n"
	"	popl %ebx		\n"
	/* Ok, this next part *really* pisses me off. I spent many hours coding
	 * to make it only pop if we pushed it, and even more hours debugging.
	 * Now I've discovered that iret doesn't pop it off, so we just pop it
	 * off anyway... bloody intel. - Mark
	 */
	"	popl %eax		\n"	// Restore the original eax
	"	addl $4, %esp		\n"	// clear the fake 0 or cpu arg
	"	iret			\n");	// Yes, iret *not* ret

/* This gets the CPU-pushed info and puts it on the stack as an extra argument
 * to exception handlers that get an argument passed to them by the CPU.
 */
asm(	".globl exceptHandler				\n"
	"exceptHandler:					\n"
	"	movl " STR(ARG_OFFSET) "(%esp), %ecx	\n"
	"	pushl %ecx				\n"
	"	jmp endIntSpecial			\n");

EXPORT_SYMBOL(setGate);
EXPORT_SYMBOL(unsetGate);

////////////////////////////////////////////////////////////////////////////////
// GDT stuff

/* comment: The GDT Under ZealOS
 * GDT under Zealos is very simply handled. As you should know, the first GDT
 * entry *Must* be NULL. The next 2 entrys are Zealos' kernel code segment and
 * kernel data segment, and the 2 after that are the user code and the user data
 * segment. After this, we only need 2 more entries, one for the current
 * process' TSS, and the other for the current process' LDT. - Mark
 */

#define MAX_GDT_ENTRIES 0x1000

union segDesc *GDTArea = NULL;
ulong GDTEntries = 0;
__uint16 kernCode, kernData;

/* function: void GDTReload()
 * description: Reloads the gdt register of the processor.
 */
void GDTReload()
{
	struct desc loadGDT;
	loadGDT.limit = GDTEntries * sizeof(struct gateDesc) - 1;
	loadGDT.base = (memory)GDTArea;
	lgdt(&loadGDT);			// Load the new IDT
}

/* function: __uint16 GDTAddEntry(__uint32 base, __uint32 limit, bool gran,
 *			bool bigSeg, __uint8 opts, char DPL, boot descType)
 * arg0: The base address (if gran == TRUE, multiplied by 4096)
 * arg1: Length of the entry, again if gran is true * 4096
 * arg2: multiply everything by 4096 or 1?
 * arg3: 32 bit addresses or 16 bit addresses? TRUE for 32 bit
 * arg4: The options you can pass to it, (use defines)
 * arg5: The privilage level for the segment (0 kernel, 3 user)
 * arg6: TRUE for gdt, FALSE for TSS
 *
 * description: Adds a entry into the GDT table. Returns the number of the GDT
 * entry it added it to or NULL if it couldnt.
 */
__uint16 GDTAddEntry(__uint32 base, __uint32 limit, bool gran, bool bigSeg,
			__uint8 opts, uchar DPL, bool descType)
{
	ulong i;
	for(i = 1; i < GDTEntries; i++)		// Search thru for a free entry
		if(!GDTArea[i].norm.ZealOS)
			break;

	if(i == GDTEntries)
		if(descResize((void **)&GDTArea, &GDTEntries, GDTEntries + 4,
					MAX_GDT_ENTRIES, sizeof(union segDesc),
					GDTReload) != SUCCESS)
			return ENOMEM;

	DPRINT("Adding GDT entry %lu\n", i);

	// Present but not activated
	GDTArea[i].norm.ZealOS = 1;

	assert(limit <= 0xFFFFF);		// Limit too big

	// Limits
	GDTArea[i].norm.limitLow = limit & 0xFFFF;
	GDTArea[i].norm.limitHigh = limit >> 16;

	// Bases
	GDTArea[i].norm.baseLow = base & 0xFFFF;
	GDTArea[i].norm.baseMed = (base >> 16) & 0xFF;
	GDTArea[i].norm.baseHigh = base >> 24;

	// Type
	GDTArea[i].norm.descType = descType;

	// Set the options
	GDTArea[i].norm.type = opts;
	GDTArea[i].norm.DPL = DPL;

	// granularity?
	GDTArea[i].norm.granularity = gran?1:0;

	// 32 or 16 bit addressing?
	GDTArea[i].norm.segmentSize = bigSeg?1:0;

	// OK, it's now present
	GDTArea[i].norm.present = 1;

	return i * sizeof(union segDesc);
}

/* function: void GDTDelEntry(char entry)
 * description: Removes the specified GDT entry.
 */
void GDTDelEntry(__uint16 entry)
{
	GDTArea[entry].dummy = 0;
}

LABEL_DECL(GDT_Reload);

/* function: initFunc GDTInit()
 * type: init
 * description: Initalises the GDT.
 */
initFunc __preInitFunc GDTInit()
{
	if(!(GDTArea = kmalloc(4 * sizeof(union segDesc))))
		return ENOMEM;

	GDTEntries = 4;
	kmemset((void *)GDTArea, 0, sizeof(union segDesc) * GDTEntries);

	// Set up kernel segments
	kernData = GDTAddEntry(0, 0xFFFFF, TRUE, TRUE, SEG_DATA | SEG_RW, 0,
			TRUE);
	assert(kernData);

	kernCode = GDTAddEntry(0, 0xFFFFF, TRUE, TRUE, SEG_CODE | SEG_RW, 0,
			TRUE);
	assert(kernCode);

	/* No need for userspace stuff here yet */

	GDTReload();

	// Reload the registers
	loadSeg("ds", kernData);
	loadSeg("es", kernData);
	loadSeg("fs", kernData);
	loadSeg("gs", kernData);
	loadSeg("ss", kernData);

	printk(KINFO, "Reloading GDT: ");

	/* Used to use && (address-of-label) here, but that is very broken on
	 * many gcc's. Zwane and Mark worked out a better system. */
	ljmp(kernCode, (memory)LABEL_ADDR(GDT_Reload));
LABEL(GDT_Reload);
	printk(KINFO, "done!\n");
	return SUCCESS;
}

EXPORT_SYMBOL(GDTAddEntry);
EXPORT_SYMBOL(GDTDelEntry);

////////////////////////////////////////////////////////////////////////////////
// LDT stuff

// FIXME: make it do dynamic-resizing of the LDT
void LDTAddEntry(union segDesc *t, ulong base, ulong limit, uchar type,
				uchar DPL)
{
	t->norm.limitLow = limit & 0xFFFF;
	t->norm.limitHigh = limit >> 16;
	t->norm.baseLow = base & 0xFFFF;
	t->norm.baseMed = (base >> 16) & 0xFF;
	t->norm.baseHigh = base >> 24;
	t->norm.type = type;
	t->norm.DPL = DPL;
	t->norm.present = TRUE;
	t->norm.descType = TRUE;
	t->norm.granularity = FALSE;
	t->norm.segmentSize = TRUE;
}

EXPORT_SYMBOL(LDTAddEntry);
