/*
 * gdb-stub.c - an i386-remote debug stub for handling gdb over serial
 *
 * Added to the Zealos project by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * This library is taken from the sigops web site and modified. All original
 *	comments etc are marked with a '**'. This is intended to be
 *	self-contained, and only uses io.h defines and a few routines from
 *	stdio. - Mark
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 23/09/01 Zwane: Updated for spinlocks
 * 31/10/01 Mark: Fixed serial code
 * 01/11/01 Mark: Removed the 'our own stack' code and stuff, was wasting space
 *	and wouldn't have worked anyway. Also made other parts of the kernel
 *	more ugly than they needed to be.
 */

/* Com ports are as follows:
 * unix name | DOS name | base address | IRQ
 * ----------+----------+--------------+----
 * ttyS0     | COM1     | 0x3f8        | 4
 * ttyS1     | COM2     | 0x2f8        | 3
 * ttyS2     | COM3     | 0x3e8        | 4
 * ttyS3     | COM4     | 0x2e8        | 3
 *
 * DIVISOR is the number we divide 115200 baud by to get the rate, common ones
 * are:
 * 1  == 115200bps
 * 3  == 38400bps
 * 12 == 9600bps
 */
#define COMBASE 0x3f8
#define COMIRQ 4
#define DIVISOR 1

// No fancy functions stuff here

#include <common.h>
#include <io.h>
#include <printk.h>
#include <video.h>
#include <stdio.h>
#include <cpu.h>
#include <stdio.h>
#include <desc.h>
#include <signals.h>
#include <irq.h>
#include <kmalloc.h>
#include <spinlock.h>

#define BUF_LEN 20
char buf[BUF_LEN];
volatile ulong bufWriteP = 0;
ulong bufReadP = 0;
spinlock_t serialLock = INIT_UNLOCKED;

// This IRQ is now unusable for anyone else...
startInt(serial, IRQAddr(COMIRQ))
{
	uchar in;
	ulong flags;

	spinlockIRQSave(&serialLock, &flags);
	EOIIntr(r->num - IRQAddr(0));
	in = inb(COMBASE + 2);

	if(in != 4) {		// Weird interrupt. We only want reads.
		DPRINT("Weird interrupt: %x\n", in);
	} else {
		in = inb(COMBASE);
		if(in == 3) {		// ^C, break now
			DPRINT("Breaking from ^C\n");
			r->eflags |= 0x100;		// Set the trace bit
		} else {
			buf[bufWriteP++] = in;

			if(bufWriteP >= BUF_LEN)
				bufWriteP = 0;
		}
	}

	spinlockIRQRestore(&serialLock, &flags);
}

void init_serial()
{
	/* Disable interrupts */
	outb_p(0, COMBASE + 4);

	/* Set then clear dlab to reprogram the transmision rate */
	outb_p(0x80, COMBASE + 3);
	outb_p(DIVISOR >> 8, COMBASE + 1);
	outb_p(DIVISOR & 0xFF, COMBASE + 0);
	outb_p(3, COMBASE + 3);

	/* Clear any pending interrupts */
	outb_p(0, COMBASE + 2);

	/* This IRQ is now unusable for anyone else... */
	setIntGate(serial, IRQAddr(COMIRQ), 0);
	IRQAdd(COMIRQ);

	/* Enable master interrupt mask line */
	outb_p(8, COMBASE + 4);

	/* Enable read interrupts */
	outb_p(1, COMBASE + 1);
}

char getDebugChar()
{
	char t;

	while(bufReadP == bufWriteP)
		;
	t = buf[bufReadP++];

	if(bufReadP >= BUF_LEN)
		bufReadP = 0;

	return t;
}

void putDebugChar(char ch)
{
	while(!(inb(COMBASE + 5) & 0x20))
		/* Wait */;

	outb(ch, COMBASE);
}

/******************************************************************************
 **
 **		THIS SOFTWARE IS NOT COPYRIGHTED
 **
 ** HP offers the following for use in the public domain. HP makes no
 ** warranty with regard to the software or it's performance and the
 ** user accepts the software "AS IS" with all faults.
 **
 ** HP DISCLAIMS ANY WARRANTIES, EXPRESS OR IMPLIED, WITH REGARD
 ** TO THIS SOFTWARE INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 ** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 **
******************************************************************************/

/*****************************************************************************
 **	Header: remcom.c,v 1.34 91/03/09 12:29:49 glenne Exp $
 **
 **	Module name: remcom.c $
 **	Revision: 1.34 $
 **	Date: 91/03/09 12:29:49 $
 **	Contributor:	Lake Stevens Instrument Division$
 **
 **	Description:	low level support for gdb debugger. $
 **
 **	Considerations:	only works on target hardware $
 **
 **	Written by:	Glenn Engel $
 **	ModuleState:	Experimental $
 **
 **	NOTES:		See Below $
 **
 **	Modified for 386 by Jim Kingdon, Cygnus Support.
 **
 **	To enable debugger support, two things need to happen.	One, a
 **	call to set_debug_traps() is necessary in order to allow any breakpoints
 **	or error conditions to be properly intercepted and reported to gdb.
 **	Two, a breakpoint needs to be generated to begin communication.	This
 **	is most easily accomplished by a call to breakpoint().	Breakpoint()
 **	simulates a breakpoint by executing a trap #1.
 **
 **	The external function exceptionHandler() is
 **	used to attach a specific handler to a specific 386 vector number.
 **	It should use the same privilege level it runs at.	It should
 **	install it as an interrupt gate so that interrupts are masked
 **	while the handler runs.
 **
 **	Because gdb will sometimes write to the stack area to execute function
 **	calls, this program cannot rely on using the supervisor stack so it
 **	uses it's own stack area reserved in the int array remcomStack.
 **
 **************
 **
 **	The following gdb commands are supported:
 **
 ** command		function				Return value
 **
 **	g		return the value of the CPU registers	hex data or ENN
 **	G		set the value of the CPU registers	OK or ENN
 **
 **	mAA..AA,LLLL	Read LLLL bytes at address AA..AA	hex data or ENN
 **	MAA..AA,LLLL: Write LLLL bytes at address AA.AA	OK or ENN
 **
 **	c		Resume at current address		SNN	( signal NN)
 **	cAA..AA	Continue at address AA..AA		SNN
 **
 **	s		Step one instruction			SNN
 **	sAA..AA	Step one instruction from AA..AA	SNN
 **
 **	k		kill
 **
 **	?		What was the last sigval ?		SNN	(signal NN)
 **
 ** All commands and responses are sent with a packet which includes a
 ** checksum.	A packet consists of
 **
 ** $<packet info>#<checksum>.
 **
 ** where
 ** <packet info> :: <characters representing the command or response>
 ** <checksum>	:: < two hex digits computed as modulo 256 sum of <packetinfo>>
 **
 ** When a packet is received, it is first acknowledged with either '+' or '-'.
 ** '+' indicates a successful transfer.	'-' indicates a failed transfer.
 **
 ** Example:
 **
 ** Host:			Reply:
 ** $m0,10#2a		+$00010203040506070809101112131415#42
 **
 *****************************************************************************/

/* BUFMAX defines the maximum number of characters in inbound/outbound buffers
 * at least NUMREGBYTES*2 are needed for register packets
 */
#define BUFMAX 400

// are we initalized yet?
static char initialized = FALSE;

// TRUE prints ill-formed commands in valid packets & checksum errors
bool remoteDebug = FALSE;

static const char hexchars[]="0123456789abcdef";

/* Here lies the exception number (since GDB only gets the signal
 * number through the usual means, and that's not very specific).
 */
int gdbVector = -1;

#define debugError(a...) if(remoteDebug) printk(a)

#define U(num) startCall(except##num, num, handleException, endIntSpecial, \
		"pushl $0");
#define M(num) startCall(except##num, num, handleException, endIntSpecial, "");
U(0)  U(1)  U(2)  U(3)	U(4)  U(5)  U(6)  U(7)  M(8)  U(9)
M(10) M(11) M(12) M(13) M(14) U(15) U(16) U(17) U(18) U(19)
#undef U
#undef M

int hex(char ch)
{
	if ((ch >= 'a') && (ch <= 'f')) return ch - 'a' + 10;
	if ((ch >= '0') && (ch <= '9')) return ch - '0';
	if ((ch >= 'A') && (ch <= 'F')) return ch - 'A' + 10;
	return -1;
}

/* scan for the sequence $<data>#<checksum> */
void getpacket(char *buffer)
{
	uchar checksum, xmitcsum;
	int i, count;
	char ch;

	do {
		/* wait around for the start character, ignore all other
		 * characters */
		while ((ch = getDebugChar()) != '$')
			/* Do nothing */;
		checksum = 0;
		xmitcsum = -1;

		count = 0;

		/* now, read until a # or end of buffer is found */
		while (count < BUFMAX) {
			ch = getDebugChar();
			if (ch == '#')
				break;
			checksum = checksum + ch;
			buffer[count] = ch;
			count = count + 1;
		}
		buffer[count] = 0;

		if (ch == '#') {
			// Get it's checksum
			xmitcsum = hex(getDebugChar()) << 4;
			xmitcsum += hex(getDebugChar());
			if (checksum == xmitcsum) {	// successful transfer
				putDebugChar('+');
				/* if a sequence char is present, reply the
				 * sequence ID */
				if (buffer[2] == ':') {
					putDebugChar( buffer[0] );

					putDebugChar( buffer[1] );
					/** remove sequence chars from buffer */
					count = length(buffer);
					for (i=3; i <= count; i++)
						buffer[i-3] = buffer[i];
				}
			} else {			// Bad checksum
				putDebugChar('-');	// failed checksum
				debugError("bad checksum. My count = 0x%x, "
						"sent=0x%x. buf=%s\n",
						checksum, xmitcsum, buffer);
			}
		}
	} while (checksum != xmitcsum);
#ifdef CONFIG_GDB & OPT_IS_DEBUG
	printk(KINFO, "> %s\n", buffer);
#endif
}

/* send the packet in buffer. */
void putpacket(char *buffer)
{
	uchar checksum;
	int count;
	char ch;

#ifdef CONFIG_GDB & OPT_IS_DEBUG
	printk(KINFO, "< %s\n", buffer);
#endif

	/* $<packet info>#<checksum>. */
	do {
		putDebugChar('$');
		checksum = 0;
		count = 0;

		while((ch = buffer[count])) {
			checksum += ch;
			count++;
			putDebugChar(ch);
		}

		putDebugChar('#');
		putDebugChar(hexchars[checksum / 16]);
		putDebugChar(hexchars[checksum % 16]);
	} while(getDebugChar() != '+');
}

char inBuf[BUFMAX];
char outBuf[BUFMAX];

/* Address of a routine to RTE to if we get a memory fault. */
static void (*volatile mem_fault_routine)() = NULL;

/* Indicate to caller of mem2hex or hex2mem that there has been an
 * error. */
static volatile bool mem_err = 0;

/* convert the memory pointed to by mem into hex, placing result in buf
 * return a pointer to the last char put in buf (null)
 * If may_fault is non-zero, then we should set mem_err in response to
 * a fault; if zero treat a fault like any other fault in the stub.	*/
void handle_mem2hex_err();
char *mem2hex(char *mem, char *buf, uint count, bool may_fault)
{
	uint i;
	uchar ch;

	if (may_fault)
		mem_fault_routine = handle_mem2hex_err;
	for (i = 0; i < count; i++) {
		ch = *mem++;
		*buf++ = hexchars[ch >> 4];
		*buf++ = hexchars[ch % 16];
	}

	// Ugly, Ugly, Ugly, but it works, unlike anything else I tried.. :/
	asm(
"	jmp 1f			\n"
"handle_mem2hex_err:		\n"
	);
	mem_err = TRUE;
asm("1:");

	*buf = 0;		// EIK!
	if(may_fault)
		mem_fault_routine = NULL;

	return buf;
}

/* convert the hex array pointed to by buf into binary to be placed in mem */
/* return a pointer to the character AFTER the last byte written */
void handle_hex2mem_err();
char *hex2mem(char *buf, char *mem, int count, bool may_fault)
{
	int i;
	unsigned char ch;

	if (may_fault)
		mem_fault_routine = handle_hex2mem_err;
	for (i = 0; i < count; i++) {
		ch = hex(*buf++) << 4;
		ch += hex(*buf++);
		*mem++ = ch;
	}

	// Ugly, Ugly, Ugly, but it works, unlike anything else I tried.. :/
	asm(
"	jmp 1f			\n"
"handle_hex2mem_err:		\n"
	);
	mem_err = TRUE;
asm("1:");

	if(may_fault)
		mem_fault_routine = NULL;
	return mem;
}

static const uchar sigTbl[] = {
	[0] 8,		/* Divide by zero */
	[1] 5,		/* debug exception */
	[3] 5,		/* breakpoint */
	[4] 16,		/* into instruction (overflow) */
	[5] 16,		/* bound instruction */
	[6] 4,		/* Invalid opcode */
	[7] 8,		/* coprocessor not available */
	[8] 7,		/* double fault */
	[9] 11,		/* coprocessor segment overrun */
	[10] 11,	/* Invalid TSS */
	[11] 11,	/* Segment not present */
	[12] 11,	/* stack exception */
	[13] 11,	/* general protection */
	[14] 11,	/* page fault */
	[16] 7,		/* coprocessor error */
};

/**********************************************/
/* WHILE WE FIND NICE HEX CHARS, BUILD AN INT */
/* RETURN NUMBER OF CHARS PROCESSED           */
/**********************************************/
int hexToInt(char **ptr, int *intValue)
{
	int numChars = 0;
	int hexValue;

	*intValue = 0;

	while (**ptr) {
		hexValue = hex(**ptr);
		if (hexValue >= 0) {
			*intValue = (*intValue << 4) | hexValue;
			numChars ++;
		} else
			break;
		(*ptr)++;
	}

	return numChars;
}

// This struct describes the sort of stack gdb wants sent to it..
struct gdbReq {
	ulong eax;
	ulong ecx;
	ulong edx;
	ulong ebx;
	ulong esp;
	ulong ebp;
	ulong esi;
	ulong edi;
	ulong eip;
	ulong eflags;
	ulong cs;
	ulong ss;
	ulong ds;
	ulong es;
	ulong fs;
	ulong gs;
};

// This function does all command procesing for interfacing to gdb.
handleInt(handleException)
{
	char *ptr;
	int addr, length;
	struct gdbReq regs;
	uchar sigval;

	/* Looks like we faulted on memory access to get here... We can handle
	 * this... - Mark */
	if(mem_fault_routine) {
//		putsColor("EIK!", FG(cNormal) | BG(cBlack));
		r->eip = (memory)mem_fault_routine;
		return;
	}

	gdbVector = r->num;

	DPRINT("vector=%d, flags=0x%x, instr=0x%x\n", r->num, r->eflags,
			r->eip);

	sigval = (r->num < sizeof(sigTbl))?sigTbl[r->num]:7;

	// reply to host that an exception has occurred
	outBuf[0] = 'S';
	outBuf[1] = hexchars[sigval >> 4];
	outBuf[2] = hexchars[sigval % 16];
	outBuf[3] = 0;

	putpacket(outBuf);

	while (1) {
		outBuf[0] = 0;
		getpacket(inBuf);
		switch (inBuf[0]) {
		case '?':			// Repeat
			outBuf[0] = 'S';
			outBuf[1] = hexchars[sigval >> 4];
			outBuf[2] = hexchars[sigval % 16];
			outBuf[3] = 0;
			break;

		case 'd':			// Toggle debug
			remoteDebug = !remoteDebug;
			break;

		case 'D':			// Remote disconnected
			initialized = FALSE;
			printk(KINFO, "Remote GDB host disconnected\n");

			/* Clear the step if it's set */
			r->eflags &= 0xfffffeff;

			return;		// Exit

		case 'g':			// Send reg's
			regs.ebx = r->ebx;
			regs.ecx = r->ecx;
			regs.edx = r->edx;
			regs.esi = r->esi;
			regs.edi = r->edi;
			regs.ebp = r->ebp;
			regs.ds = r->ds;
			regs.es = r->es;
			getSeg("fs", regs.fs);
			getSeg("gs", regs.gs);

			regs.eax = r->eax;
			regs.eip = r->eip;
			regs.cs = r->cs;
			regs.eflags = r->eflags;
			// In-kernel, doesn't push esp/ss for same-ring
			// interrupts...
			regs.esp = (ulong)r;
			getSeg("ss", regs.ss);

			/* return the value of the CPU registers */
			mem2hex((char *)&regs, outBuf, sizeof(regs), FALSE);
			break;

		case 'G':			// Set reg's
			hex2mem(&inBuf[1], (char *)&regs, sizeof(regs), FALSE);
			kstrcopy(outBuf, "OK");

			r->ebx = regs.ebx;
			r->ecx = regs.ecx;
			r->edx = regs.edx;
			r->esi = regs.esi;
			r->edi = regs.edi;
			r->ebp = regs.ebp;
			r->ds = regs.ds;
			r->es = regs.es;

			r->eax = regs.eax;
			r->eip = regs.eip;
			r->cs = regs.cs;
			r->eflags = regs.eflags;
			// Stack changing is dangerous (and not possible)
			loadSeg("fs", regs.fs);
			loadSeg("gs", regs.gs);

			break;

		case 'm':/* mAA..AA,LLLL Read LLLL bytes at address AA..AA */
			/* TRY TO READ %x,%x. IF SUCCEED, SET PTR = 0 */
			ptr = &inBuf[1];
			if(hexToInt(&ptr, &addr) && *(ptr++) == ',' &&
						hexToInt(&ptr, &length)) {
				ptr = 0;
				mem_err = 0;
				mem2hex((char*) addr, outBuf, length, TRUE);
				if(mem_err) {
					kstrcopy(outBuf, "E03");
					debugError("memory fault");
				}
			}

			if(ptr) {
				kstrcopy(outBuf,"E01");
				debugError("malformed read memory command: %s",
						inBuf);
			}
			break;

		/* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
		case 'M':
			/* TRY TO READ '%x,%x:' */
			ptr = &inBuf[1];
			if (hexToInt(&ptr,&addr) && *(ptr++) == ',' &&
						hexToInt(&ptr,&length) &&
						*(ptr++) == ':') {
				mem_err = 0;
				hex2mem(ptr, (char*) addr, length, TRUE);
				if (mem_err) {
					kstrcopy(outBuf, "E03");
					debugError("memory fault");
				} else
					kstrcopy(outBuf,"OK");

				ptr = 0;
			}
			if (ptr) {
				strcpy(outBuf,"E02");
				debugError("malformed write memory command: %s",
						inBuf);
			}
			break;

		/* cAA..AA	Continue at address AA..AA(optional) */
		/* sAA..AA	Step one instruction from AA..AA(optional) */
		case 'c':
		case 's':
			ptr = inBuf + 1;
			/* try to read optional parameter, eip unchanged if no
			 * parm */
			if(hexToInt(&ptr, &addr))
				r->eip = addr;

			/* set the trace bit if we're stepping, otherwise clear
			 * it */
			if(inBuf[0] == 's')
				r->eflags |= 0x100;
			else
				r->eflags &= 0xfffffeff;

			DPRINT("'%c' at EIP = %lx, eflags: %lx\n", inBuf[0],
					r->eip, r->eflags);

			return;	   // Will load the new values off the stack
			/* never reached */

		/* kill the program */
		case 'k':	/* do nothing */
			break;

		default:
//			printk(KINFO, "Unrecognised command: %s\n", inBuf);
		}

		putpacket(outBuf);		// reply to the request
	}
}

// Send kernel messages through to gdb (via the serial link)
// Ported from a linux patch by David Woodhouse <dwmw2@infradead.org>.

spinlock_t gdb_out_lock = INIT_UNLOCKED;
#define PACKET_CHUNK 8		// Put 8 bytes of text at a time
void gdb_puts(char *str)
{
	if(initialized) {
		char out[PACKET_CHUNK * 2 + 2];
		uint l = length(str);
		ulong flags;

		out[0] = 'O';

		spinlockIRQSave(&gdb_out_lock, &flags);
		__local_sti();
		/* If interrupts are disabled, leave.. */
		while(l) {
			uint i = (l > PACKET_CHUNK) ? PACKET_CHUNK : l;
			mem2hex(str, out + 1, i, FALSE);
			out[(i * 2) + 1] = 0;	// Null terminate
			putpacket(out);
			str += i;
			l -= i;
		}
		spinlockIRQRestore(&gdb_out_lock, &flags);
	} else
		putsColor(str, FG(cNormal) | BG(cBlack));
}

/* The following was added by Mark to interface with zealos proper */
initFunc __preInitFunc GDBInit()
{
	uchar t;

	init_serial();

#define I(x) setTrapGate(except##x,  x,  0 /* This is DPL... */);
	I(0)  I(1)  I(2)  I(3)	I(4)  I(5)  I(6)  I(7)  I(8)  I(9)
	I(10) I(11) I(12) I(13) I(14) I(15) I(16) I(17) I(18) I(19)
#undef I

	printk(KINFO, "Waiting for GDB to connect\n");

	putDebugChar('+');		// 'moo!'

	/* In case GDB is started before us, ack any packets (presumably
	 * "$?#xx") sitting there. */
	while(getDebugChar() != '$');
	while(getDebugChar() != '#');
	t = getDebugChar();		// Munge checksum
	t = getDebugChar();

	putDebugChar('+');		// And ACK

	printk(KINFO, "GDB connected, stopping right now :-)\n");

	BREAKPOINT();
	initialized = TRUE;
	return SUCCESS;
}
