/*
 * init.c - Do i386-specific kernel initalizations (with things taken from the
 *	bootloader)
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 * Fix the E820 memory map if we multiboot, as it is corrupted by grub, so we
 * need to sanify it, just fakes a memory map atm - Mark
 *
 * Changelog:
 * 13/10/01 Mark: Rewrote very large parts of theis file to support the
 * multiboot specification
 */


#include <common.h>
#include <cpu.h>
#include <init.h>
#include <stdio.h>
#include <printk.h>
#include <machine.h>

/* 2048 bytes inital stack */
#define STACK_SIZE 2048

char commandLine[MAX_COMMAND_LEN] = "";		/* The un-parsed command line */
struct multiboot_info *mbootInfo = NULL;	/* Filled in if we were loaded
						 * by a multi-boot compatable
						 * loader */
struct bootloader_info *bootInfo = NULL;	/* Filled in if we were loaded
						 * by our bootloader */
struct memMap mmap[MMAP_ENTRIES];		/* A copy of the memory map */
ulong mmapLen = 0;				/* No mmap if 0 */
memory	kernelStart = (memory)&kernStart,
	kernelLength = (memory)&kernLen;
char stack[STACK_SIZE];				/* A base stack for the kernel*/

void _start();


#ifdef MULTIBOOT
/* The multiboot structure */
struct multiboot_hdr __SECTION__(".mboot") __ALIGNED__(4) hdr = {
	magic: MULTIBOOT_HDR_MAGIC,
	flags: MULTIBOOT_FLAGS,
	checksum: -(MULTIBOOT_HDR_MAGIC + MULTIBOOT_FLAGS),
	header_addr: (memory)&hdr,
	load_addr: (memory)&textStart,
	load_end_addr: (memory)&loadEnd,
	bss_end_addr: (memory)&bssEnd,
	entry_addr: (memory)&_start,
};
#endif /* MULTIBOOT */

/* This retreves and parses the command line parameters passed by lilo at boot
 * time.
 */
errRet __preInitFunc parseLilo()
{
	__uint8 k = MAX_COMMAND_LEN - 1;
	/* A buffer of the options passed by lilo */
	char i[MAX_COMMAND_LEN], *j, *sp=i, *end=i;

	if(!commandLine[0])			/* No command line */
		return SUCCESS;

	kmemcopy(i, commandLine, MAX_COMMAND_LEN);

	while(*end && k--)
		end++;

	while(sp < end) {
		/* Move along to the next space */
		j = sp;
		/* Move sp along to the next space */
		while((*sp != ' ') && *sp) sp++;

		/* We now change the space to a null so the following work OK */
		*(sp++) = 0;
		if(kbeginswith("mem=", j))
			thisBox.memCnt = atoi(j+4, 10) * 1024;
	}
	return SUCCESS;
}

/* Inits some values, copies the data we got passed from the loader */
initFunc __preInitFunc initInit()
{
	__uint64 bytes = 0;		/* Memory size in bytes */
	ulong memSize = 0;
	char *cmdline = NULL;
	ulong i;

	if(mbootInfo) {			/* Loaded by a multiboot loader */
		printk(KINFO, "Loaded by a multiboot-compatable bootloader");
		/* Print the bootloader name if we got it passed */
		if(MBOOT_FLAG(mbootInfo, MBOOT_BNAME))
			printk(KINFO, " '%s'", mbootInfo->boot_loader_name);
		printk(KINFO, "\n");

		/* Doesn't give a sane memory map... */

#if 0
		/* Get our memory config */
		if(MBOOT_FLAG(mbootInfo, MBOOT_MMAP)) {
			/* We have a memory map! Unfortunatly there is a stupid
			 * extra parameter in it that we must remove... *sigh*
			 */
			ulong	len = (ulong)mbootInfo->mmap_length,
				start = (ulong)mbootInfo->mmap_addr,
				pos = 0;

			while(pos < len && mmapLen < MMAP_ENTRIES) {
				struct mboot_memMap *t = (void *)(pos + start);

				/* Copy the entry */
				mmap[mmapLen++] = t->map;

				/* pos now points at the beginning of the next
				 * entry */
				pos += t->size + 4;
			}
		}
#endif /* 0 */

		/* If we didn't manage to get an mmap for some reason: */
		if(mmapLen == 0 && MBOOT_FLAG(mbootInfo, MBOOT_MEM))
			/* We have a number specifying the number of kb of upper
			 * memory we have */
			memSize = (ulong)mbootInfo->mem_upper;

		/* Get a command-line string, if we have one */
		if(MBOOT_FLAG(mbootInfo, MBOOT_CMD))
			cmdline = mbootInfo->cmdline;
	} else {			/* Loaded by our loader */
		if(bootInfo->map) {	/* Memory map! */
			mmapLen = bootInfo->mapLen / sizeof(struct memMap);
			kmemcopy((char *)mmap, (char *)bootInfo->map,
					mmapLen * sizeof(struct memMap));
		} else			/* Upper memory count */
			memSize = bootInfo->memSize;

		cmdline = bootInfo->cmdline;
	}

	if(!mmapLen) {			/* No memory map */
		if(!memSize)		/* Sanity check here.. */
			panic("No memory info found!\n");

		mmap[0].base = 0x0;		// First 640Kb
		mmap[0].len = 0xA0000;
		mmap[0].type = E820_NORMAL;
		mmap[1].base = 0xA0000;	// MMIO area
		mmap[1].len = 0x60000;
		mmap[1].type = E820_RESERVED;
		mmap[2].base = 0x100000;	// The rest of memory
		mmap[2].len = memSize * 1024;
		mmap[2].type = E820_NORMAL;
		mmapLen = 3;
	}

	/* Cycle thru the list now, looking at the sizes and adding them
	 * togeather so we get a memory size
	 */
	for(i = 0; i < mmapLen; i++) {
#define PRINT_MMAP
#ifdef PRINT_MMAP
		printk(KINFO, "%16Lx -> %16Lx, type %lu\n",
				(ulong long)mmap[i].base,
				(ulong long)(mmap[i].base + mmap[i].len),
				mmap[i].type, 0, 0, 0);
#endif /* PRINT_MMAP */
		bytes += mmap[i].len;
	}

	/* Memory starts from 0, so adjust by 1 for the purpose of.. well..
	 * self-satisfaction ;)
	 */
	thisBox.memCnt = ++bytes / 1024;		// memory size in Kb

	// Copy the command line now
	if(cmdline)
		kmemcopy(commandLine, cmdline, MAX_COMMAND_LEN);

	parseLilo();

	return SUCCESS;
}

/* The *real* entry point of the kernel, just has to test if we were loaded by a
 * multi-boot loader, or ours... */
asm (
	".globl _start						\n"
	"_start:						\n"
#ifdef MULTIBOOT
	/* Booted by a multiboot-compatable bootloader? */
	"	cmpl $" STR(MULTIBOOT_BOOT_MAGIC) ", %eax	\n"
	"	jne 1f						\n"
	/* Store pointer to multiboot struct */
	"	movl %ebx, mbootInfo				\n"
	"	jmp 2f						\n"
#endif /* MULTIBOOT */
	/* Store the struct passed by our loader :) */
	"1:	movl %ebx, bootInfo				\n"
	/* Need to set up a stack for ourselves now... */
	"2:	movl $(stack + " STR(STACK_SIZE) "), %esp	\n"
	/* And jump to the real start */
	"	jmp startPoint					"
);
