/*
 * pci.c - i386 specific PCI stuff
 *
 * Added to the Zealos project and (C) by Zwane Mwaikambo
 * Parts (C) by Mark Zealey <mark@itsolve.co.uk>
 *
 * Current maintainer: Zwane Mwaikambo <zwane@commfireservices.com>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 * Most of this code could really go generic...
 * There must be a better way of listing the devices, but a 65535 entry struct
 *	doesn't really excite me much... - Mark
 *
 * Changelog:
 * 10/09/01 Zwane: HIGHLY experimental code, none of it has been tested!!
 * 21/09/01 Zwane: Ported some stuff over from a mish mash of code i found
 *	on the 'net, still struggling to find docs :( Still developing...
 * 29/09/01 Mark: Fixed so it detects my PCI stuff, more changes to come I hope
 * 08/10/01 Zwane: Started PCI Bios access code.
 * 05/11/01 Mark: Added subsystems code
 */

#include <common.h>
#include <pci.h>
#include <machine.h>
#include <spinlock.h>
#include <cpu.h>
#include <io.h>
#include <printk.h>
#include <kmalloc.h>
#include <stdio.h>
#include <subsystems.h>
#include <debug.h>


void PCICleanup(void);

PCIData_t *PCIData;	/* global PCIData struct */
spinlock_t PCILock = INIT_UNLOCKED;

/* Was gonna call this PCI_KEEP_LUSER_HAPPY but thought that that would be a bit
 * luser-unfriendly, anyway, we don't have any lusers yet... - Mark */
#ifdef PCI_PRETTY_STATS
/* XXX this actually belongs in generic pci.c */
typedef struct {
	int vendorid;
	char *name;
} PCIVendorID_t;


const PCIVendorID_t PCIVendors[] =   {
	{PCI_VENDOR_ID_INTEL,	"Intel Corporation"},
	{PCI_VENDOR_ID_REALTEK, "Realtek Semiconductor"},
	{PCI_VENDOR_ID_SIS,	"Silicon Integrated Systems"}
};

#define NUM_VENDORS	(sizeof(PCIVendors)/sizeof(PCIVendorID_t))
#endif /* PCI_PRETTY_STATS */

// Magic.. :/
#define PCI_READ_DATA(a,b,c)						\
__uint##b PCIReadData##c(__uint8 bus, __uint8 dev, __uint8 func, __uint8 reg) \
{									\
	PCIConfigAddr_t p = {						\
		enable:		1,					\
		reserved:	0,					\
		bus:		bus,					\
		dev:		dev,					\
		func:		func,					\
		reg:		reg & 0xfc				\
	};								\
									\
	outl(&p, PCI_CONF_ADDR);					\
	return in##a(PCI_CONF_DATA  + (reg & 0x3));			\
}

PCI_READ_DATA(b,8,B);
PCI_READ_DATA(w,16,W);
PCI_READ_DATA(l,32,L);


const char *PCIGetVendorName(__uint16 id)
{
#ifdef PCI_PRETTY_STATS
	__uint16 i;

	for (i = 0; i < NUM_VENDORS; i++)
		if (PCIVendors[i].vendorid == id)
			return PCIVendors[i].name;

	return "Vendor not recognised";
#else /* PCI_PRETTY_STATS */
	return "";
#endif /* PCI_PRETTY_STATS */
}

void PCIDumpConfig(PCIConfig_t *cfg, __uint8 bus, __uint8 dev)
{
	printk(KINFO, "PCI: Config @ 0x%p, device %x:%x\n", cfg, bus, dev);

	printk(KINFO, "vendor:device.revision id: %x:%x.%x (%s)\n",
			cfg->vendor_id, cfg->device_id, cfg->rev_id,
			PCIGetVendorName(cfg->vendor_id));
	printk(KINFO, "class: %x:%x\n", cfg->base_class, cfg->sub_class);
	printk(KINFO, "command %x, status %x, interface %x\n", cfg->cmd,
			cfg->status, cfg->interface);
	printk(KINFO, "cache line size %x, latency timer %x, header type %x, "
			"bist %x\n", cfg->cache_line_size,
			cfg->latency_timer, cfg->header_type, cfg->bist);
/*	if((cfg->header_type & 0x7F) == 0) {
		printk(KINFO, "cardbus pointer: %u, subsystem info: %u:%u\n",
				cfg->cardbus_ptr, cfg->subsystem_vendor_id,
				cfg->subsystem_id);
		printk(KINFO, "expansion rom pointer: %u, cap pointer: %u\n",
				cfg->expansion_rom_ptr, cfg->capabilities_ptr);
		printk(KINFO, "irq line: %u, irq pin: %u, max gnt: %u, "
				"max lat: %u\n", cfg->irq_line, cfg->irq_pin,
				cfg->max_gnt, cfg->max_lat);
	}
*/
	printk(KINFO, "\n");
}

/* probes for the BIOS32 entry point by checking for the magic
 * in memory range 0e00000h - 0100000h (paragraph aligned)
 */
bios32_t *bios32Probe(void)
{
	__uint32 length, i;
	__uint32 size = sizeof(bios32_t);	// should be 16 right now
	uchar csum;

	bios32_t *p = (bios32_t*)BIOS32_BEGIN_SEARCH;

	/* Dave is gonna hate me for all these continues ;) */
	for (;p <= (bios32_t*)BIOS32_END_SEARCH; p++) {
		if (p->signature != BIOS32_MAGIC)
			continue;

		length = p->length * size;
		DPRINT("PCI Signature found at 0x%p, length %lu\n", p, length);

		if (!length) {
			DPRINT("Length invalid\n");
			continue;
		}

		/* calculate the checksum, which should equal 0 */
		for (i=0, csum=0; i < length; i++)
			csum += ((uchar *)p)[i];

		if (csum) {
			DPRINT("Invalid checksum %u\n", csum);
			continue;
		}

		/* we got this far, so lets check the version */
		if (p->version != 0) {
			DPRINT("Invalid version %u\n", p->version);
			continue;
		}

		/* very cool, lets point our argument to what we found and bail
		 */
		return p;
	}

	DPRINT("No PCI signature detected\n");

	/* if we got this far, it aint lookin good */
	return NULL;
}

/* Returns the entry point of the BIOS32 service required, NULL on error.
 * registers should contain the following upon entry.
 * EBX = Address
 * ECX = Length
 * EDX = Entry
 * and the result returned in AL
 */
__uint32 bios32Service(__uint32 service)
{
	__uint32 eax, addr, length, entry;
	ulong flags;

	if (!PCIData->bios32Entry.offset)
		return 0;

	spinlockIRQSave(&PCILock, &flags);

	asm volatile ("lcall (%%edi)"
		: "=a" (eax), "=b" (addr), "=c" (length), "=d" (entry)
		: "0" (service), "1" (0), "D" (&PCIData->bios32Entry));

	spinlockIRQRestore(&PCILock, &flags);

	/* 0 for success, 0x80 for unsupported and lord knows for the rest ;) */
	return ((eax & 0xFF) ? 0 : (addr + entry));
}

/* Ported from Linux 2.4.9 (c) 1999--2000 Martin Mares <mj@suse.cz>
 * -Zwane
 * Upon return:
 * AH = status, AL = PCI access type (Direct (config 1/2) or BIOS)
 * BH = PCI Bios major revision, BL = minor revision
 * CL = last bus on the system
 * EDX = PCI signature "PCI "
*/
errRet PCIBiosCheck (void)
{
	__uint32 sig, eax, ebx, ecx, flags;
	__uint8 status;

	if(!(PCIData->PCIBiosEntry = bios32Service(PCI_SERVICE)))
		return ENOEXIST;

	PCIData->PCIBiosIndirect.offset = PCIData->PCIBiosEntry;
	spinlockIRQSave(&PCILock, &flags);

	asm(	"	lcall (%%edi)		\n"	/* Call the func */
		"	cld			\n"	/* Keep gcc happy */
		"	jc 1f			\n"
		/* If we didn't err, clear ah */
		"	xor %%ah, %%ah		\n"
		"1:				"
		: "=d" (sig), "=a" (eax), "=b" (ebx), "=c" (ecx)
		: "1" (PCI_PCIBIOS_PRESENT), "D" (&PCIData->PCIBiosIndirect));

	spinlockIRQRestore(&PCILock, &flags);

	status = (eax >> 8) & 0xFF;
	if (status || (sig != PCI_SIGNATURE)) {
		DPRINT("BUG Check: sig=%d status=%x\n");
		return ENOEXIST;
	}

	PCIData->access_type = eax & 0xFF;
	PCIData->major_ver = (ebx >> 8) & 0xFF;
	PCIData->minor_ver = ebx & 0xFF;
	PCIData->last_bus = (ecx & 0xFF);

	return SUCCESS;
}

void PCIReadBlock(__uint8 bus, __uint8 dev, __uint8 func, void *data, uint size)
{
	uint i;

	/* First read in 4-byte blocks */
	for(i=0; size >= sizeof(long); i += sizeof(long), size -= sizeof(long))
		*(ulong *)((char *)data + i) = PCIReadDataL(bus, dev, func, i);

	/* Then read the rest in in single-byte blocks */
	for(; size; size--, i++)
		*((char *)data + i) = (char)PCIReadDataB(bus, dev, func, i);
}

__uint8 PCIScanBus(__uint8 bus)
{
	__uint8 dev, found = 0;

	DPRINT("PCI: Scanning bus %u\n");

	for(dev = 0; dev < 32; dev++) {	/* hmm... max device? */
		PCIConfig_t c;

		PCIReadBlock(bus, dev, 0, &c, sizeof(c));

		if(c.vendor_id == PCI_VENDOR_ID_INVALID)
			continue;

		found++;

		/* Add the device to a tree here or something... */

		/* we found something, lets dump it */
		PCIDumpConfig(&c, bus, dev);
	}

	return found;
}

/* All malloced mem and general stuff you mucked with in PCI.C should
 * be reversed here and set to sane values.
 */
void PCICleanup()
{
	ulong flags;

	DPRINT("PCI: Subsystem shutting down...");

	/* lets spin on the lock and only go ahead when its free
	 * mind you the offending executing code could also have
	 * just released it temporarily and will find itself in
	 * NULL pointer land when it tries to get anywhere next
	 * time. But that should never happen if the function is
	 * coded properly!!
	 */

	spinlockIRQSave(&PCILock, &flags);	//disable IRQs too

	SUBSYSTEM_OFF("PCI");

	kfree(PCIData);
	PCIData = NULL;

	spinlockIRQRestore(&PCILock, &flags);

	DPRINT("Done\n");
}

errRet __preInitFunc PCIInit(void)
{
	int i;
	PCIData = kmalloc(sizeof(PCIData_t));

	DPRINT("PCI: Initialising subsystem...\n");

	if (!PCIData) {
		DPRINT("PCI: Unable to initialise! (OOM)\n");
		return ENOMEM;
	}

	PCIData->bios32Entry.offset = 0;
	PCIData->bios32Entry.seg = kernCode;
	PCIData->PCIBiosIndirect.offset = 0;
	PCIData->PCIBiosIndirect.seg = kernCode;

	DPRINT("PCI: Probing for PCI BIOS...\n");

	PCIData->bios32 = bios32Probe();
	if(!PCIData->bios32) {
		PCICleanup();
		return SUCCESS;		/* No error in not having PCI... */
	}

	PCIData->bios32Entry.offset = PCIData->bios32->entry;

	printk(KINFO, "PCI: BIOS32 Service Directory entry at 0x%lx\n",
			PCIData->bios32->entry);

	/* Now lets check for a PCI Bios, Note that you can have bios32 and no
	 * PCI */
	if (PCIBiosCheck() != SUCCESS) {
		PCICleanup();		/* Clean up our mess */
		return SUCCESS;
	}

	printk(KINFO, "PCI Bios Rev %x.%02x entry at 0x%lx last bus = %d "
			"access type = %x\n", PCIData->major_ver,
			PCIData->minor_ver, PCIData->PCIBiosEntry,
			PCIData->last_bus, PCIData->access_type);

	for (i = 0; i <= PCIData->last_bus; i++)
		PCIScanBus(i);

	SUBSYSTEM_ON("PCI", PCIData, PCIInit, PCICleanup);
	return SUCCESS;
}
