/*
 * smp.c - MP 1.4 implementation
 *
 * Added to the Zealos project and (C) by Zwane Mwaikambo
 *
 * Current maintainer: Zwane Mwaikambo <zwane@commfireservices.com>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * MP 1.4 implementation based on Intel specs
 *
 * Notes/Fixme's:
 * This code seriously suffers from the 80 char limit problem, i know we're not
 * all running high res but cmon, the code is next to unreadable and line breaks
 * _everywhere_ -Zwane
 * Changelog:
 * 30/09/01	Zwane: added the structures
 * 01/10/01	Zwane: added MP 1.x detection code.. in development
 * 24/10/01	Zwane: few fixups
 * 19/11/01	Zwane: Round 2 of the Lord-please-let-this-work fixes ;)
 *
 */
#include <common.h>
#include <smp.h>
#include <machine.h>
#include <cpu.h>
#include <printk.h>
#include <kmalloc.h>
#include <stdio.h>
#include <debug.h>


SMPData_t *SMPData = NULL;

/* do a search from memory range start to end for MP floating point structure
 * note that it is paragraph aligned
*/
MPFloatingPS_t* performSearch(__uint32 start, __uint32 end)
{
	MPFloatingPS_t *a = (MPFloatingPS_t *) start;
	int size = sizeof(MPFloatingPS_t);
	int len, i;
	uchar csum = 0;

	/* end - size so we don't fall of the edge of the earth */
	for (;a <= (MPFloatingPS_t*)(end - size); a++) {
		if (a->signature != MP_MAGIC)
			continue;

		DPRINT("MP Signature @ 0x%p\n", a);

		/* the length field contains the size of the structure in
		 * paragraphs so it should be 1 therefore (size * a->length) =
		 * 16bytes
		 */
		len = a->length * size;
		DPRINT("len=%d\n", len);

		if (!len) {
			DPRINT("len is 0 (a = %p)!\n", a);
			continue;
		}

		for (i=0; i<len; i++)
			csum += *((uchar *)a + i);

		/* checksum should be zero */
		if (csum) {
			/* Why does my Compaq 3000 box always report 16??? */
			DPRINT("csum is %u, should be 0 (a = %p)!\n", csum, a);
			continue;
		}

		/* we only support MP 1.1 and 1.4 boxes */
		if ((a->revision != 1) && (a->revision != 4)) {
			DPRINT("revision is %u (unsupported) (a = %p)!\n",
					a->revision, a);
			continue;
		}

		/* we got this far, return the pointer */
		return a;
	}

	/* if we get here we didn't find anything */
	return NULL;
}

/* Probe for the MP Floating pointer structure by searching the following memory
 * ranges...  the following comment is copyright Intel (intel MP 1.4 specs pg
 * 38) a. In the first kilobyte of Extended BIOS Data Area (EBDA), or b. Within
 * the last kilobyte of system base memory (e.g., 639K-640K for systems with 640
 * KB of base memory or 511K-512K for systems with 512 KB of base memory) if the
 * EBDA segment is undefined, or c. In the BIOS ROM address space between
 * 0F0000h and 0FFFFFh.
 */
MPFloatingPS_t *probeMPFloatingPS()
{
	MPFloatingPS_t *mem = NULL;

	/* SMP with 512kb memory. Pull the other leg. */
	if ((mem = performSearch((MEM_BASE_END_512 - 0x400), MEM_BASE_END_512)))
		goto end_probe;

	if ((mem = performSearch((MEM_BASE_END - 0x400), MEM_BASE_END)))
		goto end_probe;

	/* try the ROM area */
	if ((mem = performSearch(MEM_ROM_START, MEM_ROM_END)))
		goto end_probe;

	mem = performSearch(MEM_EBDA_START, MEM_EBDA_START + 0x400);

end_probe:
	return mem;
}

/* This function walks through the extended entries and sets the appropriate
 * flags, pointers in the SMPData struct to reflect what was found
*/
errRet parseConfEntries(MPConfTable_t *m)
{
	__uint16 num, size;
	uchar *ptr;		/* pointer to start of entries */

	if (!m)
		return ENOEXIST;
	
	size = m->base_length;
	num = m->entry_count;
	ptr = uchar_p(m);

	/* Do the base entries first */
	while (num--) {
		switch (*ptr) {
		case ENTRY_CPU:
			/* only add it if it's enabled (the en flag is 24 bits
			 * from the base) */
			if (*(ptr + 3) & SMP_CPU_EN) {
				/*
				SMPData->CPUEntry = krealloc(SMPData->CPUEntry,
				(SMPData->num_aps + 1) * sizeof(CPUEntry_t*));

				if (!SMPData->CPUEntry[SMPData->num_aps])
					return ENOMEM;
				*/
				if (SMPData->num_aps >= MAX_APS)
					DPRINT("SMP: Max APs Exceeded!\n");
				else {
					SMPData->CPUEntry[SMPData->num_aps] = (CPUEntry_t *)ptr;
					SMPData->num_aps++;
				}
				ptr += sizeof(CPUEntry_t);
			}
		break;

		case ENTRY_BUS:
			/*
			SMPData->BUSEntry = krealloc(SMPData->BUSEntry,
				(SMPData->num_buses + 1) * sizeof(BUSEntry_t*));

			if (!SMPData->BUSEntry[SMPData->num_buses])
				return ENOMEM;
			*/
			if (SMPData->num_buses >= MAX_BUSES)
				DPRINT("SMP: Max Buses Exceeded!\n");
			else {
				SMPData->BUSEntry[SMPData->num_buses] = (BUSEntry_t *)ptr;
				SMPData->num_buses++;
			}

			ptr += sizeof(BUSEntry_t);
		break;

		case ENTRY_IOAPIC:
			/* bitfield 24 specifies wether the APIC is enabled or not */
			if ((*ptr + 3) & SMP_IOAPIC_EN) {
				/*
				SMPData->BUSEntry = krealloc(SMPData->IOAPICEntry,
				(SMPData->num_ioapics + 1) * sizeof(IOAPICEntry_t*));

				if (!SMPData->IOAPICEntry[SMPData->num_ioapics])
					return ENOMEM;
				*/
				if (SMPData->num_ioapics >= MAX_IOAPICS)
					DPRINT("SMP: Max IOAPICs Exceeded!\n");
				else {
					SMPData->IOAPICEntry[SMPData->num_ioapics] = (IOAPICEntry_t *)ptr;
					SMPData->num_ioapics++;
				}

				ptr += sizeof(IOAPICEntry_t);
			}
		break;
			/* other entries ... */
		default:
			DPRINT("SMP: Unknown MP Configuration entry\n");
		break;
		}
	}

	/* if we have extended entries this is non zero */
	if (!m->extended_length)
		return SUCCESS;

	ptr = uchar_p(m) + size;	/* m + the base length */
	size = m->extended_length;

	/* parse the extended entries ... */
	
	return SUCCESS;
}

void SMPCleanup()
{
	DPRINT("Shutting down SMP.\n");
	SUBSYSTEM_OFF("SMP");

	/* try to free as much as possible */
	kfree(SMPData->CPUEntry);
	kfree(SMPData->BUSEntry);
	kfree(SMPData->IOAPICEntry);

	kfree(SMPData);
}

errRet __preInitFunc SMPInit()
{
	int i;
	SMPData = kmalloc(sizeof(SMPData_t));

	if (!SMPData) {
		DPRINT("Error initialising SMPData (OOM)\n");
		return ENOMEM;
	}

	SMPData->MPConfTable = NULL;
	SMPData->MPFloatingPS = NULL;
	SMPData->CPUEntry = NULL;
	SMPData->BUSEntry = NULL;
	SMPData->IOAPICEntry = NULL;

	SMPData->num_aps = 0;		/* 0 means no APs and one BSP */
	SMPData->num_buses = 0;		/* 0 means 1 bus */
	SMPData->num_ioapics = 0;

	printk(KINFO, "SMP: Probing for MP 1.x Entry...\n");
	SMPData->MPFloatingPS = probeMPFloatingPS();
	if (!SMPData->MPFloatingPS) {
		DPRINT("SMP: Unable to locate MP Floating Pointer Structure\n");
		SMPCleanup();
		return SUCCESS;		/* No error in running smp on a up box*/
	}

	printk(KINFO, "SMP: Intel MultiProcessor Specification 1.%d @ 0x%lx\n",
			SMPData->MPFloatingPS->revision, SMPData->MPFloatingPS);

	/* if info1 == 0 then there is a MP Configuration table otherwise it
	 * has the number of the predefined configuration
	 */
	switch (SMPData->MPFloatingPS->info1) {
	case 0x00:	/* we have a conf table, point to it */
		SMPData->MPConfTable = (void *) SMPData->MPFloatingPS->address;
		DPRINT("SMP: MP Conf Table @ 0x%lx\n", SMPData->MPConfTable);
		/* Let's just double check */
		if (SMPData->MPConfTable->signature != MP_CONF_MAGIC) {
			DPRINT("SMP: Invalid MP Configuration table magic!\n");
			SMPCleanup();		/* fatal */
			return SUCCESS;
		}

		DPRINT("SMP: MP 1.%x Configuration Table @ 0x%lx\n",
				SMPData->MPConfTable->spec_rev,
				SMPData->MPConfTable);

		break;

	/* i'm not quite sure how to handle some of these */
	case 0x01:
	case 0x02:
	case 0x03:
	case 0x04:
		SMPData->num_buses = 0;
		SMPData->num_aps = 1;
		break;

	case 0x05:
	case 0x06:
	case 0x07:
		SMPData->num_buses = 1;
		SMPData->num_aps = 1;
		break;

	default:
		DPRINT("SMP: Unknown MP default configuration.\n");
		SMPCleanup();
		return SUCCESS;
	}

	if (SMPData->MPConfTable) {
		printk(KINFO, "SMP: %u Base entries (%u bytes) Starting @ "
				"0x%lx\n", SMPData->MPConfTable->entry_count,
				SMPData->MPConfTable->base_length,
				SMPData->MPConfTable);

		/* start enumerating extended entries, note that they are sorted
		 * in ascending order wrt type i.e. cpus first, then bus...
		*/
		if (parseConfEntries(SMPData->MPConfTable) != SUCCESS) {
			DPRINT("SMP: Error parsing entries!\n");

			SMPCleanup();		/* this really is fatal */
			return SUCCESS;
		}

		/* we're going to have to send these off to CPUDetect, when we
		 * consolidate this code though */
		printk(KINFO, "SMP: Detected the following CPUs\n");
		for (i = 0; i <= SMPData->num_aps; i++)
			printk(KINFO, "SMP: %s CPU[%lu] APIC (v%u) = %lu Family"
					" %u Model %u Stepping %u Caps = "
					"0x%x\n",
				(SMPData->CPUEntry[i]->flags & SMP_CPU_BSP) ? "BSP" : "AP",
				i, SMPData->CPUEntry[i]->apic_ver,
				SMPData->CPUEntry[i]->apic_id,
				SMPData->CPUEntry[i]->family,
				SMPData->CPUEntry[i]->model,
				SMPData->CPUEntry[i]->stepping,
				SMPData->CPUEntry[i]->capability);

		printk(KINFO, "SMP: Detected the following system buses\n");

		for (i = 0; i <= SMPData->num_buses; i++) {
			printk(KINFO, "SMP: Bus%u id%u type %c%c%c%c%c%c\n", i,
				SMPData->BUSEntry[i]->bus_id,
				SMPData->BUSEntry[i]->bus_type_str[0]
				SMPData->BUSEntry[i]->bus_type_str[1]
				SMPData->BUSEntry[i]->bus_type_str[2]
				SMPData->BUSEntry[i]->bus_type_str[3]
				SMPData->BUSEntry[i]->bus_type_str[4]
				SMPData->BUSEntry[i]->bus_type_str[5]);
		}

		printk(KINFO, "SMP: Detected the following I/O APICs\n");
		for (i = 0; i <= SMPData->num_ioapics; i++)
			printk(KINFO, "SMP: IO/APIC ver. %u id = %u @ 0x%lx\n",
				SMPData->IOAPICEntry[i]->ioapic_ver,
				SMPData->IOAPICEntry[i]->ioapic_id,
				SMPData->IOAPICEntry[i]->ioapic_addr);
	}
	
	SUBSYSTEM_ON("SMP", SMPData, SMPInit, SMPCleanup);
	return SUCCESS;
}
