/*
 * page.c - Paging implimentation for the i386
 * 
 * Added to the Zealos project and (C) by Mark Zealey
 * 
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * Description:
 * 
 * Notes/Fixme's:
 * 
 * Changelog:
 */


#include <common.h>
#include <page.h>
#include <kmalloc.h>
#include <mem.h>
#include <init.h>
#include <printk.h>
#include <assert.h>
#include <stdio.h>
#include <machine.h>

pageDir basePageDir;

// Any better place for this function to go?
/* function: uchar pageZone(memory page)
 * desc: Returns the zone that the specified page is in
 */
uchar pageZone(memory page)
{
	if(page < 16 * 1024 * 1024 / PAGE_SIZE)
		return AREA_DMA;

	return AREA_NORMAL;
}

// Any better place for this function to go?
/* function: bool isMemory(ulong address)
 * desc: Returns TRUE if the specified address is in memory physical memory that
 * has no 'strings' attached, FALSE otherwise
 */
bool isMemory(ulong address)
{
	uint t = 0;

	/* This guards against b0rken biosen reporting that 640K..1Mb is normal
	 * memory, when we know it isn't.
	 */
	if((address > 640 * 1024) && (address < 1024 * 1024))
		return FALSE;

	for(t = 0; t < mmapLen; t++) {
		if(mmap[t].type != E820_NORMAL)
			continue;

		if((address > mmap[t].base)
				&& (address - mmap[t].base < mmap[t].len))
			return TRUE;
	}

	return FALSE;
}

/* function: struct pgTblEnt *pageLookup(pageDir *tbl, memory page)
 * desc: Returns the entry from the page table that's referenced by the page
 * table (virtually). For example, if you pass it page 0x1, it will return
 * the second entry in the first page table pointed to by 'tbl'. If the page was
 * out of bounds, for example, it will return NULL.
 */
inline struct pgTblEnt *pageLookup(pageDir *tbl, memory page)
{
	struct pgTblEnt *t;
	ulong dirPos, tblPos;

	if(tbl == NULL)
		tbl = &basePageDir;

	dirPos = page / (tbl->pageTblSize / sizeof(struct pgTblEnt));
	if(dirPos > tbl->pageTblSize)
		return NULL;

	tblPos = page % (tbl->pageTblSize / sizeof(struct pgTblEnt));
	t = (void *)PAGE_ADDRESS(tbl->dir[dirPos].entry);

	DPRINT("Found page %lu in table %lu (%p). Pos in table: %lu\n",
			page, dirPos, t, tblPos);

	return t + tblPos;
}

/* function: errRet pageAdd(pageDir *tbl, memory page, ulong options)
 * desc: Adds 'options' to the specified page in the page-table.
 */
errRet pageAdd(pageDir *tbl, memory page, ulong options)
{
	struct pgTblEnt *pg = pageLookup(tbl, page);

	if(!pg)
		return ENOEXIST;

	pg->entry |= options;
	return SUCCESS;
}

/* function: errRet pageDel(pageDir *tbl, memory page, ulong options)
 * desc: Removes 'options' to the specified page in the page-table.
 */
errRet pageDel(pageDir *tbl, memory page, ulong options)
{
	struct pgTblEnt *pg = pageLookup(tbl, page);

	if(!pg)
		return ENOEXIST;

	pg->entry &= ~options;
	return SUCCESS;
}

/* function: errRet pageSet(pageDir *tbl, memory page, ulong options)
 * desc: Sets 'options' to the specified page in the page-table. If options
 * contains a new page to map to, it will set 'page' to point at that.
 */
errRet pageSet(pageDir *tbl, memory page, ulong options)
{
	struct pgTblEnt *pg = pageLookup(tbl, page);

	if(!pg)
		return ENOEXIST;

	/* If the options passed to us contain a page address, we'll map that on
	 * for them, otherwise we map it physically - Mark
	 */
	if(PAGE_ADDRESS(options))
		pg->entry = options;
	else
		pg->entry = PAGE_ADDRESS(page) | options;

	return SUCCESS;
}

/* function: long pageGetBlock(pageDir *tbl, ulong length, uint opts)
 * desc: Returns an offset in the page table that is large enough to hold
 * 'length' entries in. If no space large enough is found, return -1. If tbl is
 * NULL, it uses the kernel's page table.
 */
long pageGetBlock(pageDir *tbl, ulong length, uint opts)
{
	// FIXME: Maybe we could use a bitmap for this? But it would be slow..
	uint i;

	assert(length);

	if(tbl == NULL)
		tbl = &basePageDir;

	if(length >= tbl->pageTblSize)
		return -1;

	for(i = 0; i < tbl->pageDirSize; i++)
		if(!tbl->dir[i].entry) {
			tbl->dir[i].entry = (memory)kmallocZeroPage() | opts
								| PAGE_PRESENT;
			return i * (tbl->pageTblSize / sizeof(struct pgTblEnt));
		}

	return -1;
}

/* function: errRet pageFreeBlock(pageDir *tbl, ulong length, long pos)
 * desc: Frees a set of page tables previously allocated with pageGetBlock(). If
 * tbl is NULL, it will sue the main kernel page table.
 */
errRet pageFreeBlock(pageDir *tbl, ulong length, long pos)
{
	if(tbl == NULL)
		tbl = &basePageDir;

	assert(length);
	assert(pos > -1);

	pos /= (tbl->pageTblSize / sizeof(struct pgTblEnt));

	kfreePage((void *)PAGE_ADDRESS(tbl->dir[pos].entry));

	tbl->dir[pos].entry = 0;

	return SUCCESS;
}

errRet __preInitFunc pageTableInit()
{
	struct pgTblEnt *curPage;
	struct pgDirEnt *dir;
	ulong curPage_p = 0;
	ulong num = 0;
	ulong dir_p = 0;

	/* We should now make space for our paging tables. If there is not
	 * enough space between the end of the kernel and the 640K memory area,
	 * we take some of the stuff over the 640K memory area. We only map what
	 * we need to access all of the physical memory, the rest we zero so
	 * that we will get a nice page-fault if that happens.
	 */

	// Get 1 page for the page dir
	dir = (void *)(PAGE_ADDRESS(kernelStart + kernelLength) + PAGE_SIZE);

	/* Search through for some memory that we can use for this page. It must
	 * be page-aligned.
	 */
	while(!isMemory((memory)dir))
		dir = (void *)((memory)dir + PAGE_SIZE);

	basePageDir.dir = dir;
	basePageDir.pageDirSize = PAGE_SIZE;
	basePageDir.pageTblSize = PAGE_SIZE;

	// Same here
	curPage = (void *)((memory)dir + PAGE_SIZE);
	while(!isMemory((memory)curPage))
		curPage = (void *)((memory)curPage + PAGE_SIZE);

	for(; num < (thisBox.memCnt * 1024); num += PAGE_SIZE) {
		/* If we're at the beginning of a new page, add the entry to
		 * the paging directory */
		if(!curPage_p)
			dir[dir_p++].entry = (memory)curPage | PAGE_PRESENT |
							PAGE_RW | PAGE_KERN;

		curPage[curPage_p].entry = num | PAGE_KERN | PAGE_RW |
								PAGE_PRESENT;

		/* We've reached the end of this page, get new page and reset
		 * curPage_p.
		 */
		if(++curPage_p == PAGE_SIZE / sizeof(struct pgTblEnt)) {
			/* This does checks against the E820 map so we make sure
			 * that we don't overwrite some stuff in the 640K+
			 * memory area. - Mark
			 */
			do
				curPage = (void *)((memory)curPage + PAGE_SIZE);
			while(!isMemory((memory)curPage));
			curPage_p = 0;
		}
	}

	/* We blank out the rest of this page */
	if(curPage_p)
		while(1) {
			curPage[curPage_p++].entry = num | PAGE_SWAPPED |
							PAGE_RO | PAGE_KERN;
			if(curPage_p == PAGE_SIZE / sizeof(struct pgTblEnt))
				break;
			num += PAGE_SIZE;
		}

	/* And blank out the rest of the page dir */
	while(dir_p < PAGE_SIZE / sizeof(struct pgTblEnt))
		dir[dir_p++].entry = 0;

	/* Update the kernel size to include all these tables we've just done */
	kernelLength = (memory)curPage + PAGE_SIZE - kernelStart;

	/* We make the first page null so that we catch null pointer
	 * dereferences and the like */
	curPage = (void *)PAGE_ADDRESS(dir[0].entry);
	curPage[0].entry = 0 | PAGE_SWAPPED | PAGE_RO | PAGE_KERN;

	/* Load the new paging directory. */
	setCr(3, dir);

#ifdef MULTIBOOT
	do {
		ulong cr0;

		/* Enable paging if we were multibooting, as bootloader might
		 * not have initalized it... */
		getCr(0, cr0);
		setCr(0, cr0 | (1 << CR0_PG));
	} while(0);
#endif /* MULTIBOOT */

	return SUCCESS;
}

EXPORT_SYMBOL(isMemory);
