/*
 * common.h - Everything that needs to be included by default in every file
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 01/11/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * So, Add more typedef's to this file. Also add structs/unions that are used
 *	right thru the module loader.
 * Also, #define's common to Zealos should either go here, or in the -D flag of
 *	GCC when compiling.
 * N.B. Anything common to modules DOES NOT GO HERE! ;-)
 * -- Mark 21/07/2000
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 14/05/2001 mark: Added some more defines here (notably CHAR_BIT) and also
 *	made it refuse to compile on anything < GCC 2.7
 * 31/08/01 Mark: Removed support for non-gcc compilers, wouldn't have worked
 *	anyway. Also rewrote MIN/MAX to do type-checking etc.
 * 28/09/01 Mark: Added BCD_TO_BIN()
 * 21/10/01 Mark: Added LABEL() and related functions.
 */
#ifndef __COMMON_H
#define __COMMON_H

/* Refuse to compile on < 2.70 */
#if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 70)
#error Old version of GCC, kernel could be fucked with...
#endif

/* Warn on < 2.90, so we don't get blamed - Mark */
#if __GNUC__ == 2 && __GNUC_MINOR__ < 90
#warning "I'll try to compile, but there might be some bugs because of your gcc"
#endif

/* macro: STR(string)
 * desc: Returns a string'ified version of string (where that is a macro or
 * something)
 */
#define __STR(x) #x		// EUCKIE!! (But it works!! ;)
#define STR(x) __STR(x)

/* function: KILL_COMPILER(error_string)
 * desc: Kills the compiler by producing some invalid assembler. Prints the
 * error message and the line number too.
 */
#define KILL_COMPILER(error) asm(".error_" error "_at_line_" STR(__LINE__) )

/* Function: WRITE_BARRIER()
 * description: Prevents gcc from thinking that some memory address is valid,
 * forcing it to reload it to a register when it's used. Like the keyword
 * 'volatile', but makes gcc invalidate /all/ memory locations.
 */
#define WRITE_BARRIER() asm("":::"memory")

/* Evulness!! - Mark */
#define MINMAX(s,a,b) ({						\
	const typeof (a) _a = (a);					\
	const typeof (b) _b = (b);					\
	(void) (&_a == &_b);						\
	_a s _b ? _a : _b;						\
})

/* function: MIN(a,b)
 * description: Returns the smallest value out of 'a' and 'b'. Type-checking and
 * signedness tests are performed on both variables. See also MAX() and __MIN().
 */
#define MIN(a,b) MINMAX(<,a,b)

/* function: MAX(a,b)
 * description: Returns the largest value out of 'a' and 'b'. Type-checking and
 * signedness tests are performed on both variables. See also __MAX() and MIN().
 */
#define MAX(a,b) MINMAX(>,a,b)

#define __norm_ROUND(a, b) ({						\
	const typeof(a) _a = (a);					\
	const typeof(b) _b = (b);					\
	_a / _b + ((_a % _b)? 1 : 0);					\
})

#define __const_ROUND(a, b) ((a) / (b) + (((a) % (b)) ? 1 : 0))

#define ROUND(a,b) ((__builtin_constant_p(a) &&				\
			__builtin_constant_p(b))?			\
				__const_ROUND(a,b):			\
				__norm_ROUND(a,b)			\
			)

/* Macro: BCD_TO_BIN(num)
 * desc: This converts 'num' from BCD to binary.
 */
#define BCD_TO_BIN(val) ({typeof (val) _v = (val); (_v&15) | (_v>>4)*10;})

/* macro: KERN_VER(major, minor, rel_major, rel_minor)
 * desc: Returns the packed version number of the arguments passed to it
 */
#define KERN_VER(a,b,c,d) (((a) << 24) | ((b) << 16) | ((c) << 8) | (d))

/* macro: CURRENT_KERN_VER
 * desc: This macro returns the current kernel version in a 'packed' format,
 * as returned by KERN_VER().
 */
#define CURRENT_KERN_VER KERN_VER(KERN_MAJOR, KERN_MINOR,		\
				KERN_REL_MAJOR,KERN_REL_MINOR)

/* Macro: PACK(a,b,c,d)
 * desc: This returns characters 'a' to 'd' in a 4-byte number with 'd' as the
 * lowest byte and 'a' as the highest. See also REVERSE_PACK().
 */
#define PACK(a,b,c,d) ((d) | ((c) << 8) | ((b) << 16) | ((a) << 24))

/* Macro: REVERSE_PACK(a,b,c,d)
 * desc: This returns characters 'a' to 'd' in a 4-byte number with 'a' as the
 * lowest byte and 'd' as the highest. See also PACK().
 */
#define REVERSE_PACK(a,b,c,d) ((a) | ((b) << 8) | ((c) << 16) | ((d) << 24))

/* Macro: LABEL_DECL(name)
 * desc: Forward-declairs the label 'name' generated by LABEL().
 */
#define LABEL_DECL(n) extern void __##n()

/* macro: LABEL_ADDR(name)
 * desc: Returns a pointer to the label 'name' created by LABEL().
 */
#define LABEL_ADDR(n) __##n

/* macro: LABEL(name)
 * desc: Creates a label that you can get the address of properly, unlike a C
 * label with the && operator...
 */
#define LABEL(n) asm(".globl __" #n "; __" #n ":")

/* macro: checkArgs(type, format_arg, formated_args_start)
 * arg0: The type of argument style. Currently 'printf', 'scanf' and 'strftime'
 * are supported (but this is just passed to gcc to varify)
 * arg1: The number of the argument which is the format string.
 * arg2: The start of the ... arguments that are to be varified.
 * desc: This macro relies on a GCC __attribute__ feature to work, so it's
 * described as a macro. Basically, it takes the type and the format string and
 * varifys that it has arguments of the right type to prevent nasty things
 * happening during runtime.
 */
#define checkArgs(a,b,c) __attribute__ ((format (a, b, c)))

/* macro: __SECTION__(name)
 * desc: Tells the compiler to put this code/data in a special section called
 * 'name'.
 */
#define __SECTION__(n) __attribute__((section(n)))

/* macro: __preInitFunc
 * desc: This macro tells the compiler to put this function into a special
 * section so that after initilization has been done using this function, it can
 * be free'd by zealos. Only use this for functions that will *never* be used
 * after the initalization of the module
 */
#define __preInitFunc __SECTION__(".text.init")

/* macro: __WEAK__
 * desc: Declaires the function or variable as global, but 'weak', so that other
 * symbols can beat it during linking.
 */
#define __WEAK__ __attribute__((weak))

/* macro: __UNUSED__
 * desc: This tells gcc that the variable is not meant to be used (sometimes
 * useful when you're *only* using it in an assembly routine)
 */
#define __UNUSED__ __attribute__((unused))

/* macro: __ALIGNED__(alignment)
 * desc: Tells the compiler to align this thing to 'alignment' bytes.
 */
#define __ALIGNED__(b) __attribute__((aligned(b)))

/* macro: __PACKED__
 * desc: This macro forces the struct or data it is applied to to be tightly
 * 'packed' togeather. This is useful for low-level system structs, but
 * shouldn't be used on others as aligned access to a struct can vastly
 * speed up access times.
 */
#define __PACKED__ __attribute__((packed))

/* macro: asmLinkage
 * desc: This macro forces the compiler to get the arguments passed to the
 * specified function off the stack. This is useful so that when we are calling
 * a function that is called from an asm routine, because they only pass their
 * arguments over the stack.
 */
#define asmLinkage __attribute__((regparm(0)))
#define fastCall __attribute__((regparm(3)))

/* macro: archBits
 * desc: Returns the number of bits the current arch has (32 on x86 for example)
 */
#define archBits (sizeof(void *) * CHAR_BIT)

// Kludge to stop gcc complaning when it buggers up with the -undef..
#undef __i386__

#include <defines.h>

/* Someone can undefine CHAR_BIT in their arch_common.h file which we include
 * later if it's not 8... - Mark
 */

/* macro: CHAR_BIT
 * desc: The number of bits that a 'char' has.
 */
#define CHAR_BIT 8

/* macro: BYTES(bits)
 * desc: Returns the number of bytes required to fit 'bits' bits in.
 */
/* GCC doesn't like this, even though it recognises that they are both
 * constant....
 * #define BYTES(x) ROUND(x, CHAR_BIT)
 */
#define BYTES(x) ((x) / CHAR_BIT)

typedef unsigned char uchar;
typedef unsigned char bool;		// Bool
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned long ulong;

#define FALSE 0
#define TRUE 1
#define ENABLE 1
#define DISABLE 0
#define ON 1
#define OFF 0
#define NULL ((void *)0)

/* macro: BYTEORDER
 * desc: Set by the arch-dependant code, a value of 0 means zealos is running on
 * a little-endian computer, a value of 1 means we're running on a big-endian
 * one.
 */
#undef BYTEORDER

#endif	// __COMMON_H

#include <arch_common.h>
#include <errors.h>
#include <functions.h>
