/*
 * llist.h - A generic linked list implimentation
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 03/11/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 * Double-linked-list implimentation looks a bit fucked... - Mark
 *
 * Changelog:
 * 23/06/01 silas: i refuse to tell you what i did :P
 * 24/06/01 Mark: I fixed the calling order to be standard
 */

#ifndef __generic_kern_llist_H
#define __generic_kern_llist_H

////////////////////////////////////////////////////////////////////////////////
// Single-lists
////////////////////////////////////////////////////////////////////////////////

/* macro: LIST(type)
 * desc: Makes an entry into your struct 'type' for a linked list.
 * For Example: struct foo { ...; LIST(struct foo); };
 */
#define LIST(type) type *next

/* macro: LIST_INIT(start)
 * desc: Statically initalizes a linked list. See also listInit()
 */
#define LIST_INIT(s) next: NULL //(s)

/* macro: listInit(start)
 * desc: Initalizes the first element of a linked list (must be in
 * function context. For a static-initalizer, please see LIST_INIT)
 */
#define listInit(s) (s)->next = NULL //(s)

/* macro: listAdd(where, element)
 * desc: Adds element into a linked list after the element 'where'.
 */
#define listAdd(b, a) do {						\
	(a)->next = (b)->next;						\
	(b)->next = (a);						\
} while(0)

/* macro: listPrepend(where, element)
 * desc: Adds element into a linked list before the element 'where'.
 */
#define listPrepend(b, a) do {						\
	(a)->next = (b);						\
	(b) = (a);							\
} while(0)

/* macro: listShift(where, element)
 * desc: Removes the first element of the list 'where' and returns it in
 * 'element'.
 */
#define listShift(a, b) do {						\
	(b) = (a);							\
	if(a) (a) = (a)->next;						\
}

/* macro: listItt(start, variable)
 * desc: Itterates over a linked list using the variable 'variable'.
 */
#define listItt(s, t)							\
	for((t) = (s); (t); (t) = (t)->next)

/* macro: listRm(start, element, where)
 * desc: Removes the element 'element' which is after 'where' in the list.
 */
#define listRm(s,e,o) ({						\
		if((o) && (e)) {					\
			(o)->next = (e)->next;				\
		} else if((e) == (s)) {					\
			(s) = (s)->next;				\
		}							\
	})

/* macro: listDel(start, variable, condition)
 * desc: Removes the first element of the linked list 'start' that matches
 * 'condition'. The value in 'variable' is the matching element. This functions
 * DOES NOT kfree() 'variable', just delinks from the list.
 */
#define listDel(start, t, condition) ({					\
	typeof (start) old = NULL;					\
	listItt(start, t) {						\
		if(condition) break;					\
		old = (t);						\
	}								\
	listRm(start, t, old);						\
})

////////////////////////////////////////////////////////////////////////////////
// Double-lists
////////////////////////////////////////////////////////////////////////////////

/* macro: DLIST(type)
 * desc: Makes an entry into your struct 'type' for a doubly-linked list.
 * For Example: struct foo { ...; DLIST(struct foo); };
 */
#define DLIST(type) type *next, *prev

/* macro: DLIST_INIT(start)
 * desc: Statically initalizes a doubly-linked list. See also dlistInit()
 */
#define DLIST_INIT(s) next: (s), prev: (s)

/* macro: dlistPrepend(where, element)
 * desc: Adds 'element' into a doubly-linked list before the element 'where'.
 */
#define dlistPrepend(a, b) do {						\
	if(a) {								\
		(b)->next = (a)->prev;					\
		(b)->prev = (a)->prev->prev;				\
		(a)->prev->next = (b);					\
		(a)->prev = (a);					\
	} else {							\
		(a) = (b);						\
		dlistInit(a);						\
	}								\
} while(0)

/* macro: dlistAdd(where, element)
 * desc: Adds 'element' into a doubly-linked list after the element 'where'.
 */
#define dlistAdd(a, b) do {						\
	if(a) {								\
		(b)->next = (a)->next;					\
		(b)->prev = (a);					\
		(a)->next->prev = (b); /* This goes before... */	\
		(a)->next = (b);					\
	} else {							\
		(a) = (b);						\
		dlistInit(a);						\
	}								\
} while(0)

/* macro: dlistDel(list, element)
 * desc: Removes 'element' from the doubly linked list 'list'. (doesn't kfree
 * any memory)
 */
#define dlistDel(s, e) do {						\
	if((e) && (e) == (e)->next && (e) == (e)->prev && (e) == (s))	\
		(s) = NULL;						\
	else {								\
		(e)->prev->next = (e)->next;				\
		(e)->next->prev = (e)->prev;				\
		(s) = (e)->next;					\
	}								\
} while(0)

/* macro: dlistInit(start)
 * desc: Initalizes the first element of a doubly-linked list (must be in
 * function context. For a static-initalizer, please see DLIST_INIT)
 */
#define dlistInit(s) do {						\
	(s)->next = (s);						\
	(s)->prev = (s);						\
} while(0)

/* macro: dlistItt(start, variable, int)
 * desc: Itterates over a doubly-linked list using the variable 'variable'.
 * 'int' is used to show the current position in the list.
 */
#define dlistItt(s, t, i)						\
	for((i) = 0, (t) = (s); (t) && (!(i) || ((t) != (s))); (t) = (t)->next, (i)++)

/* macro: dlistCount(start)
 * desc: This macro returns the number of elements in the doubly-linked list
 * 'start'.
 */
#define dlistCount(s) ({						\
	ulong count = 0;						\
	typeof (s) t;							\
	dlistItt(s, t, count) ;						\
	count;								\
})

#endif /* __generic_kern_llist_H */
