/*
 * signals.h - Signal handling
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 04/01/02
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * OK, this file provides 3 functions, but they are all *really* important to
 *	the whole process of communication between modules.
 *
 * 'connect' connects and event to a function with a given priority. 0 is the
 *	highest, 255 is the lowest. if 2 connect's to the same event, the 1st
 *	one should get there before the other one. For example:
 *		connect("DISK_ERROR", handleDiskError, 1);
 *
 * 'disconnect' disconnects a function connected to an event. For identification
 *	purposes, you must also provide the address of the handler function.
 *	For example:
 *		disconnect("DISK_ERROR",handleDiskError);
 *
 * 'trigger' triggers the specified event. For example:
 *		trigger("DISK_ERROR");
 *
 * Notes/Fixme's:
 *
 * Changelog:
 */

#ifndef __signals_H
#define __signals_H

typedef void (sigHandler)(void *, void *);

errRet sigConNum(__uint32, sigHandler *, __uint8, __uint32, void *);
errRet sigDisNum(__uint32, sigHandler *);
errRet sigTrigNum(__uint32, void *);
__uint32 sigReqReg(__uint32);

errRet sigCon(char *, sigHandler *, __uint8, __uint32, void *);
errRet sigDis(char *, sigHandler *);
errRet sigTrig(char *, void *);
initFunc signalsInit();

#define NO_DELAY 1

/* macro: DEF_PRIO
 * description: The default value for you to connect your functions with.
 */
#define DEF_PRIO 128

/* macro: SIG_HND(name, data block, trigger data)
 * desc: This macro creates a signal handler. The last two arguments passed are
 * the names and types of data which is passed when the signal is created, and
 * then triggered.
 */
#define SIG_HND(x,y,z) void x(y,z)

/* macro: SIG_DEF_HND(name)
 * desc: This macro creates a default signal handler. Data specified in signal
 * creation is passed as 'void *dataBlk' and the data passed when trigger()'d is
 * of type 'void *trigData'.
 */
#define SIG_DEF_HND(x) SIG_HND(x, void *dataBlk, void *trigData)

#if CONFIG_SIGNALS & OPT_IS_DEBUG
extern bool signalsDebug;
#endif

#endif // __signals_H
