/*
 * cpu.h - i386-specific CPU things
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 19/02/02
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 17/04/01 bodnar: Changed all #define assembly into static inline assembly.
 * 03/09/01 Zwane: Changes to archCPU struct to support L1,L2 etc caches
 * 23/09/01 Zwane: Added control register defines.
 * 24/09/01 Mark: Moved cpu.c's wbinvd() to here and simplified.
 * 14/10/01 Zwane: Adding framework for merging with SMP code
 */

#ifndef __i3_cpu_H
#define __i3_cpu_H

#include <desc.h>
#include <task.h>

/* cpu capability masks */
#define CPU_CAP_FPU		0x00
#define CPU_CAP_VME		0x01
#define CPU_CAP_DE		0x02
#define CPU_CAP_PSE		0x03
#define CPU_CAP_TSC		0x04
#define CPU_CAP_MSR		0x05
#define CPU_CAP_PAE		0x06
#define CPU_CAP_MCE		0x07
#define CPU_CAP_CX8		0x08
#define CPU_CAP_APIC		0x09
#define CPU_CAP_SEP		0x0B
#define CPU_CAP_MTRR		0x0C
#define CPU_CAP_PGE		0x0D
#define CPU_CAP_MCA		0x0E
#define CPU_CAP_CMOV		0x0F
#define CPU_CAP_PAT		0xA0
#define CPU_CAP_PSE36		0xA1
#define CPU_CAP_MMX		0x17
#define CPU_CAP_FXSR		0x18
#define CPU_CAP_SSE		0x19

/* control register bit fields */
#define CR0_PE                  0x00
#define CR0_MP                  0x01
#define CR0_EM                  0x02
#define CR0_TS                  0x03
#define CR0_ET                  0x04
#define CR0_NE                  0x05
#define CR0_WP                  0x10
#define CR0_AM                  0x12
#define CR0_NW                  0x1D
#define CR0_CD                  0x1E
#define CR0_PG                  0x1F

/* types of reboots */
#define REBOOT_COLD		0x0000
#define REBOOT_WARM		0x1234

enum {
	L1 = 0,
	L2,
	L3,
	/* L4? ;) */
	MAX_CACHE		/* 3 */
};

/* The version of the CPU's are as follows:
 * 3		i386
 * 4		i486DX, SX or i487SX
 */

/* struct: archCPU
 * description: Contains all of the info about a specific CPU in the box.
 */
struct archCPU {
	char cpuid[13];		// enough to store 3x32bit registers
	char info[43];		// 1 for a \0
	bool cache[MAX_CACHE];	// Has a cache?
	__uint8 family;		// family of CPU (see above)
	__uint8 model;
	__uint8 stepping;
	__uint16 cache_size[MAX_CACHE];	// 0 = L1, 1 = L2 ...
	__uint32 maxCPUID;	// max number of CPUID's (0 for not allowed)
	__uint32 maxExtCPUID;	// max extended cpuid op
	__uint32 capability;	// What the CPU can do :)
	struct TSS TSS;		// Only 1 TSS per-processor
	ulong TSSGDTEntry;	// GDT Entry for the TSS
#ifdef APIC
	__uint8 apic_id;		// We don't support APIC on UP
	__uint8 apic_ver;
#endif /* APIC */
};

struct CPU;			/* Forward declaration... */
void dumpRegs(struct allRegs *);
void dumpStack(ulong *);
void CPUDetectMhz(void);
void wbinvd(struct CPU *);
void CPUDetectCPUID(struct CPU *);
void CPUDetectCapability(struct CPU *);
void CPUDetectMax(struct CPU *, bool extended);
errRet CPUDetect(ulong);
errRet CPUSetup(ulong);
void reboot();
void halt();

#include <generic/kern/cpu.h>

#endif // __cpu_H
