/*
 * desc.h - Descriptor table file for i386
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 04/01/02
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * This file contains general defines and structs to setup an entry in the GDT,
 * IDT or LDT
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 25/09/00 mark: I hacked the startInt macro etc to do some neater stuff than
 *	it did before. Also added some new ones and made it safer/better than
 *	before. I finally now understand interrupts (hahaha, I think)
 * 31/10/00 mark: I re-did all of them, to take up less space :)
 * 15/05/01 mark: Minor fixups of some of the macros to make them neater
 */

#ifndef __desc_H
#define __desc_H

////////////////////////////////////////////////////////////////////////////////
// IDT Stuff
errRet setGate(void *, ulong, char, char);
void unsetGate(ulong);
bool intSetup();
initFunc IDTInit();

extern void INTHandler();

/* struct: desc
 * description: A descriptor for stuff like lgdt() and lidt()
 */
struct desc {
	__uint16 limit;
	memory base;
} __PACKED__;

/* struct: gateDesc
 * description: Structure of a gate on the i386
 */
struct gateDesc {
	ushort	offsetLow,
		TSS;
	uchar	dwordCount,
		type		:5,
		DPL		:2,
		present		:1;
	ushort	offsetHigh;
} __PACKED__;

#define CALL_GATE 12
#define INT_GATE 14
#define TRAP_GATE 15

/* function: setIntGate(handler, interrupt, DPL)
 * arg0: The address of the handler for the interrupt.
 * arg1: The interrupt number ot catch.
 * arg2: The Privilage level for the code in the interrupt to be executed in.
 * description: This function sets a interrupt gate with the specified
 * arguments.
 */
#define setIntGate(x,y,z) setGate(x,y,z, INT_GATE)

/* function: setCallGate(handler, interrupt, DPL)
 * arg0: The address of the handler for the interrupt.
 * arg1: The interrupt number ot catch.
 * arg2: The Privilage level for the code in the interrupt to be executed in.
 * description: This function sets a call gate with the specified arguments.
 */
#define setCallGate(x,y,z) setGate(x,y,z, CALL_GATE)

/* function: setTrapGate(handler, interrupt, DPL)
 * arg0: The address of the handler for the interrupt.
 * arg1: The interrupt number ot catch.
 * arg2: The Privilage level for the code in the interrupt to be executed in.
 * description: This function sets a trap gate with the specified arguments.
 */
#define setTrapGate(x,y,z) setGate(x,y,z, TRAP_GATE)

/* struct: allRegs
 * description: A struct with all the registers pushed onto the stack by an
 * interrupt.
 */
struct allRegs {
	ulong es;
	ulong ds;
	ulong ebp;
	ulong esi;
	ulong edi;
	ulong edx;
	ulong num;
	ulong ecx;
	ulong ebx;
	ulong eax;
	ulong arg;		// Arg pushed by CPU sometimes, us elsewhere

	// Pushed by CPU
	ulong eip;
	ulong cs;
	ulong eflags;

	ulong esp;		// Only pushed if there is a PL change
	ulong ss;
} __PACKED__;

// Hard coded offsets in the above struct:
#define ARG_OFFSET 44

#define startCall(real, num, dif, call, code)			\
	extern void real();					\
	asm (							\
		".globl " #real	"			\n"	\
		#real ":				\n"	\
		"	" code "			\n"	\
		"	pushl %eax			\n"	\
		"	pushl %ebx			\n"	\
		"	pushl %ecx			\n"	\
		"	movl $" #dif ", %eax		\n"	\
		"	movl $(" STR(num) "), %ebx	\n"	\
		"	movl $" #call ", %ecx		\n"	\
		"	jmp INTHandler			\n")

/* macro: startIntr(functionName, number, calls)
 * arg0: Name of the function.
 * arg1: Interrupt number.
 * arg2: The function that it calls to handle the interrupt.
 * description: Makes a basic interrupt handler called 'functionName'. This is
 * the name you should pass to set*Gate(). It the calls off 'calls'. IT IS NOT
 * RECOMENDED THAT YOU USE THIS FUNCTION FOR YOUR INTERRUPT HANDLERS UNLESS YOU
 * KNOW WHAT YOU'RE DOING, try to use startInt() instead, it's a lot easier :)
 */
#define startIntr(n, nu, fu) startCall(n, nu, fu, endIntSpecial, "pushl $0")

/* macro: startException(functionName, number, calls)
 * desc: This macro is similar to startIntr(), but it expects 'functionName' to
 * be a handleExcept() constructed function. I'd recomend you used startExcept()
 * instead of this and handleExcept().
 */
#define startException(n, nu, fu) startCall(n, nu, fu, exceptHandler, "")

/* macro: handleInt(functionName)
 * arg0: Name of the function to create
 * description: Like signalHnd(), it creates a interrupt handler. ATM, it's the
 * same as void functionName(struct allRegs *r), but this could
 * change.
 */
#define handleInt(real) void asmLinkage real(struct allRegs *r)

/* macro: handleExcept(functionName)
 * desc: This is like handleInt(), but makes a function for startException(). It
 * gives 'arg' as type ulong as well as the normal struct allRegs *r that
 * handleInt() gives. See also startExcept().
 */
#define handleExcept(real) void asmLinkage real(struct allRegs *r, ulong arg)

/* typedef: intFunc
 * description: A type that will point to a function of type handleInt().
 */
typedef void (*intFunc)(struct allRegs *);

/* macro: startInt(functionName, number)
 * arg0: The name of the function.
 * arg1: The interrupt number.
 * description: A user-friendly function for making an interrupt handler. 99% of
 * cases should use this unless you've RTFS'd and understand it.
 */
#define startInt(real, num)	startIntr(real, num, _##real);		\
				handleInt(_##real)

/* macro: startExcept(functionName, number)
 * desc: This is like startInt(), but for exceptions. It passes an extra arg
 * which the processor pushes. It should only be used for some exceptions, and
 * not the ones that don't push an argument.
 */
#define startExcept(real, num)	startException(real, num, _##real);	\
				handleExcept(_##real)

////////////////////////////////////////////////////////////////////////////////
// GDT Stuff
initFunc GDTInit();
__uint16 GDTAddEntry(__uint32, __uint32, bool, bool, __uint8, uchar, bool);
void GDTDelEntry(__uint16);

// Defines
#define SEG_DATA 0
#define SEG_CODE 8

#define SEG_GROW_DOWN 4

#define SEG_RO 0
#define SEG_RW 2

#define SEG_ACCESSED 1			// Set by the CPU itself

extern union segDesc *GDTArea;

// Some values
extern __uint16 kernData, kernCode;

////////////////////////////////////////////////////////////////////////////////
// LDT Stuff
void LDTAddEntry(union segDesc *, ulong, ulong, uchar, uchar);

////////////////////////////////////////////////////////////////////////////////
// LDT and GDT shared stuff

/* struct: segDesc
 * description: A union that describes each segment GDT (and LDT).
 */
union segDesc {
	struct {
		ushort	limitLow,
			baseLow;
		uchar	baseMed,
			type		:4,
			descType	:1,
			DPL		:2,
			present		:1,
			limitHigh	:4,
			ZealOS		:1,
			not_used	:1,
			segmentSize	:1,
			granularity	:1,
			baseHigh;
	} norm __PACKED__;
	__uint64 dummy;
};

#endif	// __desc_H
