/*
 * init.h - i386-specific initalization routines
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 04/01/02
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 * Multiboot spec can be found in grub's doc/ directory - Mark
 *
 * Changelog:
 * 13/10/01 Mark: rewrote to allow booting by a multiboot-compatable loader
 */

#ifndef __init_H
#define __init_H

initFunc initInit();

/* macro: MAX_COMMAND_LEN
 * description: Maximum length of command line that we parse from a loader
 */
#define MAX_COMMAND_LEN 128

/* macro: ARG_AREA
 * description: Address that the our bootloader puts arguments at
 */
#define ARG_AREA 0x9F000

/* macro: MMAP_LEN
 * description: Maximum length of the memory map that we get (E820 command)
 */
#define MMAP_LEN 0x100

/* macro: MMAP_ENTRIES
 * description: Maximum memory map entries we can have. See also: MMAP_LEN
 */
#define MMAP_ENTRIES (MMAP_LEN/sizeof(struct memMap))

/* struct: memMap
 * description: This is a struct that we rip as much info out of the bios as
 * possible from. It's basically what the BIOS gives us as info, however only
 * newer BIOS's actually give us this, so for older ones we have to make it up
 * as we go along ;) The meaning of each 'type' entry is shown in the table
 * below:
 *
 * value | meaning<BR>
 * ------+-----------------------------------<BR>
 * 1     | normal memory, avaliable to us!!<BR>
 * 2     | reserved (eg ROM, mmapped device)<BR>
 * 3     | ACPI Reclaim memory<BR>
 * 4     | ACPI NVS memory (dont touch!!)<BR>
 */
#define E820_NORMAL 1
#define E820_RESERVED 2

struct memMap {
	__uint64 base;			/* Base addr (in bytes) */
	__uint64 len;			/* Length (in bytes) */
	__uint32 type;			/* Type, see above */
} __PACKED__;

/* Stuff passed by our bootloader */
struct bootloader_info {
	ulong memSize;			/* Size of upper memory in Kb */
	ulong mapLen;			/* Size of our memory map (in bytes) */
	struct memMap *map;		/* Pointer to it */
	char *cmdline;			/* lilo command line */
} __PACKED__;

/* Multiboot stuff */
#define MULTIBOOT_HDR_MAGIC 0x1BADB002
#define MULTIBOOT_BOOT_MAGIC 0x2BADB002

/* Indicates we have extended options */
#define MULTIBOOT_FLAGS (1 << 16)

struct mboot_memMap {
	__uint32 size;		// Extra thing to fuck us up...
	struct memMap map;
	/* May be some padding... */
} __PACKED__;

/* struct: multiboot_hdr
 * desc: Defines the structure of the struct that tells the multiboot compliant
 * loader how/where to load us.
 */
struct multiboot_hdr {
	__uint32 magic;
	__uint32 flags;
	__uint32 checksum;
	memory header_addr;
	memory load_addr;
	memory load_end_addr;
	memory bss_end_addr;
	memory entry_addr;
	/* No video stuff */
};

/* Test for a certain flag in the multiboot_info struct */
#define MBOOT_FLAG(m,f) ((m)->flags & (f))
#define MBOOT_MEM	(1 << 0)
#define MBOOT_BOOT	(1 << 1)
#define MBOOT_CMD	(1 << 2)
#define MBOOT_MOD	(1 << 3)
#define MBOOT_SYMS4	(1 << 4)
#define MBOOT_SYMS5	(1 << 5)
#define MBOOT_MMAP	(1 << 6)
#define MBOOT_DRIVES	(1 << 7)
#define MBOOT_CONFIG	(1 << 8)
#define MBOOT_BNAME	(1 << 9)
#define MBOOT_APM	(1 << 10)
#define MBOOT_VBE	(1 << 11)

/* struct: multiboot_info
 * desc: Info that a multiboot-compatable loader will pass to us */
struct multiboot_info {
	__uint32 flags;
	__uint32 mem_lower;
	__uint32 mem_upper;
	__uint32 boot_device;
	char *cmdline;
	__uint32 mods_count;
	__uint32 mods_addr;
	__uint32 syms[4];
	__uint32 mmap_length;
	struct mboot_memMap *mmap_addr;
	__uint32 drives_length;
	__uint32 drives_addr;
	__uint32 config_table;
	char *boot_loader_name;
	/* More crap we don't worry about */
};

/* LD Symbols we use/could use */
extern char	textStart,
		textEnd,
		textLength,
		dataStart,
		dataEnd,
		dataLength,
		delableStart,
		delableEnd,
		delableLength,
		funcStart,
		funcLen,
		subStart,
		subLen,
		fsStart,
		fsLen,
		syscallStart,
		syscallLen,
		bssEnd,
		loadEnd,
		startFunc,
		kernEnd,
		kernStart,
		kernLen;

/* Stuff we let the rest of the kernel look at */
extern char commandLine[MAX_COMMAND_LEN];	/* The un-parsed command line */
extern struct memMap mmap[MMAP_ENTRIES];	/* A copy of the memory map */
extern ulong mmapLen;				/* Number of elements in mmap */
extern memory kernelStart, kernelLength;

#endif // __init_H
