/*
 * io.h - IO instructions on the i386 platform
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 03/10/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * OK, I've made this file look really neat now... a bit of macro misuse but so
 *	what :) Thanks go to arjan and prumpf who helped w/ a few of the asm
 *	questions. This looks a lot better than linux's and is much more
 *	optomized (IMHO). This file should do everything you need to do, with
 *	the minimum of fuss and the most optomized. If you specify a constant
 *	to out* on in*, it will try to optomize around it and put it in the
 *	instruction. - Mark
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 16/03/01 Mark: Now added string operations, shouldn't have broken anything
 */

#ifndef __IO_H
#define __IO_H

// 'normal' IO
#define out(l,y,v,p)						\
	do {							\
	if((__builtin_constant_p(p)) && ((p) < 0x100))		\
		asm volatile ("out" #l " %%" #y ",%0"::"g" (p), #y (v));\
	else							\
		asm volatile ("out" #l " %%" #y ",%%dx"::"d" (p), #y (v));\
	} while(0)

/* function: outb, outw, outl (value, port)
 * arg0: The value to output to the port
 * arg1: The port to put the value to
 * description: Writes a byte/word/double-word with the contents of 'value' to
 * port 'port' and tries to optomize as far as possible.
 */
#define outb(v,p) out(b,al,v,p)
#define outw(v,p) out(w,ax,v,p)
#define outl(v,p) out(l,eax,v,p)

#define in(let,reg,type,x,y) static inline type in##let##x(__uint16 p)\
{									\
	type ret=0;							\
	if((__builtin_constant_p(p)) && (p < 0x100))			\
		asm volatile("in" #let " %1,%%" #reg : "=a" (ret) : "g" (p));\
	else								\
		asm volatile("in" #let " %%dx,%%" #reg : "=a" (ret) : "d" (p));\
	y;								\
	return ret;							\
}

/* function: inb, inw, inl (port)
 * arg0: The port to get the value from
 * Description: Gets a byte/word/double-word from the specified IO port. Tries
 * to optomize as much as possible.
 */
in(b,al,__uint8,,)
in(w,ax,__uint16,,)
in(l,eax,__uint32,,)

// String operations:
#define str(t,l,r) extern inline void * t##s##l(__uint16 p, void *b, ulong c)\
{									\
	asm volatile("cld;rep;" #t "s" #l : #r (b) : "0" (b), "c" (c), "d" (p));	\
	return b;							\
}

/* function: outbs, outws, outls (__uint16 port, void *data, ulong count)
 * arg0: The port to put the value to
 * arg1: The beginning of the data to write to it
 * arg2: The number of times to do the operation
 * description: Writes 'count' bytes/words/double-words from 'data' to the
 * specified port.
 */
str(out,b,=S)
str(out,w,=S)
str(out,l,=S)

/* function: inbs, inws, inls (__uint16 port, void *data, ulong count)
 * arg0: The port to get the value from
 * arg1: The beginning of the area to read it to
 * arg2: The number of times to do the operation
 * description: Reads 'count' bytes from the specified port and puts them in
 * 'data'.
 */
str(in,b,=D)
str(in,w,=D)
str(in,l,=D)
#undef str

// Delayed IO
#ifdef DELAY_USING_IO
// Erk, unsafe..
#define DELAY_BEGIN outb(0,0x80)
#else
#define DELAY_BEGIN asm volatile("jmp 1f\n1: jmp 1f\n1: ")
#endif

/* function: pause()
 * description: Delays by a few cycles (Mostly used for gaps in IO).
 */
#ifdef REALLY_SLOW_DOWN_IO
#define pause() do { DELAY_BEGIN; DELAY_BEGIN; DELAY_BEGIN; DELAY_BEGIN; } \
						while(0);
#else
#define pause() DELAY_BEGIN
#endif

/* function: outb_p, outw_p, outl_p (value, port)
 * description: Same as outb()/outw()/outl() but pauses (via pause()) for a
 * short time afterwards.
 */
#define outb_p(v,p) do{outb(v,p); pause();} while(0)
#define outw_p(v,p) do{outw(v,p); pause();} while(0)
#define outl_p(v,p) do{outl(v,p); pause();} while(0)

/* function: inb_p, inw_p, inl_p (value, port)
 * description: Same as inb()/inw()/inl() but pauses (via pause()) for a short
 * time afterwards.
 */
in(b,al,__uint8,_p,pause())
in(w,ax,__uint16,_p,pause())
in(l,eax,__uint32,_p,pause())

#undef in

/* macro: beep()
 * desc: Not really sure where this should go. It beeps a lot to complain about
 * something.
 */
#define beep() outb(3, 0x61)

#endif
