/*
 * smp.h - MP 1.4 implementation
 *
 * Added to the Zealos project and (C) by Zwane Mwaikambo
 *
 * Current maintainer: Zwane Mwaikambo <zwane@commfireservices.com>
 * Status: Maintained
 * Last update: 04/01/02
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * MP 1.4 implementation based on Intel specs
 *
 * Notes/Fixme's:
 *
 * Changelog:
 *
 */

#ifndef __i386_smp_H
#define __i386_smp_H

#define MP_MAGIC		REVERSE_PACK('_','M','P','_')
#define MP_CONF_MAGIC		REVERSE_PACK('P','C','M','P')

/* extended entry types */
#define ENTRY_CPU		0
#define ENTRY_BUS		1
#define ENTRY_IOAPIC		2
#define ENTRY_IOIRQ		3
#define ENTRY_LCLIRQ		4
#define ENTRY_UNKNOWN		-1

/* mem location defines */
#define MEM_BASE_START		0x00000000
#define MEM_BASE_END_512	0x00080000	/* 512k */
#define MEM_BASE_END		0x0009FFFF	/* 640k */
#define MEM_EXT_START		0x00100000	/* 1024k+ */
#define MEM_EXT_END		0xFEBFFFFF	/* well it depends on how much ram you got */
#define MEM_ROM_START		0x000F0000	/* system ROM bios */
#define MEM_ROM_END		0x000FFFFF	/* 128k */
#define MEM_EBDA_START		0x000040e0
#define MEM_EBDA_END		0x000050e0	/* 4k */

/* found myself typing this too much */
#define uchar_p(x)		((uchar *)x)

/* MP Floating pointer structure */
typedef struct {
	__uint32 signature;	/* magic, should be _MP_ */
	__uint32 address;	/* pointer to the MP configuration table if any
				 */
	__uint8  length;
	__uint8  revision;	/* 0x1 == 1.1 0x4 == 1.4 */
	__uint8  csum;		/* a checksum of whole struct should be zero */

	/* MP feature stuff */
	__uint8  info1;
	__uint8  info2;		/* bits 0-6 reserved, 7 is IMCR */
	uchar  reserved[3];

} __PACKED__ MPFloatingPS_t;

/* MP Configuration table */
typedef struct {
	__uint32 signature;	/* magic should be PCMP */
	__uint16 base_length;
	__uint8 spec_rev;
	__uint8 csum;
	uchar idstring[8];
	uchar product_id[12];
	__uint32 oemid_ptr;
	__uint16 oemtable_size;
	__uint16 entry_count;
	__uint32 apic_addr;
	__uint16 extended_length;
	__uint8 extended_csum;
	__uint8 reserved;
} __PACKED__ MPConfTable_t;

/* these are all extended MP configuration table entry types */
typedef struct {
	__uint8 entry_type;		/* this would be BASE_ENTRY_CPU */
	__uint8 apic_id;		/* unique id for cpu */
	__uint8	apic_ver;

#define SMP_CPU_EN	0x01
#define SMP_CPU_BSP	0x02
	__uint8 flags;	/* bit 0 - if 0 cpu should not be used, bit 1 set if cpu
			 * is BSP, rest reserved*/

	/* Signature: 0-3 = stepping 4-7 = model family = 8-11 */
	unsigned stepping: 4;
	unsigned model: 4;
	unsigned family: 4;

	__uint32 capability;
} __PACKED__ CPUEntry_t;

typedef struct {
	__uint8 entry_type;		/* this should be BASE_ENTRY_BUS */
	__uint8 bus_id;
	uchar bus_type_str[6];
} __PACKED__ BUSEntry_t;

typedef struct {
	__uint8 entry_type;		/* this should be BASE_ENTRY_IOAPIC */
	__uint8 ioapic_id;
	__uint8 ioapic_ver;
#define SMP_IOAPIC_EN	0x01
	__uint8 flags;			/* if bit 0 is clear, this io apic is
					 * unusable */
	__uint32 ioapic_addr;		/* base address for this ioapic */
} __PACKED__ IOAPICEntry_t;

typedef struct {
	__uint8 entry_type;		/* this should be BASE_ENTRY_LCLIRQ */
	__uint8 irq_type;

#define SMP_LIRQ_EL	0x01
#define SMP_LIRQ_PO	0x02
	__uint16 flags;

	__uint8 src_bus_id;
	__uint8 src_bus_irq;
	__uint8 dest_apic_id;
	__uint8 lintin_num;
} __PACKED__ localIRQEntry_t, IOIRQEntry_t;

typedef struct {
	MPFloatingPS_t *MPFloatingPS;		/* pointer to the MP floating
						 * point struct */

	MPConfTable_t *MPConfTable;		/* pointer to the MP
						 * configuration table */

	/* i need these here till i merge with the main cpu code */
	__uint8 num_aps;			/* number of Application
						 * Processors, BSP=0 */
	__uint8 num_buses;
	__uint8 num_ioapics;			/* we start counting from 1 */

	/* Temporary solution i think i have NULL deref problems */
	#define MAX_APS		1
	#define MAX_BUSES	1
	#define MAX_IOAPICS	1

	CPUEntry_t CPUEntry[MAX_APS];
	BUSEntry_t BUSEntry[MAX_BUSES];
	IOAPICEntry_t IOAPICEntry[MAX_IOAPICS];
	
	/* other junk */
} SMPData_t;

initFunc SMPInit();

#endif /* __i386_smp_H */

