/*
 * spinlock.h - Simple spinlock implementation - i386 specific
 *
 * Added to the Zealos project by and copyright Zwane Mwaikambo
 *
 * Current maintainer: Zwane Mwaikambo <zwane@commfireservices.com>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 26/09/01 Zwane: switched to asm for SMP lock functions so i can add bus locks
 * 26/09/01 Mark: Fixed some of the locking code
 * 26/09/01 Mark: Added some debug code and rearranged code a bit
 * 30/09/01 Mark: Added deadlock detection on UP machines
 * 11/10/01 Mark: Major new stuff and optomizations :)
 * 11/10/01 Mark: Fixed the smp race in spinlock, improved the usial code path
 */

#ifndef __spinlock_H
#define __spinlock_H

#include <system.h>

/* Spinlocks arn't needed unless we are either debugging spinlocks, or on an smp
 * box, or both :) - Mark */
#if defined(__SMP__) || defined(DEBUG_SPINLOCKS)
# define SPINLOCKS_NEEDED
#else /* defined(__SMP__) || defined(DEBUG_SPINLOCKS) */
# undef SPINLOCKS_NEEDED
#endif /* defined(__SMP__) || defined(DEBUG_SPINLOCKS) */

#define SPINLOCK_MAJIK 0x1314beef

#ifdef DEBUG_SPINLOCKS
# define SPINLOCK_T , majik: SPINLOCK_MAJIK
#else /* DEBUG_SPINLOCKS */
# define SPINLOCK_T
#endif /* DEBUG_SPINLOCKS */

/* Umm, this is a bit silly, we don't know what the locked value is, any number
 * that's not 1 is locked... */
#define SPINLOCK_LOCKED		0
#define SPINLOCK_UNLOCKED	1

/* If we don't use the spinlocks, we don't need to initalize them. We don't have
 * any need to waste space... - Mark
 */
#ifdef SPINLOCKS_NEEDED
# define INIT_LOCKED		{lock: SPINLOCK_LOCKED SPINLOCK_T}
# define INIT_UNLOCKED		{lock: SPINLOCK_UNLOCKED SPINLOCK_T}
#else /* SPINLOCKS_NEEDED */
# define INIT_LOCKED		{}
# define INIT_UNLOCKED		{}
#endif /* SPINLOCKS_NEEDED */

typedef struct {
/* No need to waste memory... */
#ifdef SPINLOCKS_NEEDED
	volatile __uint32 lock;
#endif /* SPINLOCKS_NEEDED */

	__uint32 majik;

} spinlock_t;

/* If we opt to debug spinlocks, it will warn and maybe halt (if specified) if
 * we use an unintalized or corrupted spinlock. */
#ifdef DEBUG_SPINLOCKS

/* Needed for printing the warnings */
# include <debug.h>

# ifdef SPINLOCK_HALT
#  define SPINLOCK_H while(1)
# else /* SPINLOCK_HALT */
#  define SPINLOCK_H
# endif /* SPINLOCK_HALT */

# ifdef __SMP__
/* Can only detect deadlocks on UP boxen... */
#  define CHECK_DEADLOCK(l)
# else /* __SMP__ */
#  define CHECK_DEADLOCK(l)						\
	if((l)->lock != SPINLOCK_UNLOCKED) {				\
		DPRINT("Spinlock deadlock detected '" #l "'!!\n");	\
		SPINLOCK_H;						\
	}
# endif /* __SMP__ */

# define CHECK_INIT(l)							\
	if((l)->majik != SPINLOCK_MAJIK) {				\
		DPRINT("Use of uninitalized spinlock '" #l "'!!\n");	\
		SPINLOCK_H;						\
	}

# define spinlock(x) do {CHECK_INIT(x); CHECK_DEADLOCK(x); __spinlock(x); } while(0)
# define spinlockUnlock(x) do {CHECK_INIT(x); __spinlockUnlock(x); } while(0)

#else /* DEBUG_SPINLOCKS */

/* Else, we just fall throught to use normal spinlocks... */
# define spinlock(x) __spinlock(x)
# define spinlockUnlock(x) __spinlockUnlock(x)

#endif /* DEBUG_SPINLOCKS */

static inline void __spinlock(spinlock_t *lock)
{
#ifdef __SMP__
	/* Lock so that only 1 processor at a time can access the code. This is
	 * good code, only takes 2 instructions (2 cycles) on the main code
	 * path. - Mark */
	asm volatile (
		"1:						\n"
		/* Decrement the number */
		LOCK "	decl %0					\n"
		/* If we're now 0, then we have got the lock, end */
		"	jnz 2f					\n"
		/* Actual end is here */

		/* Put this in a section of it's own, it's not the common case,
		 * and it puts it out of the main code path */
		".section .text.locks, \"ax\"			\n"
		/* Lock is now some negative number, just loop round in this
		 * 'soft' loop till we see the lock being released. Then we try
		 * the lock properly again. This is because the 'lock'
		 * instruction will lock *all* the CPU's, so it's a bad thing to
		 * do it a lot as it will virtually stop the system. */
		"2:						\n"
		/* let the cpu idle a bit */
		"	rep; nop				\n"
		/* See if the lock has been unlocked */
		"	cmpl $" STR(SPINLOCK_UNLOCKED) ", %0	\n"
		/* No luck - Spin 'softly' */
		"	jne 2b					\n"
		/* It has been, might have been locked again though, go back and
		 * try to get the lock */
		"	jmp 2b					\n"
		/* End of the section */
		".previous					"
		: "=m" (lock->lock) : : "memory");
#elif defined(SPINLOCKS_NEEDED)
	lock->lock--;		/* Uhh, 'lock' it.. This is only for debugging*/
#endif
}

static inline void __spinlockUnlock(spinlock_t *lock)
{
#ifdef SPINLOCKS_NEEDED
	/* Unlock, no need to lock the bus for this.. */
	lock->lock = SPINLOCK_UNLOCKED;
#endif /* SPINLOCKS_NEEDED */
}

static inline void spinlockInit(spinlock_t *lock)
{
#ifdef SPINLOCKS_NEEDED
	lock->lock = SPINLOCK_UNLOCKED;
#endif /* SPINLOCKS_NEEDED */
	lock->majik = SPINLOCK_MAJIK;
}

/* Higher-level things, these are needed on both SMP and UP boxes */
static inline void spinlockIRQSave(spinlock_t *lock, flags *flags)
{
	*flags = get_flags();
	disable_local_interrupts();
	spinlock(lock);
}

static inline void spinlockIRQRestore(spinlock_t *lock, flags *flags)
{
	spinlockUnlock(lock);
	set_flags(*flags);
}

#endif /* __spinlock_H */
