/*
 * system.h - i386-specific functions to do stuff you can't do in C.
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 12/12/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * This file was seperated from cpu.h so that all the asm can go here.
 *
 * Notes/Fixme's:
 * 11/12/01 Mark: Fix the #defines at the bottom with checking the capabilities
 *
 * Changelog:
 * 17/04/01 bodnar: Changed all #define assembly into static inline assembly.
 * 23/09/01 Zwane: Added control register defines.
 * 24/09/01 Mark: Moved cpu.c's wbinvd() to here and simplified.
 * 14/10/01 Zwane: Adding framework for merging with SMP code
 * 11/12/01 Mark: Split into cpu.h and system.h
 */

#ifndef __i386_system_H
#define __i386_system_H

#include <task.h>

/* cpu capability masks */
#define CPU_CAP_FPU		0x00
#define CPU_CAP_VME		0x01
#define CPU_CAP_DE		0x02
#define CPU_CAP_PSE		0x03
#define CPU_CAP_TSC		0x04
#define CPU_CAP_MSR		0x05
#define CPU_CAP_PAE		0x06
#define CPU_CAP_MCE		0x07
#define CPU_CAP_CX8		0x08
#define CPU_CAP_APIC		0x09
#define CPU_CAP_SEP		0x0B
#define CPU_CAP_MTRR		0x0C
#define CPU_CAP_PGE		0x0D
#define CPU_CAP_MCA		0x0E
#define CPU_CAP_CMOV		0x0F
#define CPU_CAP_PAT		0xA0
#define CPU_CAP_PSE36		0xA1
#define CPU_CAP_MMX		0x17
#define CPU_CAP_FXSR		0x18
#define CPU_CAP_SSE		0x19

/* control register bit fields */
#define CR0_PE                  0x00
#define CR0_MP                  0x01
#define CR0_EM                  0x02
#define CR0_TS                  0x03
#define CR0_ET                  0x04
#define CR0_NE                  0x05
#define CR0_WP                  0x10
#define CR0_AM                  0x12
#define CR0_NW                  0x1D
#define CR0_CD                  0x1E
#define CR0_PG                  0x1F

/* types of reboots */
#define REBOOT_COLD		0x0000
#define REBOOT_WARM		0x1234

enum {
	L1 = 0,
	L2,
	L3,
	/* L4? ;) */
	MAX_CACHE		/* 3 */
};

/* struct: archCPU
 * description: Contains all of the info about a specific CPU in the box.
 */
struct archCPU {
	char cpuid[13];		// enough to store 3x32bit registers
	char info[43];		// 1 for a \0
	bool cache[MAX_CACHE];	// Has a cache?
	__uint8 family;		// family of CPU (see above)
	__uint8 model;
	__uint8 stepping;
	__uint16 cache_size[MAX_CACHE];	// 0 = L1, 1 = L2 ...
	__uint32 maxCPUID;	// max number of CPUID's (0 for not allowed)
	__uint32 maxExtCPUID;	// max extended cpuid op
	__uint32 capability;	// What the CPU can do :)
	struct TSS TSS;		// Only 1 TSS per-processor
	ulong TSSGDTEntry;	// GDT Entry for the TSS
#ifdef APIC
	__uint8 apic_id;		// We don't support APIC on UP
	__uint8 apic_ver;
#endif /* APIC */
};

/* struct: ljmpStr
 * desc: This struct contains the layout of a long jmp call on x86.
 */
struct ljmpStr {			// *MUST* be static
	__uint32 offset;
	__uint16 seg;
} __PACKED__;

/* typedef: flags
 * desc: This type is used for spinlocks etc, when you need to get the
 * arch-specific flags register from the cpu. Could be a struct or anything.
 */
typedef long flags;


/* macro: getReg(name, var)
 * desc: Gets the contents of the register 'name' into the variable 'var'
 */
#define getReg(y,x) asm("movl %%" y ", %0" : "=g" (x))

/* macro: setReg(name, var)
 * desc: Sets the contents of the register 'name' to the variable 'var'. USE
 * WITH CARE!!!
 */
#define setReg(y,x) asm("movl %0, %%" y : : "g" (x))

/* function: loadSeg(segName, segOff)
 * description: Loads 'segName' with the segment offset of 'segOff'.
 */
#define loadSeg(x,y) asm("movl %0, %%" x :: "g" (y))

/* function: getSeg(segName, var)
 * description: Loads the value of 'segName' into the variable 'var'.
 */
#define getSeg(x,y) asm("movl %%" x ", %0" : "=r" (y))

/* function: getCr(num, var)
 * description: Loads the contents of cr'num' into 'var'.
 */
#define getCr(y,x) asm volatile("movl %%cr" #y ", %0" : "=r" (x))

/* function: setCr(num, var)
 * description: Loads the contents of 'var' into cr'num'.
 */
#define setCr(y,x) asm volatile("movl %0, %%cr" #y :: "r" (x))


/* function: get_flags(var)
 * arg0: The variable to put EFlags into.
 * description: Loads the contents of the flags register into 'var'.
 */
static inline flags get_flags()
{
	volatile flags eflags;
	asm("pushfl\npopl %0" : "=g" (eflags));
	return eflags;
}

/* function: set_flags(var)
 * arg0: The variable to set EFlags to.
 * description: Loads the contents of 'var' into the EFlags register.
 */
static inline void set_flags(flags flags)
{
	asm("pushl %0\npopfl" :: "r" (flags));
}

/* function: void disable_local_interrupts()
 * description: Stops hardware interrupts by clearing the local CPU's
 * 'interruptable' flag (issues the instruction CLI).
 */
static inline void disable_local_interrupts()
{
	asm("cli");
}

/* function: void enable_local_interrupts()
 * description: restarts interrupts after a __local_cli() (Issues the
 * instruction STI)
 */
static inline void enable_local_interrupts()
{
	asm("sti");
}

/* function: cpuid(__uint32 op, __uint32 *eax, __uint32 *ebx, __uint32 *ecx,
 *			__uint32 *edx)
 * arg0: Operation to cpuid.
 * arg1..4: Values of the registers eax to edx on the return of the function.
 * description: Issues a CPUID to the CPU. 'op' is the operation number to issue
 * along with the CPUID instruction. eax to edx are the values that are returned
 * by the system.
 */
static inline void cpuid(__uint32 op, __uint32 *eax, __uint32 *ebx,
			__uint32 *ecx, __uint32 *edx)
{
	asm("cpuid" : "=a" (*eax), "=b" (*ebx), "=c" (*ecx), "=d" (*edx)
		: "a" (op) : "cc");
}

/* function: paging_enable()
 * desc: Enables paging
 */
static inline void paging_enable()
{
	ulong t;
	getCr(0, t);
	t |= CR0_PG;
	setCr(0, t);
}

/* function: paging_disable()
 * desc: Disables paging
 */
static inline void paging_disable()
{
	ulong t;
	getCr(0, t);
	t &= ~CR0_PG;
	setCr(0, t);
}

/* function: hlt()
 * description: Issues the HLT command to the processor. HLT causes the CPU to
 * wait until the next interrupt or reset, so calling HLT with interrupts
 * diabled is a good way to shutdown.
 */
static inline void hlt()
{
	asm("hlt");
}

/* function: interrupt(uint number)
 * arg0: The number of the interrupt to call
 * description: issues the INT instruction with the specified number as the
 * argument. (i.e. calls the interrupt 'number').
 */
static inline void interrupt(uint x)
{
	asm("int %0" :: "g" (x));
}

/* function: lidt(void *address)
 * arg0: Address of the IDT descriptor to load
 * description: Loads the IDT descriptor from location 'address'.
 */
static inline void lidt(void *address)
{
	asm("lidt (%0)" :: "r" (address));
}

/* function: lgdt(void *address)
 * arg0: Address of the GDT descriptor to load
 * description: Loads the GDT descriptor from location 'address'.
 */
static inline void lgdt(void *address)
{
	asm("lgdt (%0)" :: "r" (address));
}

/* function: lldt(__uint16 seg)
 * description: loads the LDT register with 'seg' (entry into the GDT).
 */
static inline void lldt(__uint16 seg)
{
	asm("lldt %0" : : "g" (seg));
}

/* function: ltr(__uint16 seg)
 * description: loads the task register with 'seg' (entry into the GDT).
 */
static inline void ltr(__uint16 seg)
{
	asm("ltr %0" : : "r" (seg));
}

/* function: ljmp(__uint16 cs, __uint32 offset)
 * description: Does a long jump (reloading CS with 'cs') and offset 'offset'
 */
static inline void ljmp(__uint16 cs, __uint32 offset)
{
	static struct ljmpStr t;
	t.offset = offset;
	t.seg = cs;
	asm("ljmp %0" :: "m" (t));
}

/* function: ulong getStackPointer()
 * desc: Returns the current stack pointer.
 */
static inline ulong get_stack_pointer() {
	ulong t;
	getReg("esp", t);
	return t;
}

/* function: wb_invalidate_cache()
 * description: Write back and invalidate cache (if it exists).
 */
#define wb_invalidate_cache() ({ if(curCPU->arch.cache[L1]) asm("wbinvd"); })


/* XXX On these next few we should actually check the capability flags!! */

/* function: rdtsc(highVal, lowVal)
 * arg0: The high bit of the 64-bit integer returned.
 * arg1: The low bit of the 64-bit integer returned.
 * description: Issues a RDTSC (read the clock-counter) instruction to the CPU
 * and returns the 64-bit number in 'highVal' and 'lowVal'. (Only works if the
 * kernel is compiled for a PROCESSOR > 4)
 */
#if PROCESSOR > 4
#define rdtsc(h, l) asm volatile("rdtsc" : "=a" (l), "=d" (h))
#define rdtscl(l) asm volatile("rdtsc" : "=a" (l) : : "edx")
#else
#define rdtsc(h, l) KILL_COMPILER("Kernel attempted to use pentium-specific " \
		"RDTSC instruction when PROCESSOR < 586")
#define rdtscl(l) rdtsc(1,1)
#endif

/* function: rd/wrmsr
 * arg0: the MSR to read/write to
 * arg1: the upper 32bits of the MSR to read/write to
 * arg2: the lower 32bits of the MSR to read/write from
 */
#if PROCESSOR > 4
#define rdmsr(m, h, l) asm volatile("rdmsr" : "=a"(l), "=d"(h): "c"(m))
#define wrmsr(m, h, l) asm volatile("wrmsr" : :"a"(l), "d"(h), "c"(m))
#endif /* rd/wrmsr block */

#ifdef HAVE_CYCLE_COUNT
#include <math64.h>
#define cycle_count() ({__uint64 t; __uint32 a,b; rdtsc(a,b); to64(a,b,t); t;})
#endif /* HAVE_CYCLE_COUNT */

#include <generic/kern/system.h>

#endif // __system_H
