/*
 * cpu.h - mips-specific CPU things
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 04/01/02
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 *
 */

/* Euckie kludge... */
#ifndef __mips_eflags
#define __mips_eflags

/* function: getEFlags(var)
 * arg0: The variable to put eflags into.
 * description: Loads the contents of the EFlags register into 'var'.
 */
static inline ulong getEFlags()
{
	volatile ulong eflags;
	asm("pushfl\npopl %0" : "=g" (eflags));
	return eflags;
}

/* function: setEFlags(var)
 * arg0: The variable to set EFlags to.
 * description: Loads the contents of 'var' into the EFlags register.
 */
static inline void setEFlags(ulong eflags)
{
	asm("pushl %0\npopfl" :: "r" (eflags));
}

/* function: void __local_cli()
 * description: Stops hardware interrupts by clearing the local CPU's
 * 'interruptable' flag (issues the instruction CLI).
 */
static inline void __local_cli()
{
	asm("cli");
}

/* function: __local_sti()
 * description: restarts interrupts after a __local_cli() (Issues the
 * instruction STI)
 */
static inline void __local_sti()
{
	asm("sti");
}
#endif /* !__mips_eflags */

#ifndef __ONLY_WANT_EFLAGS
#ifndef __mips_cpu_H
#define __mips_cpu_H

/* types of reboots */
#define REBOOT_COLD		0
#define REBOOT_WARM		0

enum {
	L1 = 0,
	L2,
	L3,
	/* L4? ;) */
	MAX_CACHE		/* 3 */
};

/* struct: archCPU
 * description: Contains all of the info about a specific CPU in the box.
 */
struct archCPU {
	bool cache[MAX_CACHE];	// Has a cache?
	__uint8 model;
	__uint16 cache_size[MAX_CACHE];	// 0 = L1, 1 = L2 ...
	__uint32 capability;	// What the CPU can do :)
};

struct CPU;			/* Forward declaration... */
void dumpRegs(struct allRegs *);
void dumpStack(ulong *);
void CPUDetectMhz(void);
void wbinvd(struct CPU *);
void CPUDetectCPUID(struct CPU *);
void CPUDetectCapability(struct CPU *);
void CPUDetectMax(struct CPU *, bool extended);
errRet CPUDetect(ulong);
errRet CPUSetup(ulong);
void reboot(__uint32);
void halt();

/* function: interrupt(uchar number)
 * arg0: The number of the interrupt to call
 * description: issues the INT instruction with the specified number as the
 * argument. (i.e. calls the interrupt 'number').
 */
static inline void interrupt(uchar x)
{
	asm("int %0" :: "g" (x));
}

#if 0
/* struct: ljmpStr
 * desc: This struct contains the layout of a long jmp call on mips
 */
struct ljmpStr {			// *MUST* be static
	__uint32 offset;
	__uint16 seg;
} __PACKED__;

/* function: ljmp(__uint16 cs, __uint32 offset)
 * description: Does a long jump (reloading CS with 'cs') and offset 'offset'
 */
static inline void ljmp(__uint16 cs, __uint32 offset)
{
	static struct ljmpStr t;
	t.offset = offset;
	t.seg = cs;
	asm("ljmp %0" :: "m" (t));
}
#endif /* 0 */

/* macro: getReg(name, var)
 * desc: Gets the contents of the register 'name' into the variable 'var'
 */
#define getReg(y,x) asm("movl %%" y ", %0" : "=g" (x))

/* macro: setReg(name, var)
 * desc: Sets the contents of the register 'name' to the variable 'var'. USE
 * WITH CARE!!!
 */
#define setReg(y,x) asm("movl %0, %%" y : : "g" (x))

/* function: ulong getStackPointer()
 * desc: Returns the current stack pointer.
 */
static inline ulong getStackPointer() {
	ulong t;
	getReg("esp", t);
	return t;
}

/* function: void wbinvd()
 * description: Write back and invalidate cache (if it exists).
 */
#define wbinvd() if(curCPU->arch.cache[L1]) { asm("wbinvd"); }

/* function: getCr(num, var)
 * description: Loads the contents of cr'num' into 'var'.
 */
#define getCr(y,x) asm volatile("movl %%cr" #y ", %0" : "=r" (x))

/* function: setCr(num, var)
 * description: Loads the contents of 'var' into cr'num'.
 */
#define setCr(y,x) asm volatile("movl %0, %%cr" #y :: "r" (x))

#define disablePaging() do { \
	__uint32 __tmp; \
	getCr(0, __tmp); \
	btc(&__tmp, CR0_PG); \
	setCr(0, __tmp); \
} while (0)

#define enablePaging() do { \
	__uint32 __tmp; \
	getCr(0, __tmp); \
	bts(&__tmp, CR0_PG); \
	setCr(0, __tmp); \
} while (0)

/* function: get_cycle()
 * description: Returns the 32-bit number from the count reg in cp0
 */
static inline ulong get_cycle()
{
	ulong t;
	asm("mfc0 %0, $9" : "=r" (t));
	return t;
}

/* function: set_cycle(ulong number)
 * description: Sets the cycle count to 'number'
 */
static inline void set_cycle(ulong number)
{
	asm("mtc0 %0, $9" : : "r" (number));
}

#ifdef HAVE_CYCLE_COUNT
#define cycle_count() get_cycle()
#endif /* HAVE_CYCLE_COUNT */

#include <generic/kern/cpu.h>

#endif // __mips_cpu_H
#endif /* !__ONLY_WANT_EFLAGS */
