/*
 * cpu.h - x86-64 specific CPU things
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 04/01/02
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 17/04/01 bodnar: Changed all #define assembly into static inline assembly.
 * 01/06/2001 Mark: Ported to x86-64 arch
 */

#ifndef __x86_cpu_H
#define __x86_cpu_H

#include <desc.h>
#include <task.h>

/* struct: archCPU
 * description: Contains all of the info about a specific CPU in the box.
 */
struct archCPU {
	char cpuid[13];
	char info[43];		// 1 for a \0
	__uint8 version;	// version of CPU (see above)
	__uint32 maxCPUID;	// max number of CPUID's (0 for not allowed)
	__uint32 capability;	// What the CPU can do :)
	struct TSS TSS;		// Only 1 TSS per-processor
	ulong TSSGDTEntry;	// GDT Entry for the TSS
};

#include <generic/kern/cpu.h>

struct CPU;
handleInt(dumpRegs);
void wbinvd(struct CPU *);
void CPUDetectCPUID(struct CPU *);
void CPUDetectCapability(struct CPU *);
void CPUDetectMax(struct CPU *);
errRet CPUDetect(__uint32);
errRet CPUSetup(ulong);

/* function: cpuid(__uint32 op, __uint32 *eax, __uint32 *ebx, __uint32 *ecx,
 *			__uint32 *edx)
 * arg0: Operation to cpuid.
 * arg1..4: Values of the registers eax to edx on the return of the function.
 * description: Issues a CPUID to the CPU. 'op' is the operation number to issue
 * along with the CPUID instruction. eax to edx are the values that are returned
 * by the system.
 */
// Shamlessly taken from linux ;)
static inline void cpuid(__uint32 op, __uint32 *eax, __uint32 *ebx,
			__uint32 *ecx, __uint32 *edx)
{
	asm("cpuid" : "=a" (*eax), "=b" (*ebx), "=c" (*ecx), "=d" (*edx)
		: "a" (op) : "cc");
}

/* function: void __local_cli()
 * description: Stops hardware interrupts by clearing the local CPU's
 * 'interruptable' flag (issues the instruction CLI).
 */
static inline void __local_cli()
{
	asm("cli");
}

/* function: __local_sti()
 * description: restarts interrupts after a __local_cli() (Issues the
 * instruction STI)
 */
static inline void __local_sti()
{
	asm("sti");
}

/* function: hlt()
 * description: Issues the HLT command to the processor. HLT causes the CPU to
 * wait until the next interrupt or reset, so calling HLT with interrupts
 * diabled is a good way to shutdown.
 */
static inline void hlt()
{
	asm("hlt");
}

/* function: interrupt(uchar number)
 * arg0: The number of the interrupt to call
 * description: issues the INT instruction with the specified number as the
 * argument. (i.e. calls the interrupt 'number').
 */
static inline void interrupt(uchar x)
{
	asm("int %0" :: "g" (x));
}

/* function: lidt(void *address)
 * arg0: Address of the IDT descriptor to load
 * description: Loads the IDT descriptor from location 'address'.
 */
static inline void lidt(void *address)
{
	asm("lidt (%0)" :: "r" (address));
}

/* function: lgdt(void *address)
 * arg0: Address of the GDT descriptor to load
 * description: Loads the GDT descriptor from location 'address'.
 */
static inline void lgdt(void *address)
{
	asm("lgdt (%0)" :: "r" (address));
}

/* function: loadSeg(segName, segOff)
 * description: Loads 'segName' with the segment offset of 'segOff'.
 */
#define loadSeg(x,y) asm("movl %0, %%" x :: "r" (y))

/* function: getSeg(segName, var)
 * description: Loads the value of 'segName' into the variable 'var'.
 */
#define getSeg(x,y) asm("movl %%" x ", %0" : "=r" (y))

/* function: getEFlags(var)
 * arg0: The variable to put eflags into.
 * description: Loads the contents of the EFlags register into 'var'.
 */
static inline ulong getEFlags()
{
	ulong eflags;
	asm("pushfq\npopq %0" : "=g" (eflags));
	return eflags;
}

/* function: setEFlags(var)
 * arg0: The variable to set EFlags to.
 * description: Loads the contents of 'var' into the EFlags register.
 */
static inline void setEFlags(ulong eflags)
{
	asm("pushq %0\npopfq" :: "g" (eflags));
}

/* function: clearNT()
 * description: Clears the Nested Task flag in the EFlags register
 */
static inline void clearNT()
{
	ulong a = getEFlags();
	a &= 0xFFFFFFFFFFFFBFFF;
	setEFlags(a);
}

/* function: ljmp(__uint16 cs, __uint32 offset)
 * description: Does a long jump (reloading CS with 'cs') and offset 'offset'
 */
static inline void ljmp(__uint16 cs, __uint32 offset)
{
	static struct {		// *MUST* be static
		__uint32 offset;
		__uint16 seg;
	} __PACKED__t;
	t.offset = offset;
	t.seg = cs;
	asm("ljmp %0" :: "m" (t));
}

/* function: lldt(__uint16 seg)
 * description: loads the LDT register with 'seg' (entry into the GDT).
 */
static inline void lldt(__uint16 seg)
{
	asm("lldt %0" : : "g" (seg));
}

/* function: ltr(__uint16 seg)
 * description: loads the task register with 'seg' (entry into the GDT).
 */
static inline void ltr(__uint16 seg)
{
	asm("ltr %0" : : "r" (seg));
}

/* macro: getReg(name, var)
 * desc: Gets the contents of the register 'name' into the variable 'var'
 */
#define getReg(y,x) asm("movl %%" y ", %0" : "=g" (x))

/* macro: setReg(name, var)
 * desc: Sets the contents of the register 'name' to the variable 'var'. USE
 * WITH CARE!!!
 */
#define setReg(y,x) asm("movl %0, %%" y : : "g" (x))

/* function: ulong getStackPointer()
 * desc: Returns the current stack pointer.
 */
static inline void getStackPointer() {
	ulong t;
	getReg("rsp", t);
	return t;
}

/* function: getCr(num, var)
 * description: Loads the contents of cr'num' into 'var'.
 */
#define getCr(y,x) asm("movq %%cr" #y ", %0" : "=r" (x))

/* function: setCr(num, var)
 * description: Loads the contents of 'var' into cr'num'.
 */
#define setCr(y,x) asm("movq %0, %%cr" #y :: "r" (x))

/* function: rdtsc(highVal, lowVal)
 * arg0: The high bit of the 64-bit integer returned.
 * arg1: The low bit of the 64-bit integer returned.
 * description: Issues a RDTSC (read the clock-counter) instruction to the CPU
 * and returns the 64-bit number in 'highVal' and 'lowVal'.
 */
#define rdtsc(h, l) asm volatile("rdtsc" : "=a" (l), "=d" (h));
#define rdtscl(l) asm volatile("rdtsc" : "=a" (l) : : "edx");

#include <generic/kern/cpu.h>

#endif // __cpu_H
// END
