/*
 * desc.h - Descriptor table file for x86-64
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 04/01/02
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 * This file contains general defines and structs to setup an entry in the GDT,
 * IDT or LDT
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 25/09/00 Mark: I hacked the startInt macro etc to do some neater stuff than
 *	it did before. Also added some new ones and made it safer/better than
 *	before. I finally now understand interrupts (hahaha, I think)
 * 31/10/00 Mark: I re-did all of them, to take up less space :)
 * 15/05/01 Mark: Minor fixups of some of the macros to make them neater
 * 01/06/01 Mark: Ported to x86-64
 */

#ifndef __desc_H
#define __desc_H

////////////////////////////////////////////////////////////////////////////////
// IDT Stuff
errRet setGate(void *, ulong, char, char);
void unsetGate(ulong);
initFunc IDTInit();
extern void INTHandler();

/* struct: desc
 * description: A descriptor for stuff like lgdt() and lidt()
 */
struct desc {
	__uint16 limit;
	memory base;
} __PACKED__;

/* struct: gateDesc
 * description: Structure of a gate on the x86-64
 */
struct gateDesc {
	__uint16 offsetLow,
		 TSS;
	__uint8  oldDWordCount,		// Must be 0
		 type		:5,
		 DPL		:2,
		 present	:1;
	__uint16 offsetMid;
	__uint32 offsetHigh;
	__uint32 reserved;		// Must be 0
} __PACKED__;

#define CALL_GATE 12
#define INT_GATE 14
#define TRAP_GATE 15

/* function: setIntGate(handler, interrupt, DPL)
 * arg0: The address of the handler for the interrupt.
 * arg1: The interrupt number ot catch.
 * arg2: The Privilage level for the code in the interrupt to be executed in.
 * description: This function sets a interrupt gate with the specified
 * arguments.
 */
#define setIntGate(x,y,z) setGate(x,y,z, INT_GATE)

/* function: setCallGate(handler, interrupt, DPL)
 * arg0: The address of the handler for the interrupt.
 * arg1: The interrupt number ot catch.
 * arg2: The Privilage level for the code in the interrupt to be executed in.
 * description: This function sets a call gate with the specified arguments.
 */
#define setCallGate(x,y,z) setGate(x,y,z, CALL_GATE)
 
/* function: setTrapGate(handler, interrupt, DPL)
 * arg0: The address of the handler for the interrupt.
 * arg1: The interrupt number ot catch.
 * arg2: The Privilage level for the code in the interrupt to be executed in.
 * description: This function sets a trap gate with the specified arguments.
 */
#define setTrapGate(x,y,z) setGate(x,y,z, TRAP_GATE)

/* struct: allRegs
 * description: A struct with all the registers pushed onto the stack by an
 * interrupt.
 */
struct allRegs {
	ulong r15;		// Pushed by interrupt handler
	ulong r14;
	ulong r13;
	ulong r12;
	ulong r11;
	ulong r10;
	ulong r9;
	ulong r8;
	ulong rbp;
	ulong rsi;
	ulong rdi;
	ulong rdx;
	ulong rcx;
	ulong num;
	ulong rbx;
	ulong rax;

	ulong rip;		// Pushed by CPU
	ulong cs;
	ulong eflags;
	ulong rsp;
	ulong ss;
} __PACKED__;

/* macro: startIntr(functionName, number, calls)
 * arg0: Name of the function.
 * arg1: Interrupt number.
 * arg2: The function that it calls to handle the interrupt.
 * description: Makes a basic interrupt handler called 'functionName'. This is
 * the name you should pass to set*Gate(). It the calls off 'calls'. IT IS NOT
 * RECOMENDED THAT YOU USE THIS FUNCTION FOR YOUR INTERRUPT HANDLERS UNLESS YOU
 * KNOW WHAT YOU'RE DOING, try to use startInt() instead, it's a lot easier :)
 */
#define startIntr(real, num, dif)				\
	extern void real();					\
	asm(	".globl " #real	"			\n"	\
		#real ":				\n"	\
		"	pushq %rax			\n"	\
		"	pushq %rbx			\n"	\
		"	movl $" #dif ", %eax		\n"	\
		"	movl $(" STR(num) "), %ebx	\n"	\
		"	jmp INTHandler			\n")

/* macro: handleInt(functionName)
 * arg0: Name of the function to create
 * description: Like signalHnd(), it creates a interrupt handler. ATM, it's the
 * same as void functionName(struct allRegs *r), but this could
 * change.
 */
#define handleInt(real) void asmLinkage real(struct allRegs *r)

/* typedef: intFunc
 * description: A type that will point to a function of type handleInt().
 */
typedef void (*intFunc)(struct allRegs *);

/* macro: startInt(functionName, number)
 * arg0: The name of the function.
 * arg1: The interrupt number.
 * description: A user-friendly function for making an interrupt handler. 99% of
 * cases should use this unless you've RTFS'd and understand it.
 */
#define startInt(real, num) startIntr(real, num, _##real);handleInt(_##real)

////////////////////////////////////////////////////////////////////////////////
// GDT Stuff
initFunc GDTInit();
__uint16 GDTAddEntry(__uint32, __uint32, bool, bool, __uint8, uchar, bool);
void GDTDelEntry(__uint16);

// Defines
#define SEG_DATA 0
#define SEG_CODE 8

#define SEG_GROW_DOWN 4

#define SEG_RO 0
#define SEG_RW 2

#define SEG_ACCESSED 1			// Set by the CPU itself

extern union segDesc *GDTArea;

// Some values
extern __uint16 kernData, kernCode, userData, userCode;

////////////////////////////////////////////////////////////////////////////////
// LDT Stuff
void LDTAddEntry(union segDesc *, ulong, ulong, uchar, uchar);

////////////////////////////////////////////////////////////////////////////////
// LDT and GDT shared stuff

/* struct: segDesc
 * description: A union that describes each segment GDT (and LDT).
 */
union segDesc {
	struct {
		ushort	limitLow,
			baseLow;
		uchar	baseMed,
			type		:4,
			descType	:1,
			DPL		:2,
			present		:1,
			limitHigh	:4,
			ZealOS		:1,
			not_used	:1,
			segmentSize	:1,
			granularity	:1,
			baseHigh;
	} norm __PACKED__;
	__uint64 dummy;
};

#endif	// __desc_H
