#!/usr/bin/perl -w
# bug-reporting system for the ZealOS module loading system. Copyright Mark
# Zealey 2001, released under the GPL.
# Last update: 10/10/01
# Arg1: (optional) name of smtp server

use strict;
use Cwd;

my $zealos_bug = 'zealos-bugs@zealos.org';
my $ver = '0.1';
my $netsmtp;			# false
my $quesnum = 1;
my $smtp = shift;
my ($sev, $bugreport, $confpath, $conf, $send, $email, $sub, $mod);
sub ques;

BEGIN {
	$netsmtp = 0;
	eval 'use Net::SMTP';
	if($@) {
		print "Can't load Net::SMTP, falling back to piping thru 'mail'\n"
	} else {		# OK, we have Net::SMTP
		$netsmtp = 1;
	}
}

print <<EOF;
Welcome to the ZealOS bug-reporter. This is aiming to be a relativly pain-free
process, so I'll start off by asking you a few questions.

EOF

ques '"Please type in the URL of your SMTP server (press enter to use default)',
		'SMTP server' , \$smtp if $netsmtp and !$smtp;
ques 'What is your email address?', 'Email address', \$email;
ques 'What module is this bug in (loader for the base-module loader)?',
		'Module', \$mod;
ques 'Please write a short description (for the subject line)', 'Subject',
		\$sub;
ques <<EOF, 'Severity', \$sev, sub {return 1 if shift =~ /^[0-4]$/; return 0};
What is the severity of your bug report? Different levels are:
        0) Feature request

        1) A small, non-critical failure that. Most bugs are in this category

        2) This bug makes this module unstable for a stable release.

        3) A dangerous bug which makes the module unstable, causes data loss or
           introduces a security hole.

        4) A critical bug. Makes the whole system unstable, causes serious
           data-loss etc...
EOF

ques <<EOF, 'Config file path', \$confpath, \&checkpath;
The config file is needed to check your configuration so that we can try to
reproduce the problem. Please type in the path of your config file, or press
<enter> to check the current directory.
EOF

chomp $conf;
my $report = <<EOF;
Bug report from $email using zealos bugreport version $ver on
@{[scalar localtime]}.

Module: $mod

Report:
*** Insert bug report here ***

config:
$conf

End of message
EOF

my $tmp_file = &mkTmp;
open F, ">$tmp_file" or die "Can't open temp file for writing: $!";
print F $report;
close F;

my $ed = $ENV{EDITOR} || 'vim';
my $t;
ques "What editor do you want to use? (enter = $ed)", 'Editor', \$t,
		sub{return !system(shift || $ed, $tmp_file)};

ques '', 'Send bug report? (y/n)', \$send, sub { return shift =~ /^[yn]/i};

if($send !~ /^y/i) {
	print "Aborting bug report.\n";
	exit 1;
}

&slurp($tmp_file, \$report);

my $send_data = <<EOF;
To: $zealos_bug
From: $email
Subject: [BUG:$sev] $mod: $sub

$report
EOF

&smtp_send || &mail_send || do {print "Send failed. Please try manually :(\n";
				unlink $tmp_file;
				exit 1};

print "Bug report sent successfully. Thanks!\n";
unlink $tmp_file;
exit;

sub smtp_send {
	return 0 unless $netsmtp;

	my $s = new Net::SMTP $smtp or do {print STDERR "Can't open up an smtp".
				" connection to $smtp: $!\n"; return 0};
	$s->mail($email);
	$s->to($zealos_bug);
	$s->data;
	$s->datasend($send_data);
	$s->dataend;
	$s->quit;
	return 1;
}

sub mail_send {
	open P, "|mail $zealos_bug" or do {print STDERR "Can't open a pipe to ".
						"mail: $!\n"; return 0};
	print P $send_data;
	close P;
	return 1;
}

sub ques {
	my ($ques, $prompt, $var, $sub) = @_;
	print $quesnum++, ') ', $ques, "\n" if $ques;
	$sub ||= sub {return shift};

	print "$prompt: ";
	chomp($$var = <STDIN>);
	while(!&$sub($$var)) {
		print "Sorry, that is not a valid answer. Please re-enter\n";
		print "$prompt: ";
		chomp($$var = <STDIN>);
	}
	print "\n";
}

sub checkpath {
	my $path = shift;

	$path =~ s/^\s+//g;
	$path =~ s/\s+$//g;

	$path = cwd.$path if $path !~ m!^[./~]!;
	if($path =~ m!config[^/]*$! and -e $path) {
		return &slurp($path, \$conf);
	}
	$path .= '/' if $path !~ m!/$!;
	return &slurp("$path/config", \$conf) if -e "$path/config";
	return &slurp("$path/config.bak", \$conf) if -e "$path/config.bak";
}

sub slurp {
	my ($file, $str) = @_;
	open F, $file or do {print STDERR "Can't open $file: $!\n", return 0};
	local $/ = undef;
	$$str = <F>;
	close F;

	return 1;
}

sub mkTmp {
	my @chars = ('a'..'z', 'A'..'Z', 0..9);
	my $fn;
	
	do {
		$fn = '/tmp/';
		map {$fn .= $chars[rand @chars]} 0..(4 + rand 4);
	} while -e $fn;

	return $fn;
}
