#!/usr/bin/perl -w
# Last update: 22/09/01
# Copyright Mark Zealey, 2001. A part of the Zealos project.
# Released under the GPL.
use strict;
use Term::Complete;
require 'tools/Parse.pm';

my $conf_file = 'config';
my $old_conf_file = 'config.bak';
my $defaults = 'defaults';
my $fast = 0;
my @out;

my %types = (
	no_module => [qw/Y N D/],
	yes_no => [qw/Y N/],
	normal => [qw/Y N D M MD/],
	no_debug => [qw/Y N M/],
	number => [],
	string => [],
);

my %config = Parse::Config($old_conf_file);
my %defaults = Parse::Config($defaults);

for(grep /^-/, @ARGV) {
	/^-f(ast)?$/ and $fast = 1
}

my $out = <<OUT;
# This config file was automatically generated by $0 at @{[scalar localtime]}.
# This file will be moved to 'config.bak' next time you 'make config'.
OUT

# A hash table with details of all the stuff that's been processed.
my %options;

&do_file($_) for grep !/^-/, @ARGV;

open C, ">$conf_file" or &error("Can't open $conf_file for writing: $!");
print C $out;
close C;
exit;

sub do_file {
	my $file = shift;

	open F, $file or &error("Can't open $file: $!");
	my $source;
	{
		local $/ = undef;
		$source = <F>;
	}
	close F;

	&do_block([&tokenize($source)]);
}

sub do_block {
	my $token_list = shift;
	my @tokens = @$token_list;

	while(defined($_ = shift @tokens)) {
#		print "Processing token \"$_\"\n";
		if($_ eq 'write') {	# write <device> <text>;
			my $thing = shift @tokens;
			my $text = &interpolate(scalar &getline(\@tokens));

			if($thing eq 'screen') {
				print $text;
			} elsif($thing eq 'config') {
				$out .= $text;
			} else {
				&error("$_: Device '$thing' not known about.");
			}
		} elsif($_ eq 'option') {	# option <name> {...}
			my $name = shift @tokens;
			my $ask = !$fast;
			my $o = $options{$name} = {};
			my @block = &getblock(\@tokens);
			&error("No name for option block") if $name eq '{';
			&error("Option block is invalid") if shift @block ne'{';

			# We now process <name> = <value>; lines:
			my @line;
			while(@line = &getline(\@block) and @block > 0) {
				if($line[1] ne '=') {
					&warn("Line ``@line'' is invalid");
					next;
				}
				$o->{$line[0]} = &interpolate(
							"@line[2..$#line]");
#				print "\$o->{$line[0]} = .",
#						$o->{$line[0]}, ".\n"
			}

			$o->{type} ||= 'normal';
			if($o->{type} ne 'select') {
				if(!exists $types{$o->{type}}) {
					&warn("Type $o->{type} is not valid");
					next;
				}
				$o->{choices} = $types{$o->{type}}
			} elsif($o->{choices}) {
				$o->{choices} = [split /\s*,\s*/,$o->{choices}];
			} else {
				&error("`choices' does not exist for option $name");
			}

			if(exists $config{$name}) {
				my $r = $config{$name};
				$r = &is_in($o->{choices}, $config{$name}) if
							@{$o->{choices}};
				$o->{default} = $r if defined $r
			} else {		# New option
				$ask = 1
			}

			if(exists $o->{default}) {
				my $r = &is_in($o->{choices}, $o->{default});
				$o->{default} = $r if defined $r
			}

			if(!$ask and exists $o->{default}) {
				$o->{choice} = $o->{default}
			} else {
				$o->{choice} = &ask($name, $o->{question},
						$o->{help}, $o->{choices},
						$o->{default}, $o->{type})
			}

			$out .= "$name = $o->{choice}\n";
		} elsif($_ eq 'include') {	# include <file>;
			&do_file(&interpolate(&getline(\@tokens)));
		} elsif($_ eq 'if') {		# if (...) {...}
			my $res = &do_expr([&getblock(\@tokens, '(', ')')]);
			my @tokes = &getblock(\@tokens);
			if($res) {
				&error("if block does not begin with a {") if
							shift @tokes ne '{';
				&error("if block does not end with a }") if
							pop @tokes ne '}';
#				print "@tokes\n";
				&do_block(\@tokes);
			}
		} elsif($_ eq 'eval') {		# eval <perl code>;
			eval &interpolate(scalar &getline(\@tokens));
		} else {
			&error("Command '$_' is not known about");
		}
	}
}

# returns 1 or 0 depending on the result of a boolean expression
sub do_expr {
	my $token = shift;
	my @tokens = @$token;

	# Sanity checking
	$tokens[0] eq '(' and shift @tokens;
	$tokens[-1] eq ')' and pop @tokens;

	my ($t, @c);

	while(defined($t = shift @tokens)) {
		if($t eq '(') {
			unshift @tokens, $t;
			unshift @tokens, &do_expr([&getblock(\@tokens, '(',
									')')]);
			next;
		}

		push @c, $t;
#print "@c\n";
		if(@c == 3) {
			&interpolate(\$c[0], \$c[2]);
#			print "Arith: @c: ";

			if($c[1] eq '==') {
				push @tokens, $c[0] eq $c[2]
			} elsif($c[1] eq '!=') {
				push @tokens, $c[0] ne $c[2]
			} elsif($c[1] eq '&&') {
				push @tokens, $c[0] && $c[2]
			} elsif($c[1] eq '||') {
				push @tokens, $c[0] || $c[2]
			} else {
				&warn("$c[1] is not a valid operator");
			}
#			print "$tokens[0]\n";
			@c = ();
		}
	}

	if(@c != 1) {
		&error("Error: @c is not 1 arg only");
	}
	return $c[0];
}

# Die because of an error
sub error {
	print STDERR "Error: @_\n";
	exit 1;
}

# Report internal error
sub internal {
	print STDERR "Internal error: @_\n";
	die;
}

# Warn of an error
sub warn {
	print STDERR "Warning: @_\n"
}

# Tokenize it (don't ask...):
sub tokenize {
	my $source = "@_";
	my @tokens;

	while($source =~ /\G\s*/cg) {
		if($source =~ /\G<<(\w+)/cg) {		# Multi-line text
			my $t = $1;
			my $y = pos $source;
			# Thanks to japhy who made this work!
			substr($source, $y) =~ s/([^\n]*\n)(.*?)^$t$/$1/sm;
			push @tokens, $2;
			pos($source) = $y;
		} elsif($source =~ /\G(['"])/cg) {	# Text
			my $m = $1;
			$source =~ /\G([^\\$m]*(?:\\.[^\\$m]*)*)$m\s*/cg;
			push @tokens, $1 if defined $1;
		} elsif($source =~ /\G#/cg) {		# comment
			$source =~ /\G[^\n]*\n/cg;
		} elsif($source =~ /\G([!=&|]+)/cg) {	# Operation
			push @tokens, $1;
		} elsif($source =~ /\G([][\w\/.\$-]+)/cg) {# Plain text
			push @tokens, $1 if defined $1;
		} elsif($source =~ /\G([^-\/.\$\w])/cg) {# statment start/end
			push @tokens, $1;
		}
	}

	return @tokens;
}

# Returns all the tokens up to the ';'
sub getline {
	my $a = shift;
	my @elem;
	my $t;

	push @elem, $t while defined($t = shift @$a) and $t !~ /^[;}]$/;
	defined $t and $t eq '}' and unshift @$a, $t;

#print "returned .@elem.\n";
	return @elem if wantarray;
	return "@elem";
}

# Returns all the tokens up to the $close
sub getblock {
	my ($a, $open, $close) = @_;
	$open ||= '{';
	$close ||= '}';
	my $depth = 0;
	my $v = 0;
	my (@block, $t);

	while(defined($t = shift @$a)) {
		push @block, $t;
		$t eq $open and do {
			$depth++;
			$v = 1
		};

		$t eq $close and do {
			$depth--;
			last if $v and !$depth
		}
	}

	return @block if wantarray;
	return "@block";
}

# Interpolates all the $variables in the strings given
sub interpolate {
	my $t;

	for(@_) {
		if(ref and ref ne 'SCALAR') {
			&internal("argument passed to me was not a scalar ref");
		}

		$t = ref() ? $$_ : $_;
		$t =~ s/(\\.)/"qq!$1!"/eeg;
		$t =~ s/\$OPT\[(\w+?)]/(exists $options{$1}) ?
						$options{$1}{choice} :
						'N'/eg;
		$t =~ s/\$DEF\[(\w+?)]/exists $defaults{$1}?$defaults{$1}:''/eg;
#		print "$t\n";
		ref() ? $$_ : $_ = $t;
	}

	return $t;
}

# Looks for an element in the array that matches the element specified and
# returns it in the correct case.
sub is_in {
	my ($a, $e) = @_;
	for(@$a) {
		return $_ if lc $_ eq lc $e;
	}

	return undef;
}

# Ask the user a question and make sure we get a valid response
sub ask {
	my ($name, $q, $help, $choices, $def, $type) = @_;
	my $in;

	$def = $choices->[0] if @$choices and !defined $def;
	$help ||= 'Sorry, no help is avalable for this option';

	while(1) {
		my $prompt = "[$name] $q ";
		$prompt .= "(" . join('/', @$choices) . ") " if @$choices;
		$prompt .= "[$def] " if defined $def;

		if($type eq "select") {
			$in = Complete($prompt, @$choices);
		} else {
			print $prompt;
			chomp($in = <STDIN>);
		}

		if(lc $in eq 'h' or lc $in eq '?') {
			print $help, "\n";
			next
		}

		return $def if $in eq '' and defined $def;	# default

		if($choices) {
			# We have a valid response
			my $t = &is_in($choices, $in);
			return $t if defined $t;
		}

		if($type eq 'string') {
			return $in
		} elsif($type eq 'number') {
			return $in if $in =~ /^\d+$/
		}

		print "Invalid response, please try again\n";
	}
}
