/*
 * cpu.c - Non-generic CPU handling functions for the x86-64 arch
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 29/06/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 24/06/01 Mark: Fixed up stuff for the new archCPU stuff
 * 26/06/01 Mark: Fixed in_lock to not __local_sti before we have the interrupt
 *	table set up.
 */

#include <common.h>
#include <cpu.h>
#include <desc.h>
#include <printk.h>
#include <stdio.h>
#include <signals.h>
#include <irq.h>
#include <kmalloc.h>
#include <atomic.h>


#ifndef __SMP__
/* We initalize this value to 1 at boot-up because before we load the new GDT
 * and IDT tables, we don't want to __local_sti() on unlock(). When we've loaded
 * the IDT, we atomicDec() this so that locks really work. - Mark
 */
atomic in_lock = 1;
#endif

handleInt(exception)
{
	dumpRegs(r);
	printk("That's all folks, maybe someone wants to debug? ;-)\n");
	__local_cli();			// Halt
	while(1);
}

/* function: void dumpRegs(struct allRegs *)
 * description: This function is the end point of most exceptions, dumps all the
 * registers as the function name suggests, and then hangs the box. Not a very
 * nice way to go.
 */
void dumpRegs(struct allRegs *r)
{
	ulong t[2];
	uint v[1];

	// RIP.. :)
	printk("ERROR: %u\n", (uint)r->num);
	printk("CS: %4x RIP: %16lx\n", (uint)r->cs, r->rip);
	printk("SS: %4x RSP: %16lx\n", (uint)r->ss, r->rsp);
	printk("EFLAGS: %16lx\n", r->eflags);
	printk("RAX: %16lx RBX: %16lx RCX: %16lx\n", r->rax, r->rbx, r->rcx);
	printk("RDX: %16lx RSI: %16lx RDI: %16lx\n", r->rdx, r->rsi, r->rdi);

	getSeg("ds", v[0]);
	getSeg("es", v[1]);
	printk("DS: %4x ES: %4X\n", v[0], v[1]);

	printk("RBP: %16lx R08: %16lx R09: %16lx\n", r->rbp, r->r8, r->r9);
	printk("R10: %16lx R11: %16lx R12: %16lx\n", r->r10, r->r11, r->r12);
	printk("R13: %16lx R14: %16lx R15: %16lx\n", r->r13, r->r14, r->r15);

	getCr(0, t[0]);
	getCr(2, t[1]);
	getCr(3, t[2]);
	printk("CR0: %16lx CR2: %16lx CR3: %16lx\n", t[0], t[1], t[2]);

	getCr(4, t[0]);
	printk("CR4: %16lx\n", t[0]);
	printk("That's all folks, maybe someone wants to debug? ;-)\n");
}

/* function: void CPUDetect(__uint32 num)
 * description: Detects the specified CPU and adds it into the current box.
 */
errRet CPUDetect(__uint32 num)
{
	struct CPU *CPU;
	
	CPU = kmalloc(sizeof(struct CPU));
	thisBox.CPU[num] = CPU;

	kmemset((char *)CPU, 0, sizeof(struct CPU));
	v = (__uint32 *)CPU->arch.cpuid;

	/* Get CPUID */
	cpuid(0, &v[0], &v[1], &v[2], &v[3]);	// Do a standard cpuid
	CPU->arch.cpuid[12] = '\0';		// Null-terminate it

	/* FIXME: Extend this... */

	return SUCCESS;
}

/* function: void CPUDetectMax(struct CPU *p)
 * description: Detects the max CPUID command.
 */
void CPUDetectMax(struct CPU *p)
{
	__uint32 d;

	if(!p->arch.maxCPUID)		// 0 for not allowed to CPUID
		return;

	cpuid(0x80000000, &p->arch.maxCPUID, &d, &d, &d);
}

/* function: void CPUDetectCapability(struct CPU *p)
 * description: Gets the capability of the CPU.
 */
void CPUDetectCapability(struct CPU *p)
{
	__uint32 d;
	
	if(p->arch.maxCPUID < 0x80000001)
		return;

	cpuid(0x80000001, &d, &d, &d, &p->arch.capability);
}

/* function: void CPUDetectCPUID(struct CPU *p)
 * description: Detects the CPU by using CPUID.
 */
void CPUDetectCPUID(struct CPU *p)
{
	__uint32 *v = (__uint32 *) p->arch.info;

	if(p->arch.maxCPUID < 0x80000004)   // Return if it doesn't support it
		return;

	// Grab the name etc
	cpuid(0x80000002, &v[0], &v[1], &v[2], &v[3]);
	cpuid(0x80000003, &v[4], &v[5], &v[6], &v[7]);
	cpuid(0x80000004, &v[8], &v[9], &v[10], &v[11]);

	p->arch.info[48] = '\0';		// we add a \0
}

errRet CPUSetup(ulong num)
{
	if(CPUDetect(num) == SUCCESS)
		CPUDelaySetup(thisBox.CPU[num]);

	return SUCCESS;
}

EXPORT_SYMBOL(dumpRegs);
