/*
 * video.c - Video handling routines for the x86-64
 *
 * Added to the Zealos project and (C) by Mark Zealey
 *
 * Current maintainer: Mark Zealey <mark@itsolve.co.uk>
 * Status: Maintained
 * Last update: 29/06/01
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description:
 *
 * Notes/Fixme's:
 *
 * Changelog:
 * 28/09/00 Mark: changed the sloooo IO routines to x and y variables. This
 *	is cos of a problem that I suspect the IO is involved in. We will prolly
 *	want to put the IO routines in if someone wants a little blinky cursor
 *	:)
 * 16/12/00 Mark: added the cursor stuff back in, but keep cached values so it
 *	moves the cursor w/o the lag
 * 28/12/00 Mark: Changed it all to __uint16 in an attempt to solve a problem
 *	with kernel memory getting overwritten. Still trying to solve this
 *	prob...
 * 30/12/00 mark: fixed this problem - bootloader probs.
 */


#include <common.h>
#include <video.h>
#include <io.h>
#include <assert.h>
#include <stdio.h>
#include <lock.h>

#define Y 25
#define X 80
#define compress(x,y) ((x)+((y)*X))

__uint16 *device = (__uint16 *) 0xB8000;
lock_t vidLock;
short x=0, y=0;

void updateCursor();
bool updateScreen();

void putcColor(char character, __uint8 color)
{
	switch(character) {
	case '\n':		// Enter
		x = 0;
		y++;
		updateScreen();
		break;

	case '\b':		// Backspace
		x--;
		updateScreen();
		device[compress(x,y)] = 0;
		break;

	default:
		device[compress(x,y)] = color << 8 | character;
		x++;
		updateScreen();		// If it changed something
	}
}

bool updateScreen()
{
	bool ret = FALSE;

	lock(&vidLock);
	while(x >= X) {		// Down 1 line
		x -= X;
		y++;
		ret = TRUE;
	}

	while(x < 0) {		// Back 1 line
		x += X;
		y--;
		ret = TRUE;
	}

	if(y >= Y) {		// Scroll down
		y -= Y - 1;
		kmemcopy((char *)device, (char *)(device + X*y), 2*X*(Y-y));
		y = Y - 1;
		kmemset((char *)(device + X*y), 0, 2*X);
		ret = TRUE;
	}
	unlock(&vidLock);

	updateCursor();

	return ret;
}

void cls()		// This just sets the colour to black...
{
	lock(&vidLock);
	kmemset((char *)device, 0, 2*X*Y);
	unlock(&vidLock);

	x = 0;			// Cursor top-left corner
	y = 0;

	updateCursor();
}

// Set the hardware cursor
void updateCursor()
{
	outb(14, 0x3D4);
	outb(compress(x,y) >> 8, 0x3D5);
	outb(15, 0x3D4);
	outb(compress(x,y), 0x3D5);
}

// Initalizes the video driver
initFunc __preInitFunc videoInit()
{
	__uint16 a=0;

	lockInit(&vidLock);

	outb(14, 0x3D4);		// Get cursor pos
	a = inb(0x3D5) << 8;
	outb(15, 0x3D4);
	a |= inb(0x3D5);

	y = a / X;		// Set x and y
	x = 0;

	return SUCCESS;
}

EXPORT_SYMBOL(putcColor);
EXPORT_SYMBOL(cls);
